/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/
#ifndef __ANY_DISPLAY_H
#define __ANY_DISPLAY_H

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "lpc_types.h"
#include "ea_lcd_board.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define BACK_BUFFER_INVALID  0

typedef int32_t BackBuffer_t;

typedef struct
{
  uint16_t      width;
  uint16_t      height;
  uint32_t      buffersize;  // width*height*bitsPerPixel
  int32_t       framebuffer;
  int32_t       dev_lcd;
} AnyDisplayParams_t;


/******************************************************************************
 * Public Functions
 *****************************************************************************/

/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Attempts to identify which type of display is connected to the
 *    OemBaseBoard by reading the display board's EEPROM. If the display
 *    cannot be identified it defaults to the 3.2" display because it does
 *    not have an EEPROM.
 *
 *    The display is then initialized.
 *
 * Params:
 *    [in] framebuffer   - A memory region to use as the frame buffer for
 *                         the display
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_init(int32_t framebuffer);


/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Returns the display parameters such as width and height.
 *
 * Params:
 *    [out] pParams   - The parameters
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_getParameters(AnyDisplayParams_t* pParams);


/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Returns the name of the display
 *
 * Params:
 *    [in] buf    - The buffer to write the name into. Must exist.
 *    [in] len    - The size of the buffer
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_getName(char* buf, uint32_t len);


/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Returns the name of the display's manufacturer
 *
 * Params:
 *    [in] buf    - The buffer to write the name into. Must exist.
 *    [in] len    - The size of the buffer
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_getManufacturer(char* buf, uint32_t len);


/*****************************************************************************
 *
 * Description:
 *    Allocates memory for a back buffer the size of the display.
 *    The back buffer will take width*height*2 bytes of RAM.
 *
 *    The back buffer is guaranteed to be 8-byte aligned which is
 *    what is required according to the User's Manual.
 *
 *    Use anydisplay_freeBackBuffer() to release the buffer when it is
 *    no longer needed.
 *
 * Params:
 *    None
 *
 * Return:
 *    The allocated back buffer, or NULL if there was not enough free RAM.
 *
 ****************************************************************************/
BackBuffer_t anydisplay_createBackBuffer(void);


/*****************************************************************************
 *
 * Description:
 *    Allocates memory for a each of the specified pointers that are not NULL.
 *    The back buffers will take width*height*2 bytes of RAM each.
 *
 *    The back buffers are guaranteed to be 8-byte aligned which is
 *    what is required according to the User's Manual.
 *
 *    Use anydisplay_freeBackBuffers() to release the buffer when they are
 *    no longer needed.
 *
 *    If one of the requested back buffers cannot be allocated then all the
 *    already allocated buffers are released.
 *
 * Params:
 *    [out] pBB1   - Pointer to back buffer, or NULL to not allocate anything
 *    [out] pBB2   - Pointer to back buffer, or NULL to not allocate anything
 *    [out] pBB3   - Pointer to back buffer, or NULL to not allocate anything
 *
 * Return:
 *    TRUE if all wanted back buffers were allocated.
 *
 ****************************************************************************/
Bool anydisplay_createBackBuffers(BackBuffer_t* pBB1, BackBuffer_t* pBB2, BackBuffer_t* pBB3);


/*****************************************************************************
 *
 * Description:
 *    Releases the allocates memory for the specified back buffer if
 *    it is not NULL.
 *
 * Params:
 *    [in] buff   - The back buffer
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void anydisplay_freeBackBuffer(BackBuffer_t buff);


/*****************************************************************************
 *
 * Description:
 *    Releases the allocates memory for each of the non-NULL back buffers.
 *
 * Params:
 *    [in] pBB1   - Pointer to back buffer
 *    [in] pBB1   - Pointer to back buffer
 *    [in] pBB1   - Pointer to back buffer
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void anydisplay_freeBackBuffers(BackBuffer_t* pBB1, BackBuffer_t* pBB2, BackBuffer_t* pBB3);


/*****************************************************************************
 *
 * Description:
 *    Makes the specified back buffer the active one (i.e. the one that the
 *    display is actually using).
 *
 * Params:
 *    [in] buff   - The back buffer
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void anydisplay_activateBuffer(BackBuffer_t buff);

#endif /* end __ANY_DISPLAY_H */

