/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/
#ifndef __IMAGE_H
#define __IMAGE_H

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "lpc_types.h"
#include "trans_fs.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

typedef enum
{
  IMAGE_FORMAT_BMP,
  IMAGE_FORMAT_PNG,
  IMAGE_FORMAT_JPEG,
  IMAGE_FORMAT_UNKNOWN,
} ImageFormat_t;

typedef enum
{
  IMAGE_ERROR_OK,
  IMAGE_ERROR_WRONG_FORMAT,
  IMAGE_ERROR_FSOPEN,
  IMAGE_ERROR_MALLOC,
  IMAGE_ERROR_FSSIZE_ZERO,
  IMAGE_ERROR_BMP_DECODE,
  IMAGE_ERROR_PNG_DECODE,
  IMAGE_ERROR_JPEG_DECODE,
  IMAGE_ERROR_BAD_IMAGE_SIZE,
} ImageError_t;

/******************************************************************************
 * Public Functions
 *****************************************************************************/

/*****************************************************************************
 *
 * Description:
 *    Loads an image file from the ROM file system and then decodes it
 *    into the 565 format used by the display.
 *
 *    Memory will be allocated for the new image and returned in the
 *    ppDecoded parameter. If the return value is IMAGE_ERROR_OK then it
 *    is the responsibility of the caller to deallocate (using lpc_free())
 *    the image when it is no longer needed.
 *
 * Params:
 *    [in]  format     - The type of image file. If UNKNOWN then the file
 *                       extension will be looked at
 *    [in]  pFileName  - The name of the file to decode.
 *    [out] ppDecoded  - The decoded image in 565 format
 *    [out] pWidth     - The width of the decoded image
 *    [out] pHeight    - The height of the decoded image
 *
 * Return:
 *    IMAGE_ERROR_OK if an image was decoded and stored in ppDecoded,
 *    otherwise one of the IMAGE_ERROR_* error codes.
 *
 ****************************************************************************/
ImageError_t decodeImageFileRomFS(ImageFormat_t format,
    char* pFileName,
    uint8_t** ppDecoded,
    uint16_t* pWidth,
    uint16_t* pHeight);


/*****************************************************************************
 *
 * Description:
 *    Loads an image file from one of the available file systems and then
 *    decodes it into the 565 format used by the display.
 *
 *    The fileLocation parameter decides the search order used when attempting
 *    to locate the image file.
 *
 *    Memory will be allocated for the new image and returned in the
 *    ppDecoded parameter. If the return value is IMAGE_ERROR_OK then it is
 *    the responsibility of the caller to deallocate (using lpc_free()) the
 *    image when it is no longer needed.
 *
 * Params:
 *    [in]  fileLocation  - Which location (FATFS, ROMFS) to start searching in
 *    [in]  pFileName     - The name of the file to decode.
 *    [out] ppDecoded     - The decoded image in 565 format
 *    [out] pWidth        - The width of the decoded image
 *    [out] pHeight       - The height of the decoded image
 *
 * Return:
 *    IMAGE_ERROR_OK if an image was decoded and stored in ppDecoded,
 *    otherwise one of the IMAGE_ERROR_* error codes.
 *
 ****************************************************************************/
ImageError_t decodeImageFile(
    tFsOverride fileLocation,
    char* pFileName,
    uint8_t** ppDecoded,
    uint16_t* pWidth,
    uint16_t* pHeight);


#endif /* end __IMAGE_H */

