/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/
#ifndef __ROM_FS_H
#define __ROM_FS_H

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "lpc_types.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

/* Maximum file name length */
#define FS_FILENAME_MAX 256

/* Flag bits */
#define FS_MASK_TYPE 0x01 /* --- BIT 0 --- */
#define  FS_DIR_TYPE 0x01 /* File is a directory */
#define  FS_REG_TYPE 0x00 /* File is a NOT a directory */
#define FS_MASK_LAST 0x02 /* --- BIT 1 --- */
#define  FS_IS_LAST  0x02 /* This is the last entry in the directory */
#define  FS_NOT_LAST 0x00 /* This is not the last entry */

/*
 * File meta structure, as stored in a directory...
 *  - A directory is a special kind of file, ordered as an
 *    array of file meta structures.
 *  - The file system begins with a meta structure, describing the
 *    root directory.
 */
typedef struct tag_FSFile_t
{
  uint32_t __fileStart;
  uint32_t __fileNameStart;
  uint32_t __fileSize;
  uint16_t __fileNameLength;
  uint16_t __fileFlags;
} FSFile_t;


/* return structure from fsOpen. */
typedef struct
{
  Bool         __isUsed;

  Bool         __seenLast;     /* Last directory entry has been seen */
  FSFile_t    *__pEntry;       /* Identifies this file */
  uint8_t     *__pName;        /* Name of current directory entry */
  uint32_t    __offset;        /* Current read position within the file */
  Bool        __flags;         /* file start */
  uint32_t    __size;
} File_t;

/* A directory is a special kind of file.  */
typedef File_t Dir_t;


/* return structure from fsReadDir */
typedef File_t DirEnt_t;

/******************************************************************************
 *
 * Description:
 *    Get the size of an open file.
 *
 * Prototype:
 *    uint16_t fsFileSize(File_t *pFile)
 *
 * Params:
 *    [in] pFile - file to check the size of.
 *
 * Returns:
 *    the size of the file
 *
 *****************************************************************************/
#define fsFileSize(pFile) ((pFile)->__pEntry->__fileSize)

/******************************************************************************
 *
 * Description:
 *    Get the type of a file.
 *
 * Prototype:
 *    uint8_t fsFileType(File_t *pFile)
 *
 * Params:
 *    [in] pFile - file to check the type of.
 *
 * Returns:
 *    0 if it is a file. 1 if it is a directory.
 *
 *****************************************************************************/
#define fsFileType(pFile) ((pFile)->__pEntry->__fileFlags & FS_MASK_TYPE)

/******************************************************************************
 *
 * Description:
 *    Check if a file is a directory
 *
 * Prototype:
 *    Bool fsIsDirectory(File_t *pFile)
 *
 * Params:
 *    [in] pFile - file to check if it is a directory.
 *
 * Returns:
 *    TRUE if a directory; otherwise FALSE.
 *
 *****************************************************************************/
#define fsIsDirectory(pFile) ( fsFileType(pFile) == FS_DIR_TYPE )

/******************************************************************************
 *
 * Description:
 *    Get the name of a directory.
 *
 * Prototype:
 *    uint8_t* fsDirEntName(Dir_t *pDirEnt)
 *
 * Params:
 *    [in] pDir - directory to get the name of.
 *
 * Returns:
 *    name of the directory.
 *
 *****************************************************************************/
#define fsDirEntName(pDirEnt)  (pDirEnt)->__pName


/******************************************************************************
 * Public functions
 *****************************************************************************/


/*****************************************************************************
 *
 * Description:
 *    Initialize the ROM file system.
 *
 * Params:
 *    [in] pImage - Address of ROM file system image.
 *
 ****************************************************************************/
void fsInit(void* pImage);


/*****************************************************************************
 *
 * Description:
 *    Open a file.
 *
 * Params:
 *    [in] pName - Absolute path name to the file that should be opened.
 *                 Directory names end with a '/'.
 *
 * Returns:
 *    Reference to file meta structure
 *
 ****************************************************************************/
File_t* fsOpen(uint8_t* pName);


/*****************************************************************************
 *
 * Description:
 *    Close an open file.
 *
 * Params:
 *    [in] pFile - An open file to be closed.
 *
 ****************************************************************************/
void fsClose(File_t* pFile);


/*****************************************************************************
 *
 * Description:
 *    Return a reference to the file data, which is located as a continuous
 *    array in memory.
 *
 * Params:
 *    [in]  pFile - Open file.
 *    [out] pSize - The size of the file.
 *
 * Returns:
 *    A reference to the file data.
 *
 ****************************************************************************/
uint8_t* fsGetFile(File_t * pFile,
               uint32_t *  pSize);


/*****************************************************************************
 *
 * Description:
 *    Open a directory for reading.
 *
 * Params:
 *    [in] pName - Absolute path name to the directory.
 *
 * Returns:
 *    A reference to the directory.
 *
 ****************************************************************************/
Dir_t * fsOpenDir(uint8_t * pName);


/*****************************************************************************
 *
 * Description:
 *    Fetch the next entry from a directory.
 *
 * Params:
 *    [in] pDir - An open directory.
 *
 * Returns:
 *    A reference to an entry structure. The __d_name member will indicate the
 *    name of this entry.
 *
 ****************************************************************************/
DirEnt_t * fsReadDir(Dir_t * pDir);


/*****************************************************************************
 *
 * Description:
 *    Checks if a directory entry is a directory or a file.
 *
 * Params:
 *    [in] pDir - A directory entry
 *
 * Returns:
 *    TRUE of the entry is a directory; otherwise FALSE
 *
 ****************************************************************************/
Bool
fsIsEntryDir(DirEnt_t *pDir);

/*****************************************************************************
 *
 * Description:
 *    Get the size of a directory entry
 *
 * Params:
 *    [in] pDir - A directory entry
 *
 * Returns:
 *    the size of the entry
 *
 ****************************************************************************/
uint32_t
fsEntrySize(DirEnt_t *pDir);

/*****************************************************************************
 *
 * Description:
 *    Close an open directory.
 *
 * Params:
 *    [in] pDir - The open directory to close.
 *
 ****************************************************************************/
void fsCloseDir(Dir_t* pDir);

#endif /* end __ROM_FS_H */
