/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include "any_display.h"
#include "ea_lcd_board.h"
#include "truly_display.h"
#include "log.h"
#include "lpc_heap.h"
#include "lcd_driver.h"
#include "globals.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/


/******************************************************************************
 * Global variables
 *****************************************************************************/


/******************************************************************************
 * Local variables
 *****************************************************************************/

static AnyDisplayParams_t DisplayParams;
static Bool Initialized = FALSE;

/******************************************************************************
 * Local Functions
 *****************************************************************************/


/******************************************************************************
 * Public Functions
 *****************************************************************************/


/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Attempts to identify which type of display is connected to the
 *    OemBaseBoard by reading the display board's EEPROM. If the display
 *    cannot be identified it defaults to the 3.2" display because it does
 *    not have an EEPROM.
 *
 *    The display is then initialized.
 *
 * Params:
 *    [in] framebuffer   - A memory region to use as the frame buffer for
 *                         the display
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_init(int32_t framebuffer)
{
  LCD_PARAM_T lp;
  int32_t dev_lcd = 0;
  int32_t width = 0;
  int32_t height = 0;

  lcdb_result_t result = LCDB_RESULT_OK;
  if (!Initialized)
  {
    do {
      if ((result = ea_lcdb_open(NULL, NULL, &dev_lcd)) == LCDB_RESULT_OK) {

        if ((result = ea_lcdb_getLcdParams(&lp)) != LCDB_RESULT_OK) {
          log_e("ea_lcdb_getLcdParams FAILED (%d)\r\n", result);
          break;
        }

        width  = lp.pixels_per_line;
        height = lp.lines_per_panel;

        memset((void*)(framebuffer), 0x00, width*height*2);

        if ((result = ea_lcdb_enable(dev_lcd, framebuffer)) != LCDB_RESULT_OK) {
          log_e("ea_lcdb_enable FAILED (%d)\r\n", result);
          break;
        }
      }
      else {
        /* Assuming that the unidentified display is a QVGA display */
        width  = 240;
        height = 320;
        memset((void*)(framebuffer), 0x00, width*height*2);

        if ((dev_lcd = truly_lcd_open(framebuffer)) == 0) {
          log_e("truly_lcd_open FAILED\r\n");
          break;
        }

        log_i("truly_lcd_open: Assuming it is EA QVGA display\r\n");
        result = LCDB_RESULT_OK;
      }
      DisplayParams.width = width;
      DisplayParams.height = height;
      DisplayParams.buffersize = width * height * 2;
      DisplayParams.framebuffer = framebuffer;
      DisplayParams.dev_lcd = dev_lcd;
      Initialized = TRUE;
    } while(0);
  }
  return result;
}


/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Returns the display parameters such as width and height.
 *
 * Params:
 *    [out] pParams   - The parameters
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_getParameters(AnyDisplayParams_t* pParams)
{
  pParams->width = 0;
  pParams->height = 0;

  if (Initialized) {
    *pParams = DisplayParams;
  }
  return LCDB_RESULT_OK;
}

/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Returns the name of the display
 *
 * Params:
 *    [in] buf    - The buffer to write the name into. Must exist.
 *    [in] len    - The size of the buffer
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_getName(char* buf, uint32_t len)
{
  /*
   * Retrieves the name of the display. If it cannot be read from EEPROM then
   * it is assumed to be the Truly display.
   */
  if (ea_lcdb_getDisplayName(buf, len) != LCDB_RESULT_OK) {
    snprintf(buf, len, "TFT-G240320LTSW-118W-E");
  }
  return LCDB_RESULT_OK;
}


/*****************************************************************************
 *
 * Description:
 *    A common interface to the three supported displays: 3.2", 4.3" and 7".
 *
 *    Returns the name of the display's manufacturer
 *
 * Params:
 *    [in] buf    - The buffer to write the name into. Must exist.
 *    [in] len    - The size of the buffer
 *
 * Return:
 *    LCDB_RESULT_OK if the display was initialized,
 *    otherwise on of the LCD* error codes.
 *
 ****************************************************************************/
lcdb_result_t anydisplay_getManufacturer(char* buf, uint32_t len)
{
  /*
   * Retrieves the name of the display manufacturer. If it cannot be
   * read from EEPROM then it is assumed to be the Truly display.
   */
  if (ea_lcdb_getDisplayMfg(buf, len) != LCDB_RESULT_OK) {
    snprintf(buf, len, "TRULY SEMICONDUCTORS");
  }
  return LCDB_RESULT_OK;
}

/*****************************************************************************
 *
 * Description:
 *    Allocates memory for a back buffer the size of the display.
 *    The back buffer will take width*height*2 bytes of RAM.
 *
 *    The back buffer is guaranteed to be 8-byte aligned which is
 *    what is required according to the User's Manual.
 *
 *    Use anydisplay_freeBackBuffer() to release the buffer when it is
 *    no longer needed.
 *
 * Params:
 *    None
 *
 * Return:
 *    The allocated back buffer, or NULL if there was not enough free RAM.
 *
 ****************************************************************************/
BackBuffer_t anydisplay_createBackBuffer(void)
{
  uint32_t* bb = lpc_new(DisplayParams.buffersize + 4);
  if (bb == NULL)
  {
    return BACK_BUFFER_INVALID;
  }
  if ((((uint32_t)bb) % 8) == 4)
  {
    // lpc_new has returned a 4-byte aligned address,
    // advance pointer 4 bytes to get it 8-byte aligned.
    bb++;
  }
  return (BackBuffer_t) bb;
}


/*****************************************************************************
 *
 * Description:
 *    Allocates memory for a each of the specified pointers that are not NULL.
 *    The back buffers will take width*height*2 bytes of RAM each.
 *
 *    The back buffers are guaranteed to be 8-byte aligned which is
 *    what is required according to the User's Manual.
 *
 *    Use anydisplay_freeBackBuffers() to release the buffer when they are
 *    no longer needed.
 *
 *    If one of the requested back buffers cannot be allocated then all the
 *    already allocated buffers are released.
 *
 * Params:
 *    [out] pBB1   - Pointer to back buffer, or NULL to not allocate anything
 *    [out] pBB2   - Pointer to back buffer, or NULL to not allocate anything
 *    [out] pBB3   - Pointer to back buffer, or NULL to not allocate anything
 *
 * Return:
 *    TRUE if all wanted back buffers were allocated.
 *
 ****************************************************************************/
Bool anydisplay_createBackBuffers(BackBuffer_t* pBB1, BackBuffer_t* pBB2, BackBuffer_t* pBB3)
{
  if (pBB1 != NULL)
  {
    *pBB1 = anydisplay_createBackBuffer();
    if (*pBB1 == BACK_BUFFER_INVALID)
    {
      return FALSE;
    }
  }
  if (pBB2 != NULL)
  {
    *pBB2 = anydisplay_createBackBuffer();
    if (*pBB2 == BACK_BUFFER_INVALID)
    {
      return FALSE;
    }
  }
  if (pBB3 != NULL)
  {
    *pBB3 = anydisplay_createBackBuffer();
    if (*pBB3 == BACK_BUFFER_INVALID)
    {
      return FALSE;
    }
  }
  return TRUE;
}


/*****************************************************************************
 *
 * Description:
 *    Releases the allocates memory for the specified back buffer if
 *    it is not NULL.
 *
 * Params:
 *    [in] buff   - The back buffer
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void anydisplay_freeBackBuffer(BackBuffer_t buff)
{
  if (buff != BACK_BUFFER_INVALID)
  {
    if (!lpc_free((void*)buff))
    {
      // Found no allocated memory at buff. This is
      // probably because the anydisplay_createBackBuffer
      // function aligned it to an 8-byte boundary by
      // adding 4 bytes.
      lpc_free((void*)(buff-4));
    }
  }
}


/*****************************************************************************
 *
 * Description:
 *    Releases the allocates memory for each of the non-NULL back buffers.
 *
 * Params:
 *    [in] pBB1   - Pointer to back buffer
 *    [in] pBB1   - Pointer to back buffer
 *    [in] pBB1   - Pointer to back buffer
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void anydisplay_freeBackBuffers(BackBuffer_t* pBB1, BackBuffer_t* pBB2, BackBuffer_t* pBB3)
{
  if (pBB1 != NULL)
  {
    anydisplay_freeBackBuffer(*pBB1);
    *pBB1 = BACK_BUFFER_INVALID;
  }
  if (pBB2 != NULL)
  {
    anydisplay_freeBackBuffer(*pBB2);
    *pBB2 = BACK_BUFFER_INVALID;
  }
  if (pBB3 != NULL)
  {
    anydisplay_freeBackBuffer(*pBB3);
    *pBB3 = BACK_BUFFER_INVALID;
  }
}


/*****************************************************************************
 *
 * Description:
 *    Makes the specified back buffer the active one (i.e. the one that the
 *    display is actually using).
 *
 * Params:
 *    [in] buff   - The back buffer
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void anydisplay_activateBuffer(BackBuffer_t buff)
{
  lcd_ioctl(GlobalParameters.display.dev_lcd, LCD_SET_BUFFER, buff);
}
