/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "application.h"
#include "apptable.h"
#include "lpc_heap.h"
#include "log.h"
#include <string.h>
#include "lpc177x_8x_timer.h"
#include "globals.h"
#include "any_display.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define BRD_TIMER_USED     (LPC_TIM2)
#define BRD_TIM_INTR_USED  (TIMER2_IRQn)


/******************************************************************************
 * Global variables
 *****************************************************************************/


/******************************************************************************
 * Local variables
 *****************************************************************************/

static AppCallback_t* instances[APP_TABLE_SIZE];

static TIM_TIMERCFG_Type TIM_ConfigStruct;
static TIM_MATCHCFG_Type TIM_MatchConfigStruct;

static void (*timeoutFunc)(void) = NULL;

/******************************************************************************
 * Local functions
 *****************************************************************************/

/*****************************************************************************
 *
 * Description:
 *    Checks if the App has specified a timeout() function and a
 *    intervalInMs > 0 in it's callback information. If it has then TIMER2
 *    is configured to cause an interrupt with a period of the specified
 *    number of milliseconds. Each time the interrupt is called, the
 *    specified timeout() function will be called.
 *
 * Params:
 *    [in] pInstance   - The App's information
 *
 * Return:
 *    None
 *
 ****************************************************************************/
static void enableTimer(AppCallback_t* pInstance)
{
  if (pInstance->intervalInMs == 0 || pInstance->timeout == NULL) {
    return;
  }
  timeoutFunc = pInstance->timeout;

  // Initialize timer 0, prescale count time of 100uS
  TIM_ConfigStruct.PrescaleOption = TIM_PRESCALE_USVAL;
  TIM_ConfigStruct.PrescaleValue  = 100;

  // use channel 0, MR2
  TIM_MatchConfigStruct.MatchChannel = 2;
  // Enable interrupt when MR0 matches the value in TC register
  TIM_MatchConfigStruct.IntOnMatch   = TRUE;
  //Enable reset on MR2: TIMER will reset if MR2 matches it
  TIM_MatchConfigStruct.ResetOnMatch = TRUE;
  //Stop on MR2 if MR2 matches it
  TIM_MatchConfigStruct.StopOnMatch  = FALSE;
  //Toggle MR0.2 pin if MR2 matches it
  TIM_MatchConfigStruct.ExtMatchOutputType = TIM_EXTMATCH_NOTHING;//TIM_EXTMATCH_TOGGLE;
  // Set Match value, count value of 10000 (10000 * 100uS = 1000000us = 1s --> 1 Hz)
  TIM_MatchConfigStruct.MatchValue   = 10 * pInstance->intervalInMs;//10000;

  // Set configuration for Tim_config and Tim_MatchConfig
  TIM_Init(BRD_TIMER_USED, TIM_TIMER_MODE, &TIM_ConfigStruct);
  TIM_ConfigMatch(BRD_TIMER_USED, &TIM_MatchConfigStruct);

  /* preemption = 1, sub-priority = 1 */
  NVIC_SetPriority(BRD_TIM_INTR_USED, ((0x01<<3)|0x01));

  /* Enable interrupt for timer 2 */
  NVIC_EnableIRQ(BRD_TIM_INTR_USED);

  // To start timer
  TIM_Cmd(BRD_TIMER_USED, ENABLE);
}


/*****************************************************************************
 *
 * Description:
 *    Stop the App requested time if there is one.
 *
 * Params:
 *    None
 *
 * Return:
 *    None
 *
 ****************************************************************************/
static void disableTimer(void)
{
  if (timeoutFunc != NULL) {
    TIM_Cmd(BRD_TIMER_USED, DISABLE);
    NVIC_DisableIRQ(BRD_TIM_INTR_USED);
    timeoutFunc = NULL;
  }
}


/*****************************************************************************
 *
 * Description:
 *    Starts the App with the specified id (i.e. position in
 *    the "App Table" in apptable.h) and waits for it to end it's execution.
 *
 *    The sequence looks like this:
 *      1) Locate App
 *      2) Note how much memory is used before starting it
 *      3) If the App needs a timer, create it
 *      4) Run the App's start() function and wait for it to return
 *      5) If a timer was created in step 3, stop it
 *      6) Run the App's stop() function and wait for it to return
 *      7) Restore the frame buffer to avoid sloppy Apps from damaging
 *         the system
 *      8) Note how much memory is used now, and if it differs from step 2
 *         then print information to aid the developer in removing memory
 *         leaks.
 *
 * Params:
 *    [in] id   - ID of the App to start
 *
 * Return:
 *    TRUE if the App was started
 *
 ****************************************************************************/
static Bool executeApp(int appTableIndex)
{
  Bool result = FALSE;
  do
  {
    if (appTableIndex < 0 || appTableIndex >= APP_TABLE_SIZE)
    {
      log_e("Invalid application %d\r\n", appTableIndex);
      break;
    }
    if (instances[appTableIndex] != NULL)
    {
      log_e("Application already started\r\n");
      break;
    }

    instances[appTableIndex] = appTable[appTableIndex].init();

    log_i("Creating app %d - %s, prio=%d\r\n", appTableIndex,
        appTable[appTableIndex].pName,
        appTable[appTableIndex].prio);

    if (instances[appTableIndex] != NULL)
    {
      // Prepare and then start application
      uint32_t mem_size = lpc_get_allocatedsize();
      log_i("Starting app %d with handle 0x%x\r\n", appTableIndex, instances[appTableIndex]);
      enableTimer(instances[appTableIndex]);
      instances[appTableIndex]->stopped = FALSE;
      instances[appTableIndex]->start();

      // App has finished. Clean up.
      disableTimer();
      instances[appTableIndex]->stopped = TRUE;
      instances[appTableIndex]->stop();
      instances[appTableIndex] = NULL;
      anydisplay_activateBuffer(GlobalParameters.display.framebuffer);
      mem_size -= lpc_get_allocatedsize();
      if (mem_size != 0)
      {
        log_i("Application leaked %d bytes\r\n", mem_size);
      }
      result = TRUE;
    }
  } while (0);

  return result;
}

/******************************************************************************
 * Public Functions
 *****************************************************************************/

/*****************************************************************************
 *
 * Description:
 *    IRQ handler function for the timer used by executeApp().
 *
 *    This is used to call the App specific timeout() function.
 *
 * Params:
 *    None
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void TIMER2_IRQHandler(void)
{
  if (TIM_GetIntStatus(BRD_TIMER_USED, TIM_MR2_INT)== SET)
  {
    if (timeoutFunc != NULL) {
      timeoutFunc();
    }
  }

  TIM_ClearIntPending(BRD_TIMER_USED, TIM_MR2_INT);
}


/*****************************************************************************
 *
 * Description:
 *    Executes the first App in the "App Table" (found in appTable.h) marked as
 *    autoStart=TRUE.
 *
 *    This function is called from the main() function to start the default
 *    App, which is normally the menu system.
 *
 * Params:
 *    None
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void appStartInitial(void)
{
  int i = 0;

  for (i = 0; i < APP_TABLE_SIZE; i++)
  {
    if (appTable[i].autoStart)
    {
      executeApp(i);
      break;
    }
  }
}


/*****************************************************************************
 *
 * Description:
 *    Starts the App with the specified name.
 *
 * Params:
 *    [in] pName   - Name of the App to start
 *
 * Return:
 *    TRUE if the App was started
 *
 ****************************************************************************/
Bool appStartApp(const char* pName)
{
  int i = 0;

  for (i = 0; i < APP_TABLE_SIZE; i++)
  {
    if (strcmp(appTable[i].pName, pName) == 0)
    {
      return executeApp(i);
    }
  }
  return FALSE;
}


/*****************************************************************************
 *
 * Description:
 *    Starts the App with the specified id (i.e. position in the "App Table"
 *    in apptable.h).
 *
 * Params:
 *    [in] id   - ID of the App to start
 *
 * Return:
 *    TRUE if the App was started
 *
 ****************************************************************************/
Bool appStartAppID(int id)
{
  return executeApp(id);
}


/*****************************************************************************
 *
 * Description:
 *    Stops the currently running App.
 *
 * Params:
 *    None
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void appClose(void)
{
}


/*****************************************************************************
 *
 * Description:
 *    Returns the name of the App with the specified id (i.e. position in
 *    the "App Table" in apptable.h).
 *
 *    Note that the returned string is not allocated and hence should not
 *    be freed.
 *
 * Params:
 *    [in] id   - ID of the App to get the name for
 *
 * Return:
 *    The name of the App, or NULL if the App could not be found.
 *
 ****************************************************************************/
const char* appGetAppName(int id)
{
  if ((id >= 0) && (id < APP_TABLE_SIZE)) {
    return appTable[id].pName;
  }
  return NULL;
}
