/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "application.h"
#include <stdlib.h>
#include <string.h>
#include "image.h"
#include "log.h"
#include "lpc_swim.h"
#include "globals.h"
#include "swim_extensions.h"
#include "joystick.h"
#include "lpc177x_8x_timer.h"
#include "pca9532.h"
#include "lpc_heap.h"
#include "acc.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define BACKGROUND_COLOR          BLACK  //0x5800
#define BALL_COLOR                0x03ff //Color of player's ball
#define FALL_COLOR                0x0000 //Fall color

#define LARGE_CIRCLE_COLOR        0x39e7 //DARK_GRAY
#define SMALL_CIRCLE_FRONT_COLOR  WHITE
#define SMALL_CIRCLE_BACK_COLOR   0x7bef //LIGHT_GRAY

// Screen width
#define WIDTH GlobalParameters.display.width
// Screen height
#define HEIGHT GlobalParameters.display.height

// Delay between frame updates (20 = 20 ms = 50Hz)
#define RENDER_DELAY_MS 30

// Radius of the player's ball
#define RADIUS 12

// Tile size (width and height)
#define TILESIZE 32

// Player's thrust value
#define THRUST 550 //450

// Slowdown due to friction, etc.
#define SLOWDOWN 970 //885

// Slowdown due to collision
#define COLLISIONSLOWDOWN 980

#define NUM_OF_DOTS 10

#define DEG_0   0x0000
#define DEG_45  0x2000
#define DEG_90  0x4000
#define DEG_135 0x6000
#define DEG_180 0x8000
#define DEG_225 0xA000
#define DEG_270 0xC000
#define DEG_315 0xE000

#define GAME_ICON_START_PATH "/games/ballTexture"
#define GAME_ICON_END_PATH ".png"


typedef struct
{
  uint16_t* pIconData;
} tIcons;


/******************************************************************************
 * Public variables
 *****************************************************************************/


/******************************************************************************
 * Local variables
 *****************************************************************************/

static BackBuffer_t myLcdHnd1 = BACK_BUFFER_INVALID;
static BackBuffer_t myLcdHnd2 = BACK_BUFFER_INVALID;
static BackBuffer_t myLcdHnd3 = BACK_BUFFER_INVALID;
static BackBuffer_t myBackground = BACK_BUFFER_INVALID;

static uint8_t   activeArea;
static int16_t  zeroAccX;
static int16_t  zeroAccY;

static SWIM_WINDOW_T MainWin;
static Bool started = FALSE;
static Bool wantToShutDown = FALSE;

static tIcons
icons[] =
{
  {NULL},
  {NULL},
  {NULL}
};

// Width of level data
static int levelWidth = 7;
static int levelLeftOffset = 8;

// Height of level data
static int levelHeight = 10;
static int levelTopOffset = 0;
static uint16_t* gLevelBackground = NULL;

// Player's position * 1000
static int32_t gXPos;
static int32_t gYPos;

// Player's motion vector * 1000
static int32_t gXMov;
static int32_t gYMov;

static int32_t gRoll;

// Player's key status
static int gKeyLeft;
static int gKeyRight;
static int gKeyUp;
static int gKeyDown;

// Level data
static unsigned char* gLevel = NULL;

// Vertex structure
typedef struct
{
  long x, y, z;
}tVertex;

// Transformed vertices
static tVertex *gRVtx;

static const tVertex gVtx[NUM_OF_DOTS] = {
  {-21585, 17597, -17265},
  {28493, -7790, 14183},
  {13031, 27845, 11338},
  {10822, 29162, -10304},
  {19517, -25865, -4876},
  {1283, 30949, -10687},
  {-23097, 12297, 19723},
  {-17162, 1188, 27888},
  {-29007, -1547, 15163},
  {-8182, -31729, -197}
};

/******************************************************************************
 * Local function prototypes
 *****************************************************************************/

static void start(void);
static void stop(void);
static void timeout(void);
static void handleAcc(int16_t x, int16_t y);


/******************************************************************************
 * Local functions
 *****************************************************************************/

static unsigned short
iatan2(int a, int b)
{
  if ((a == 0) && (b >= 0))
    return DEG_90;
  else if ((a == 0) && (b < 0))
    return DEG_270;
  else if ((b == 0) && (a >= 0))
    return DEG_0;
  else if ((b == 0) && (a < 0))
    return DEG_180;

  if ((a > 0) && (b > 0) && (a > b))
  {
    //between 0..45 degrees
    return DEG_0 + 0x1000;
  }
  else if ((a > 0) && (b > 0) && (a <= b))
  {
    //between 45..90 degrees
    return DEG_45 + 0x1000;
  }
  else if ((a < 0) && (b > 0) && (-a <= b))
  {
    //between 90..135 degrees
    return DEG_90 + 0x1000;
  }
  else if ((a < 0) && (b > 0) && (-a > b))
  {
    //between 135..180 degrees
    return DEG_135 + 0x1000;
  }
  else if ((a < 0) && (b < 0) && (-a > -b))
  {
    //between 180..225 degrees
    return DEG_180 + 0x1000;
  }
  else if ((a < 0) && (b < 0) && (-a <= -b))
  {
    //between 225..270 degrees
    return DEG_225 + 0x1000;
  }
  else if ((a > 0) && (b < 0) && (a < -b))
  {
    //between 270..315 degrees
    return DEG_270 + 0x1000;
  }
  else if ((a > 0) && (b < 0) && (a >= -b))
  {
    //between 315..360 degrees
    return DEG_315 + 0x1000;
  }
  else
    return DEG_0;
}

static void
drawball(int32_t myLcdHnd, int x, int y, int r, int colorb, int color0,
         int color1, uint32_t roty, uint32_t rotz)
{
  swimex_fill_circle(&MainWin, x, y, r, colorb);
}

static void
uiLoadIcons(void)
{
  int i = 0;
  char path[30];
  ImageError_t result;
  uint16_t width;
  uint16_t height;
  uint8_t* pTmp;

  for (i = 0; i < 3; i++)
  {
    pTmp = NULL;
    sprintf(path, GAME_ICON_START_PATH "%d" GAME_ICON_END_PATH, i+1);
    result = decodeImageFile(ONLY_USE_ROMFS, path, &pTmp, &width, &height);
    if (result == IMAGE_ERROR_OK)
    {
    	icons[i].pIconData = (uint16_t*)pTmp;
    }
    else
    {
    	log_i("[%s:%d] Failed to load icon %s, result=%d\n", __FILE__, __LINE__, path, result);
    }
  }
}

static void
init()
{
  int i, j;

  gXPos = (WIDTH  / 2) * 1000;
  gYPos = (HEIGHT / 2) * 1000;

  gXMov = 0;
  gYMov = 0;

  gRoll = 0;

  gKeyLeft  = 0;
  gKeyRight = 0;
  gKeyUp    = 0;
  gKeyDown  = 0;

  gRVtx = (tVertex*)lpc_new(sizeof(tVertex) * NUM_OF_DOTS);

  /*
   * Prepare the "level" which is a levelWidth by levelHeight
   * array where boundaries are marked as 2 and open space
   * with 0.
   *
   * This first setup uses just four surrounding walls.
   */
  levelWidth = GlobalParameters.display.width / TILESIZE;
  levelHeight = GlobalParameters.display.height / TILESIZE;
  levelLeftOffset = (GlobalParameters.display.width - levelWidth * TILESIZE) / 2;
  levelTopOffset = (GlobalParameters.display.height - levelHeight * TILESIZE) / 2;
  gLevel = lpc_new(levelWidth * levelHeight);
  memset(gLevel, 2, levelWidth * levelHeight);
  for (i = levelHeight - 2; i > 0; i--)
  {
    for (j = levelWidth - 2; j > 0; j--)
    {
      gLevel[i*levelWidth+j] = 0;
    }
  }
}

static void render(int32_t myLcdHnd, uint32_t tick)
{
  MainWin.fb = (COLOR_T*)myLcdHnd;

  if (gKeyLeft)  gXMov -= gKeyLeft;
  if (gKeyRight) gXMov += gKeyRight;
  if (gKeyUp)    gYMov -= gKeyUp;
  if (gKeyDown)  gYMov += gKeyDown;

  gXMov = (gXMov * SLOWDOWN) / 1000;
  gYMov = (gYMov * SLOWDOWN) / 1000;

  gXPos += gXMov;
  gYPos += gYMov;

  // Collision with the screen borders
  if (((gXPos+500)/1000) > WIDTH - RADIUS - TILESIZE - levelLeftOffset - 1)
  {
    gXPos -= gXMov;
    gXMov = -(gXMov * COLLISIONSLOWDOWN) / 1000;
  }

  if (((gXPos+500)/1000) < RADIUS + TILESIZE + levelLeftOffset)
  {
    gXPos -= gXMov;
    gXMov = -(gXMov * COLLISIONSLOWDOWN) / 1000;
  }

  if (((gYPos+500)/1000) > HEIGHT - RADIUS - TILESIZE - levelTopOffset - 1)
  {
    gYPos -= gYMov;
    gYMov = -(gYMov * COLLISIONSLOWDOWN) / 1000;
  }

  if (((gYPos+500)/1000) < RADIUS + TILESIZE + levelTopOffset)
  {
    gYPos -= gYMov;
    gYMov = -(gYMov * COLLISIONSLOWDOWN) / 1000;
  }

  if (myLcdHnd != myBackground)
  {
    // Use the pre-rendered background
    memcpy((uint8_t*)myLcdHnd, (uint8_t*)myBackground, GlobalParameters.display.buffersize);
  }
  else
  {
    // rendering here (only done once during init)
    // fill background
    uint32_t i, j;
    swim_clear_screen(&MainWin, BACKGROUND_COLOR);
    for (i = 0; i < levelHeight; i++)
    {
      for (j = 0; j < levelWidth; j++)
      {
        swimex_put_image_at_position(&MainWin,
            (COLOR_T*)icons[gLevel[i * levelWidth + j]].pIconData,
            j * TILESIZE + levelLeftOffset,  //X-pos
            i * TILESIZE + levelTopOffset,   //y-pos
            TILESIZE,                        //x-width
            TILESIZE);                       //y-width
      }
    }
    // don't want to render the ball on the background
    return;
  }

  memcpy(gRVtx, gVtx, sizeof(tVertex) * NUM_OF_DOTS);

  // draw the player object
  drawball(myLcdHnd, (gXPos+500)/1000, (gYPos+500)/1000, RADIUS, BALL_COLOR,
           SMALL_CIRCLE_BACK_COLOR, SMALL_CIRCLE_FRONT_COLOR, gRoll/10,
           iatan2((gYPos+500)/1000, (gXPos+500)/1000));

  gRoll = 0;
}

static void
start(void)
{
  uint8_t joyState = 0;
  int8_t x;
  int8_t y;
  int8_t z;

  init();

  if (!anydisplay_createBackBuffers(&myLcdHnd1, &myLcdHnd2, &myLcdHnd3))
  {
    log_e("Failed to allocate back buffer\r\n");
    return;
  }
  myBackground = anydisplay_createBackBuffer();
  if (myBackground == BACK_BUFFER_INVALID)
  {
    log_e("Failed to allocate back buffer\r\n");
    return;
  }

  /* Create a SWIM window */
  swim_window_open(&MainWin, GlobalParameters.display.width,
      GlobalParameters.display.height, (COLOR_T*)myLcdHnd1, 0, 0,
      (GlobalParameters.display.width - 1),
      (GlobalParameters.display.height - 1), 0, WHITE, BACKGROUND_COLOR, BACKGROUND_COLOR);

  /* Clear all three buffers */
  MainWin.fb = (COLOR_T*)myLcdHnd1;
  swim_clear_screen(&MainWin, BACKGROUND_COLOR);
  MainWin.fb = (COLOR_T*)myLcdHnd2;
  swim_clear_screen(&MainWin, BACKGROUND_COLOR);
  MainWin.fb = (COLOR_T*)myLcdHnd3;
  swim_clear_screen(&MainWin, BACKGROUND_COLOR);
  MainWin.fb = (COLOR_T*)myLcdHnd3;

  uiLoadIcons();

  render(myBackground, 0);
  render(myLcdHnd1, msClock);
  anydisplay_activateBuffer(myLcdHnd1);
  activeArea = 1;
  render(myLcdHnd2, msClock);

  acc_init();
  acc_read(&x, &y, &z);
  zeroAccX = x;
  zeroAccY = y;

  wantToShutDown = FALSE;
  started = TRUE;

  while(started)
  {
    TIM_Waitms(100);

    /* Update based on joystick position */
    joyState = joystick_read();
    gKeyLeft  = ((joyState & JOYSTICK_LEFT) != 0) ? THRUST : 0;
    gKeyRight = ((joyState & JOYSTICK_RIGHT)!= 0) ? THRUST : 0;
    gKeyUp    = ((joyState & JOYSTICK_UP)   != 0) ? THRUST : 0;
    gKeyDown  = ((joyState & JOYSTICK_DOWN) != 0) ? THRUST : 0;

    /* Update based on accelerometer changes */
    acc_read(&x, &y, &z);
    handleAcc(x, y);

    /* Check SW2 button which causes Exit */
    if (pca9532_getLedState(FALSE) & KEY1) {
      wantToShutDown = TRUE;
    }
  }
}

static void
stop(void)
{
  int i;
  started = FALSE;
  swim_window_close(&MainWin);

  anydisplay_freeBackBuffer(myLcdHnd1);
  anydisplay_freeBackBuffer(myLcdHnd2);
  anydisplay_freeBackBuffer(myLcdHnd3);
  anydisplay_freeBackBuffer(myBackground);
  myLcdHnd1 = myLcdHnd2 = myLcdHnd3 = myBackground = BACK_BUFFER_INVALID;

  lpc_free(gRVtx);
  gRVtx = NULL;

  lpc_free(gLevel);
  gLevel = NULL;

  lpc_free(gLevelBackground);
  gLevelBackground = NULL;

  for (i = 0; i < 3; i++) {
    if (icons[i].pIconData != NULL) {
      lpc_free(icons[i].pIconData);
      icons[i].pIconData = NULL;
    }
  }
}

static void
handleAcc(int16_t x, int16_t y)
{
  int16_t xDiff;
  int16_t yDiff;

  xDiff = x - zeroAccX;
  if ((xDiff < 8) && (xDiff > -8))
	xDiff = 0;
  yDiff = y - zeroAccY;
  if ((yDiff < 8) && (yDiff > -8))
	yDiff = 0;

	if (xDiff > 0)
	{
    gKeyDown = xDiff*15;
    gKeyUp   = 0;
	}
	else if (xDiff < 0)
	{
    gKeyUp   = -xDiff*15;
    gKeyDown = 0;
	}

	if (yDiff < 0)
	{
    gKeyLeft  = -yDiff*15;
    gKeyRight = 0;
	}
	else if (yDiff > 0)
	{
    gKeyRight = yDiff*15;
    gKeyLeft  = 0;
	}
}

static void timeout(void)
{
  if (started)
  {
    if (wantToShutDown) {
      started = FALSE;
    }
    else if (activeArea == 3)
    {
      anydisplay_activateBuffer(myLcdHnd1);
      render(myLcdHnd2, msClock);
      activeArea = 1;
    }
    else if (activeArea == 1)
    {
      anydisplay_activateBuffer(myLcdHnd2);
      render(myLcdHnd3, msClock);
      activeArea = 2;
    }
    else
    {
      anydisplay_activateBuffer(myLcdHnd3);
      render(myLcdHnd1, msClock);
      activeArea = 3;
    }
  }
}


/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

static AppCallback_t bouncingballCallback = {
    &start,
    &stop,
    &timeout,
    100,
    FALSE,
};

AppCallback_t* bouncingballAppInit(void)
{
  return &bouncingballCallback;
}

