/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include "application.h"
#include "image.h"
#include "globals.h"
#include "log.h"
#include "swim_extensions.h"
#include "lpc_heap.h"
#include "lpc_swim.h"
#include "lpc_swim_font.h"
#include "lpc_swim_image.h"
#include "joystick.h"
#include "pca9532.h"
#include "lpc177x_8x_timer.h"
#include "random.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define GAME_NOT_STARTED 0
#define GAME_RUNNING     1
#define GAME_PAUSE       3
#define GAME_OVER        4
#define GAME_END         5

#define BOARD_WIDTH  15
#define BOARD_HEIGHT 21

#define FIGURE_SIZE   4
#define NUM_OF_FIGURE 19

#define GREEN_FIGURE       1
#define RED_FIGURE         2
#define YELLOW_FIGURE      3
#define WHITE_FIGURE       4
#define BLUE_FIGURE        5

#define ICON_WIDTH   8
#define ICON_HEIGHT  8
#define ICON_BPP     24

#define GAME_ICON_START_PATH "/games/8block00"
#define GAME_ICON_END_PATH ".png"
#define GAME_BKG_PATH "/slides/240x320/short/05.jpg"   //"/games/background.png"

#define KEY_BACK  (KEY1)
#define JOY_MASK 0x1f
#define ANYTHING_PRESSED(_keystate, _joystate)  (((_keystate) & KEY_MASK) || ((_joystate) & JOY_MASK))


typedef struct
{
  uint16_t* pIconData;
} tIcons;


/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static SWIM_WINDOW_T GameWindow;
static SWIM_WINDOW_T MessageWin;
static Bool MessageWinActive = FALSE;

static uint32_t play;
static uint8_t gameStatus;
static uint8_t score;

static int32_t shift[NUM_OF_FIGURE]={1,0,3,4,5,2,7,8,9,6,11,12,13,10,15,14,17,16,18};
static int32_t board[BOARD_WIDTH][BOARD_HEIGHT];
static int32_t currFigure;
static int32_t currXpos;
static int32_t currYpos;
static uint8_t pause = 0;
static volatile uint32_t lastUpdate;
static volatile uint32_t delayTimeMs;

static BackBuffer_t gameLcdHnd1 = BACK_BUFFER_INVALID;
static BackBuffer_t gameLcdHnd2 = BACK_BUFFER_INVALID;

static uint8_t lastJoyState = 0;

static tIcons
icons[] =
{
  {NULL},
  {NULL},
  {NULL},
  {NULL},
  {NULL},
  {NULL},
  {NULL}
};

static unsigned char figures[NUM_OF_FIGURE][FIGURE_SIZE + 1][FIGURE_SIZE] =
{{".X..",
  ".X..",
  ".X..",
  ".X..",
  {2}},  //red
 {"....",
  "XXXX",
  "....",
  "....",
  {2}},

 {"XXX.",
  "X...",
  "....",
  "....",
  {1}},  //green
 {".XX.",
  "..X.",
  "..X.",
  "....",
  {1}},
 {"..X.",
  "XXX.",
  "....",
  "....",
  {1}},
 {".X..",
  ".X..",
  ".XX.",
  "....",
  {1}},
 {"XXX.",
  "..X.",
  "....",
  "....",
  {1}},
 {"..X.",
  "..X.",
  ".XX.",
  "....",
  {1}},
 {"X...",
  "XXX.",
  "....",
  "....",
  {1}},
 {".XX.",
  ".X..",
  ".X..",
  "....",
  {1}},

 {".X..",
  "XXX.",
  "....",
  "....",
  {3}},   //yellow
 {".X..",
  ".XX.",
  ".X..",
  "....",
  {3}},
 {"XXX.",
  ".X..",
  "....",
  "....",
  {3}},
 {".X..",
  "XX..",
  ".X..",
  "....",
  {3}},

 {"X...",
  "XX..",
  ".X..",
  "....",
  {4}},  //white
 {".XX.",
  "XX..",
  "....",
  "....",
  {4}},
 {".X..",
  "XX..",
  "X...",
  "....",
  {4}},
 {"XX..",
  ".XX.",
  "....",
  "....",
  {4}},

 {"XX..",
  "XX..",
  "....",
  "....",
  {5}}}; //blue


/******************************************************************************
 * Local function prototypes
 *****************************************************************************/

static void start(void);
static void stop(void);
static void timeout(void);
static void restartGame(void);
static void quitGame(void);

/******************************************************************************
 * Local functions
 *****************************************************************************/

static void
uiLoadIcons(void)
{
  int i = 0;
  char path[30];
  ImageError_t result;
  uint16_t width;
  uint16_t height;
  uint8_t* pTmp;

  for (i = 0; i < 6; i++)
  {
    pTmp = NULL;
    sprintf(path, GAME_ICON_START_PATH "%d" GAME_ICON_END_PATH, i+1);
    result = decodeImageFile(ONLY_USE_ROMFS, path, &pTmp, &width, &height);
    if (result == IMAGE_ERROR_OK)
    {
      icons[i].pIconData = (uint16_t*)pTmp;
    }
    else
    {
      log_i("[%s:%d] Failed to load icon %s, result=%d\r\n", __FILE__, __LINE__, path, result);
    }
  }

  //load background
  pTmp = NULL;
  result = decodeImageFile(ONLY_USE_ROMFS, GAME_BKG_PATH, &pTmp, &width, &height);
  if (result == IMAGE_ERROR_OK)
  {
    icons[6].pIconData = (uint16_t*)pTmp;
  }
  else
  {
    log_i("[%s:%d] Failed to load background %s, result=%d\r\n", __FILE__, __LINE__, GAME_BKG_PATH, result);
  }
}

/*****************************************************************************
 *
 * Description:
 *    Check if current place for figure is valid
 *
 ****************************************************************************/
static int8_t
isValid(int32_t x, int32_t y, int32_t shape)
{
  int32_t i,j;

  for (i=0; i<FIGURE_SIZE; i++)
    for (j=0; j<FIGURE_SIZE; j++)
      if((figures[shape][i][j] == 'X') &&
         ((i+x >= BOARD_WIDTH) || (j+y >= BOARD_HEIGHT) || (i+x < 0)))
        return 0;

  for (i=0; i<FIGURE_SIZE; i++)
    for (j=0; j<FIGURE_SIZE; j++)
      if((i+x < BOARD_WIDTH) &&
         (j+y < BOARD_HEIGHT) &&
         (figures[shape][i][j] == 'X') &&
         (board[i+x][j+y] != 0))
        return 0;

  return 1;
}


/*****************************************************************************
 *
 * Description:
 *    Draw/refresh game board
 *
 ****************************************************************************/
static void
drawGame(void)
{
  int32_t i,j;

  for (i=1; i<=BOARD_WIDTH; i++)
    for (j=1; j<=BOARD_HEIGHT; j++)
     {
        uint8_t x = 50+(i*8);
        uint8_t y = 54+(j*8);
        if(board[i-1][j-1] == 0)
        {
          swimex_fill_rect(&GameWindow, x, y, x+ICON_WIDTH, y+ICON_HEIGHT, BLACK);
        }
        else
        {
          uint8_t index = board[i-1][j-1] -1;
          swimex_put_image_at_position(&GameWindow,
              (COLOR_T*)icons[index].pIconData,
              x,
              y,
              ICON_WIDTH,
              ICON_HEIGHT);
        }
    }
}


/*****************************************************************************
 *
 * Description:
 *    Insert one figure at a specific x,y position on the game board
 *
 ****************************************************************************/
static void
insertFigure(int32_t shape, int32_t x, int32_t y)
{
  int32_t i,j;

  for (i=0; i<FIGURE_SIZE; i++)
  {
    for (j=0; j<FIGURE_SIZE; j++)
    {
      if(figures[shape][i][j] == 'X')
      {
        board[x+i][y+j] = (int32_t)figures[shape][FIGURE_SIZE][0];
      }
    }
  }
}


/*****************************************************************************
 *
 * Description:
 *    Delete one figure at a specific x,y position on the game board
 *
 ****************************************************************************/
static void
deleteFigure(int32_t shape, int32_t x, int32_t y)
{
  int32_t i,j;

  for (i=0; i<FIGURE_SIZE; i++)
    for (j=0; j<FIGURE_SIZE; j++)
      if(figures[shape][i][j] == 'X')
        board[x+i][y+j] = 0;
}


/*****************************************************************************
 *
 * Description:
 *    Returns TRUE if one row of the board is full
 *
 ****************************************************************************/
static char
full(int32_t j)
{
  int32_t i;

  for(i=0; i<BOARD_WIDTH; i++)
    if(board[i][j] == 0)
      return 0;

  return 1;
}


/*****************************************************************************
 *
 * Description:
 *    Creates a dialog and displays the message in it.
 *
 ****************************************************************************/
static void
displayMessageWin(BackBuffer_t lcdHnd, char* msg)
{
  if (!MessageWinActive)
  {
    int w = 180;
    int h = 60;
    int x = (GlobalParameters.display.width - w) / 2;
    int y = (GlobalParameters.display.height - h) / 2;
    swim_window_open(&MessageWin,
        GlobalParameters.display.width,
        GlobalParameters.display.height,
        (COLOR_T*)lcdHnd,
        x,y,x+w,y+h,1,BLACK,WHITE,WHITE);
    swimex_put_text_centered(&MessageWin, (h - MessageWin.font->font_height)/2, msg);
    MessageWinActive = TRUE;
  }
}

/*****************************************************************************
 *
 * Description:
 *    Destroys the dialog
 *
 ****************************************************************************/
static void
hideMessageWin(void)
{
  swim_window_close(&MessageWin);
  MessageWinActive = FALSE;
}

/*****************************************************************************
 *
 * Description:
 *    Draw the current score
 *
 ****************************************************************************/
static void
drawScore(void)
{
  char str[4];
  char strText[16] = " Score: ";

  str[0] = score/100 + '0';
  str[1] = (score/10)%10 + '0';
  str[2] = score%10 + '0';
  str[3] = 0;
  if (str[0] == '0')
  {
    str[0] = ' ';
    if (str[1] == '0')
      str[1] = ' ';
  }
  strcat(strText, str);
  swim_set_pen_color(&GameWindow, BLACK);
  swim_put_text_xy(&GameWindow, strText, 100, GameWindow.yvsize - 30);
}


/*****************************************************************************
 *
 * Description:
 *    Check if time to update score
 *
 ****************************************************************************/
static void
checkScore(void)
{
  int32_t i,j,k;

  for (j=0; j<BOARD_HEIGHT; j++)
  {
    //check if one row is full
    if(full(j))
    {
      //if so, erase that row...
      for(k=j; k>0; k--)
        for(i=0; i<BOARD_WIDTH; i++)
          board[i][k] = board[i][k-1];

      //and update score
      score++;
      drawScore();
    }
  }
}


/*****************************************************************************
 *
 * Description:
 *    Returns TRUE if game is over
 *
 ****************************************************************************/
static char
gameOver(void)
{
  int32_t i;

  for(i=0; i<NUM_OF_FIGURE; i++)
    if(!isValid(7,0,i))
      return 1;

  return 0;
}


/*****************************************************************************
 *
 * Description:
 *    Calculate new update/advance time, based in current score
 *    The higher score, the shorter update time
 *
 ****************************************************************************/
static void
updateDelayTime(void)
{
  if (score >= 75)
    delayTimeMs = 200;
  else if (score >= 50)
    delayTimeMs = 300;
  else if (score >= 25)
    delayTimeMs = 350;
  else if (score >= 10)
    delayTimeMs = 400;
  else
    delayTimeMs = 500;
}


/*****************************************************************************
 *
 * Description:
 *    Advance the game, check if game is over
 *
 ****************************************************************************/
static void
advanceGame(void)
{
  insertFigure(currFigure, currXpos, currYpos);
  drawGame();
  deleteFigure(currFigure, currXpos, currYpos);

  /*******************************************************
   * Check if time to update figure position
   *******************************************************/
  updateDelayTime();
  if ((msClock - lastUpdate) > delayTimeMs)
  {
    lastUpdate = msClock;

    if (isValid(currXpos, currYpos+1, currFigure))
      currYpos++;

    //figure has hit bottom
    //insert new, random figure on the game board
    else
    {
      insertFigure(currFigure, currXpos, currYpos);
      currYpos  = 0;
      currXpos  = 7;
      currFigure = random() % NUM_OF_FIGURE;

      //check if game is over
      if (gameOver())
        gameStatus = GAME_OVER;
    }

    //check score, i.e., if any rows are full
    checkScore();
  }
}


/*****************************************************************************
 *
 * Description:
 *    Draw game background and game board
 *
 ****************************************************************************/
static void
drawBoard(void)
{
  int j;

  swimex_put_image_at_position(&GameWindow, (COLOR_T*)(icons[6].pIconData), 0, 0, 240, 320);

  drawScore();

  swimex_fill_rect(&GameWindow, 58, 62, 177, 230, BLACK);

  //LEFT
  for (j=0; j < BOARD_HEIGHT; j++)
   {
      uint8_t x = 50;
      uint8_t y = 62+(j*8);
      swimex_put_image_at_position(&GameWindow, (COLOR_T*)icons[5].pIconData, x, y, 8, 8);
   }

   //RIGHT
   for (j=0; j < BOARD_HEIGHT; j++)
   {
      uint8_t x = 178;
      uint8_t y = 62+(j*8);
      swimex_put_image_at_position(&GameWindow, (COLOR_T*)icons[5].pIconData, x, y, 8, 8);
   }

   //BOTTOM
   for (j=0; j < BOARD_WIDTH + 2; j++)
   {
      uint8_t x = 50+(j*8);
      uint8_t y = 230;
      swimex_put_image_at_position(&GameWindow, (COLOR_T*)icons[5].pIconData, x, y, 8, 8);
   }

   //TOP
   for (j=0; j < BOARD_WIDTH + 2; j++)
   {
      uint8_t x = 50+(j*8);
      uint8_t y = 54;
      swimex_put_image_at_position(&GameWindow, (COLOR_T*)icons[5].pIconData, x, y, 8, 8);
   }
}


/*****************************************************************************
 *
 * Description:
 *    Draw game background and game board, initialize all variables
 *
 ****************************************************************************/
static void
setupGame (Bool redrawBoard)
{
  int32_t i,j;

  //initialize random generator and reset score
  random_init(msClock);
  score = 0;

  if (redrawBoard)
  {
    // board and background
    drawBoard();
  }
  else
  {
    // only background
    swimex_put_image_at_position(&GameWindow, (COLOR_T*)(icons[6].pIconData), 0, 0, 240, 320);
  }

  // Initialize game board
  for (i=0; i<BOARD_WIDTH; i++)
    for (j=0; j<BOARD_HEIGHT; j++)
      board[i][j] = 0;

  currFigure = random() % NUM_OF_FIGURE;
  currXpos  = 7;
  currYpos  = 0;
  insertFigure(currFigure, currXpos, currYpos);

  lastUpdate = msClock;
}


/*****************************************************************************
 *
 * Description:
 *    Run one round of the game
 *
 ****************************************************************************/
static void
playGame(BackBuffer_t lcdHnd)
{
  uint8_t joyState = 0;
  uint16_t buttonState = 0;

  GameWindow.fb = (COLOR_T*)lcdHnd;

  joyState = joystick_read();
  buttonState = pca9532_getLedState(FALSE);

  switch(gameStatus)
  {
    case GAME_NOT_STARTED:
      if (buttonState & KEY_BACK)
      {
        gameStatus = GAME_END;
        quitGame();
        return;
      }
      else if (ANYTHING_PRESSED(buttonState, joyState))
      {
        hideMessageWin();
        gameStatus = GAME_RUNNING;
        setupGame(TRUE);
      }
      break;

    case GAME_RUNNING:
      if (ANYTHING_PRESSED(buttonState, joyState))
      {
        if (joyState & JOYSTICK_UP)
        {
          if (!(lastJoyState & JOYSTICK_UP)) {
            if(isValid(currXpos, currYpos, shift[currFigure]))
              currFigure = shift[currFigure];
          }
        }
        else if (joyState & JOYSTICK_DOWN)
        {
          if(isValid(currXpos, currYpos+1, currFigure))
            currYpos++;
        }
        else if (joyState & JOYSTICK_RIGHT)
        {
          if(isValid(currXpos+1, currYpos, currFigure))
            currXpos++;
        }
        else if (joyState & JOYSTICK_LEFT)
        {
          if(isValid(currXpos-1, currYpos, currFigure))
            currXpos--;
        }
        else if (joyState & JOYSTICK_CENTER)
        {
          // avoid repeated presses on joystick
          if (!(lastJoyState & JOYSTICK_CENTER))
          {
            gameStatus = GAME_PAUSE;
            pause = !pause;
          }
        }
        else if (buttonState & KEY_BACK)
        {
          gameStatus = GAME_END;
          quitGame();
          return;
        }

        advanceGame();
      }
      else
      {
        advanceGame();
      }
      break;

    case GAME_PAUSE:
    {
      // avoid repeated presses on joystick
      if (joyState & JOYSTICK_CENTER)
      {
        if (!(lastJoyState & JOYSTICK_CENTER))
        {
          pause = !pause;
        }
      }

      else if (buttonState & KEY_BACK)
      {
        gameStatus = GAME_END;
        quitGame();
        return;
      }

      if(pause)
      {
        displayMessageWin(lcdHnd, "PAUSED");
      }
      else
      {
        hideMessageWin();
        drawBoard();
        drawGame();
        gameStatus = GAME_RUNNING;
      }
    }
    break;

    case GAME_OVER:
    {
      displayMessageWin(lcdHnd, "Game Over! Press key to restart");

      if (buttonState & KEY_BACK)
      {
        gameStatus = GAME_END;
        quitGame();
        return;
      }
      else if (ANYTHING_PRESSED(buttonState, joyState))
      {
        hideMessageWin();
        restartGame();
      }
    }
    break;

    default:
      gameStatus = GAME_END;
      break;
  }

  // Save last state of joystick to be able to filter out repeated presses
  lastJoyState = joyState;
}

static Bool
initGame(void)
{
  int w, h, x, y;
  MessageWinActive = FALSE;
  gameStatus = GAME_NOT_STARTED;
  if (!anydisplay_createBackBuffers(&gameLcdHnd1, &gameLcdHnd2, NULL))
  {
    log_e("Failed to allocate back buffers\r\n");
    return FALSE;
  }
  memset((void*)gameLcdHnd1, 0, GlobalParameters.display.buffersize);

  log_i("initGame, gameLcdHnd 0x%x\r\n", gameLcdHnd1);

  /* Create a SWIM window that is centered on the display and has a
   * maximum size of 240x320
   */
  w = MIN(GlobalParameters.display.width, 240);
  h = MIN(GlobalParameters.display.height, 320);
  x = (GlobalParameters.display.width - w) / 2;
  y = (GlobalParameters.display.height - h) / 2;
  swim_window_open(&GameWindow, GlobalParameters.display.width,
      GlobalParameters.display.height, (COLOR_T*)gameLcdHnd1, x, y,
      x+w-1,
      y+h-1, 0, BLACK, BLACK, BLACK);
  swim_set_font_trasparency(&GameWindow, 0);

  anydisplay_activateBuffer(gameLcdHnd1);
  setupGame(FALSE);

  displayMessageWin(gameLcdHnd1, "Press joystick to start");
  return TRUE;
}

void
restartGame()
{
  log_i("restart game\r\n");
  gameStatus = GAME_RUNNING;
  setupGame(TRUE);
}

void
quitGame(void)
{
  log_i("quit game\r\n");
}


static void
start(void)
{
  BackBuffer_t bbOld, bbNew, bbTmp;
  log_i("fallingBlockGameApp: start");

  uiLoadIcons();

  if (!initGame())
  {
    return;
  }

  log_i("fallingBlockGameApp: start game\r\n");

  play = 0;
  bbOld = gameLcdHnd2;
  bbNew = gameLcdHnd1;
  while(gameStatus != GAME_END) {
    if (play > 0)
    {
      bbTmp = bbOld;
      bbOld = bbNew;
      bbNew = bbTmp;

      memcpy((uint8_t*)bbNew, (uint8_t*)bbOld, GlobalParameters.display.buffersize);
      playGame(bbNew);
      anydisplay_activateBuffer(bbNew);
      play = 0;
    }
  }
}

static void
stop(void)
{
  uint8_t i;
  log_i("fallingBlockGameApp: stop\r\n");
  gameStatus = GAME_END;

  swim_window_close(&GameWindow);
  anydisplay_freeBackBuffers(&gameLcdHnd1, &gameLcdHnd2, NULL);

  for(i = 0; i < 7; ++i)
  {
     if(icons[i].pIconData)
     {
        lpc_free(icons[i].pIconData);
     }
  }
}

static void timeout(void)
{
  if (gameStatus != GAME_END)
  {
    play++;
  }
}


/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

static AppCallback_t fallingBlockGameCallback = {
    &start,
    &stop,
    &timeout,
    50,
    FALSE,
};

AppCallback_t* fallingBlockGameAppInit(void)
{
  return &fallingBlockGameCallback;
}
