/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "application.h"
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include "image.h"
#include "lpc_swim.h"
#include "lpc_swim_font.h"
#include "lpc_swim_image.h"
#include "globals.h"
#include "lpc_heap.h"
#include "log.h"
#include "pca9532.h"
#include "lpc177x_8x_timer.h"
#include "swim_extensions.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define BACKGROUND_COLOR          BLACK
#define LARGE_CIRCLE_COLOR        0x39e7 //DARK_GRAY
#define SMALL_CIRCLE_FRONT_COLOR  WHITE
#define SMALL_CIRCLE_BACK_COLOR   0x7bef //LIGHT_GRAY

// Screen width
#define WIDTH GlobalParameters.display.width
// Screen height
#define HEIGHT GlobalParameters.display.height

// Delay between frame updates (20 = 20 ms = 50Hz)
#define RENDER_DELAY_MS 30

#define NUM_OF_DOTS 50//40


/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static SWIM_WINDOW_T MainWin;
static Bool started = FALSE;
static Bool wantToShutDown = FALSE;

static BackBuffer_t myLcdHnd1 = BACK_BUFFER_INVALID;
static BackBuffer_t myLcdHnd2 = BACK_BUFFER_INVALID;
static BackBuffer_t myLcdHnd3 = BACK_BUFFER_INVALID;

static uint8_t   activeArea;

// Vertex structure
typedef struct
{
  long x, y, z;
}tVertex;

// Transformed vertices
tVertex *gRVtx;

static const tVertex gVtx[NUM_OF_DOTS] = {
{-21585, 17597, -17265},
{28493, -7790, 14183},
{13031, 27845, 11338},
{10822, 29162, -10304},
{19517, -25865, -4876},
{1283, 30949, -10687},
{-23097, 12297, 19723},
{-17162, 1188, 27888},
{-29007, -1547, 15163},
{-8182, -31729, -197},
{11599, -28575, -11073},
{25963, 12182, 15850},
{106, 10547, -31023},
{-24312, 3053, -21755},
{9966, -18803, -24916},
{17598, -6037, -26973},
{23321, 17149, -15353},
{-3265, 8867, -31376},
{-23639, 13701, 18087},
{-2433, -22123, -24049},
{21284, -11349, -22179},
{-21407, 24683, -2486},
{-32011, -4336, 5495},
{-20191, 22605, 12450},
{14752, -23540, 17376},
{-8961, -17292, -26351},
{-18078, 13561, -23727},
{10730, -23639, 19994},
{-18718, 17555, 20376},
{13626, -22837, 19144},
{-24695, -19036, -10073},
{11935, 22275, 20859},
{-2182, -28801, -15474},
{21428, -21867, -11678},
{-19601, 21558, -14991},
{24512, 10876, -18830},
{12385, 27881, 11956},
{26982, 15618, -10088},
{-16954, 19591, 20061},
{-6027, 22699, -22850},
{5453, 28825, -14598},
{-20155, -16252, -20083},
{-15962, 11757, -26089},
{-29175, -11289, -9750},
{-15370, 604, -28933},
{26009, 19868, -1575},
{24722, -17277, -12806},
{-4527, 25836, -19639},
{-22224, 10442, 21697},
{-10388, 24393, -19255}};

/******************************************************************************
 * Local function prototypes
 *****************************************************************************/

static void start(void);
static void stop(void);
static void timeout(void);


/******************************************************************************
 * Local functions
 *****************************************************************************/

static short _sin(short y){
	static short s1 = 0x6487;
	static short s3 = 0x2951;
	static short s5 = 0x4f6;
	long z, prod, sum;

	z = ((long)y * y) >> 12;
	prod = (z * s5) >> 16;
	sum = s3 - prod;
	prod = (z * sum) >> 16;
	sum = s1 - prod;

	// for better accuracy, round here
	return (short)((y * sum) >> 13);
 }

static short _cos(short y){
	static short c0 = 0x7fff;
	static short c2 = 0x4eea;
	static short c4 = 0x0fc4;
	long z, prod, sum;
	z = ((long)y * y) >> 12;
	prod = (z * c4) >> 16;
	sum = c2 - prod;

	// for better accuracy, round here
	prod = (z * sum) >> 15;
	return (short)(c0 - prod);
}

static short isine(short x){
	unsigned short n = (((unsigned short)x + 0x2000) >> 14) & 0x3;
	x -= n * 0x4000;
	switch(n){
		case 0:
			return _sin(x);
		case 1:
			return _cos(x);
		case 2:
			return - _sin(x);
		case 3:
			return  - _cos(x);
	}
  return -1; // this will never happen but eliminates compiler warning
}


static short icosine(short x){
	return isine(x + 0x4000);
}

static void
init()
{
  gRVtx = (tVertex*)lpc_new(sizeof(tVertex) * NUM_OF_DOTS);
}

static void rotate_z(uint32_t angle)
{
  uint32_t i;
  long x,y,ca,sa;

  ca = icosine(angle);
  sa = isine(angle);
  for (i = 0; i < NUM_OF_DOTS; i++)
  {
    x = (gRVtx[i].x * ca - gRVtx[i].y * sa) / 0x7fff;
    y = (gRVtx[i].x * sa + gRVtx[i].y * ca) / 0x7fff;
    gRVtx[i].x = x;
    gRVtx[i].y = y;
  }
}

static void rotate_y(uint32_t angle)
{
  uint32_t i;
  long x,z,ca,sa;

  ca = icosine(angle);
  sa = isine(angle);
  for (i = 0; i < NUM_OF_DOTS; i++)
  {
    z = (gRVtx[i].z * ca - gRVtx[i].x * sa) / 0x7fff;
    x = (gRVtx[i].z * sa + gRVtx[i].x * ca) / 0x7fff;
    gRVtx[i].z = z;
    gRVtx[i].x = x;
  }
}

static void
render(int32_t myLcdHnd, uint32_t tick)
{
  uint32_t i;
  uint32_t rotz;
  uint32_t roty;

  // rendering here
  memset((char*)myLcdHnd, BACKGROUND_COLOR, WIDTH*HEIGHT*2);
  MainWin.fb = (COLOR_T*)myLcdHnd;


  swim_set_pen_color(&MainWin, WHITE);
  swim_set_bkg_color(&MainWin, BACKGROUND_COLOR);
  swimex_put_text_centered(&MainWin, GlobalParameters.display.height - 20, "End application with button click");

  swimex_fill_circle(&MainWin, WIDTH / 2, HEIGHT / 2, HEIGHT / 4, LARGE_CIRCLE_COLOR);

  memcpy(gRVtx, gVtx, sizeof(tVertex) * NUM_OF_DOTS);

  rotz = tick * 5;
  roty = tick * 20;
  rotate_y(roty);
  rotate_z(rotz);

  for (i = 0; i < NUM_OF_DOTS; i++)
  {
    uint16_t c = SMALL_CIRCLE_FRONT_COLOR;

    if (gRVtx[i].z < 0)
      c = SMALL_CIRCLE_BACK_COLOR;
    swimex_fill_circle(&MainWin,
                   (int)((gRVtx[i].x * (HEIGHT / 4)) / 0x7fff + WIDTH / 2),
                   (int)((gRVtx[i].y * (HEIGHT / 4)) / 0x7fff + HEIGHT / 2),
                   2,
                   c);
  }
  swimex_fill_circle(&MainWin,
                 (int)(((HEIGHT / 3) * icosine(rotz))/0x7fff + WIDTH / 2),
                 (int)(((HEIGHT / 3) * isine(rotz))  /0x7fff + HEIGHT / 2),
                 7,
                 SMALL_CIRCLE_FRONT_COLOR);
}

static void
start(void)
{
  init();

  if (!anydisplay_createBackBuffers(&myLcdHnd1, &myLcdHnd2, &myLcdHnd3))
  {
    log_e("Failed to allocate back buffers\r\n");
    return;
  }

  /* Create a SWIM window */
  swim_window_open(&MainWin, GlobalParameters.display.width,
      GlobalParameters.display.height, (COLOR_T*)myLcdHnd1, 0, 0,
      (GlobalParameters.display.width - 1),
      (GlobalParameters.display.height - 1), 0, WHITE, BLACK, BLACK);

  render(myLcdHnd1, msClock);
  anydisplay_activateBuffer(myLcdHnd1);
  activeArea = 1;
  render(myLcdHnd2, msClock);

  wantToShutDown = FALSE;
  started = TRUE;

  while(started)
  {
    TIM_Waitms(100);

    /* Check SW2 button which causes Exit */
    if (pca9532_getLedState(FALSE) & KEY1) {
      wantToShutDown = TRUE;
    }
  }
}

static void
stop(void)
{
  started = FALSE;
  swim_window_close(&MainWin);
  lpc_free(gRVtx);

  anydisplay_freeBackBuffers(&myLcdHnd1, &myLcdHnd2, &myLcdHnd3);
}

static void timeout(void)
{
  if (started)
  {
    if (wantToShutDown) {
      started = FALSE;
    }
    else if (activeArea == 3)
    {
      anydisplay_activateBuffer(myLcdHnd1);
      render(myLcdHnd2, msClock);
      activeArea = 1;
    }
    else if (activeArea == 1)
    {
      anydisplay_activateBuffer(myLcdHnd2);
      render(myLcdHnd3, msClock);
      activeArea = 2;
    }
    else
    {
      anydisplay_activateBuffer(myLcdHnd3);
      render(myLcdHnd1, msClock);
      activeArea = 3;
    }
  }
}

/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

static AppCallback_t globeCallback = {
    &start,
    &stop,
    &timeout,
    RENDER_DELAY_MS,
    FALSE,
};

AppCallback_t* globeAppInit(void)
{
  return &globeCallback;
}
