/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include "jaw_game.h"
#include "jaw_types.h"
#include "jaw_graphics.h"
#include "log.h"
#include "lpc_heap.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define JAW_IS_VALID_CELL(r,c) (((r) >= 0) && ((r) < JAW_NUM_ROWS) && ((c) >= 0) && ((c) < JAW_NUM_COLUMNS))
#define JAW_IS_EMPTY_CELL(r,c) (gpBoard->board[r][c].state == CELL_STATE_EMPTY)
#define JAW_CELL_TYPE(r,c)      gpBoard->board[r][c].type


/******************************************************************************
 * Public variables
 *****************************************************************************/

tBoard* gpBoard = NULL;

/******************************************************************************
 * External variables
 *****************************************************************************/

extern volatile uint32_t msClock;

/******************************************************************************
 * Local variables
 *****************************************************************************/

/******************************************************************************
 * Local function prototypes
 *****************************************************************************/

static tJawError SelectIfMatching(tCell* pCell, int row, int col);
static tJawError Select(tCell* pCell);
static tJawError Remove(tArrayList* pList);
static void Syncronize(void);
static void MoveCol(int col);
static uint32_t GetScore(int num);
static void Initialize(void);
static Bool GameFinished(void);
static tJawError ClearSelection(void);
static tJawError UpdateSelection(void);
static void ShowGameOverScreen(void);
static void UpdateScore(void);
static void SetExpectedScore(uint32_t score);


/******************************************************************************
 * Local functions
 *****************************************************************************/

static tJawError
SelectIfMatching(tCell* pCell, int row, int col)
{
  if ((row >= 0) && (row < JAW_NUM_ROWS) && (col >= 0) && (col < JAW_NUM_COLUMNS))
  {
    tCell* pNew = &(gpBoard->board[row][col]);
    if (pNew->state != CELL_STATE_EMPTY)
    {
      if (pNew->type == pCell->type)
      {
        if (!jaw_ArrayList_Contains(&gpBoard->selected, pNew))
        {
          return Select(pNew);
        }
      }
    }
  }
  return E_OK;
}

static tJawError
Select(tCell* pCell)
{
  tJawError result;
  int row = pCell->row;
  int col = pCell->col;

  do
  {
    result = jaw_ArrayList_Add(&gpBoard->selected, pCell);
    if (result != E_OK)
      break;

    result = SelectIfMatching(pCell, row-1, col);
    if (result != E_OK)
      break;

    result = SelectIfMatching(pCell, row,   col-1);
    if (result != E_OK)
      break;

    result = SelectIfMatching(pCell, row,   col+1);
    if (result != E_OK)
      break;

    result = SelectIfMatching(pCell, row+1, col);
    if (result != E_OK)
      break;

  } while (0);

  return result;
}

static tJawError
Remove(tArrayList* pList)
{
  if (pList == NULL)
  {
    return E_INVALID_POINTER;
  }
  else if (pList->numUsed > 0)
  {
    int i;
    for (i = 0; i < pList->numUsed; i++)
    {
      pList->list[i]->state = CELL_STATE_EMPTY;
    }

    //Sync the board
    Syncronize();

    //update the score
    gpBoard->totalScore += GetScore(pList->numUsed);

    // clear the list
    pList->numUsed = 0;
  }

  return E_OK;
}

static void
Syncronize(void)
{
  int times, i, j;
  tCell* pMyCell;
  tCell* pCellBelow;

  //Use bruteforce to sync for the time being
  //Use a cool algorithm here to speed up things
  for (times=0; times < JAW_NUM_ROWS; times++)
  {
    for (i = 0; i < (JAW_NUM_ROWS - 1); i++)
    {
      for (j = 0; j < JAW_NUM_COLUMNS; j++)
      {
        pMyCell = &(gpBoard->board[i][j]);

        if (pMyCell->state != CELL_STATE_EMPTY)
        {
          pCellBelow = &(gpBoard->board[pMyCell->row+1][pMyCell->col]);
          if (pCellBelow->state == CELL_STATE_EMPTY)
          {
            // move me down
            pCellBelow->state = pMyCell->state;
            pCellBelow->type = pMyCell->type;
            pMyCell->state = CELL_STATE_EMPTY;
          }
        }
      }
    }
  }

  //Check if any empty columns-- and move accordingly
  //Again use simple algorithm
  for (times = 0; times < JAW_NUM_COLUMNS; times++)
  {
    for (i = 1; i < JAW_NUM_COLUMNS; i++)
    {
      //if (JAW_IS_EMPTY_COLUMN(i))
      if (gpBoard->board[JAW_NUM_ROWS-1][i].state == CELL_STATE_EMPTY)
      {
        MoveCol(i);
      }
    }
  }

  // update all cells
  jaw_UpdateAllCells();
}

static void
MoveCol(int col)
{
  int i;
  tCell* pMyCell;
  tCell* pLeftCell;
  for (i = 0; i < JAW_NUM_ROWS; i++)
  {
    pMyCell = &(gpBoard->board[i][col]);
    pLeftCell = &(gpBoard->board[i][col-1]);
    if (pLeftCell->state != CELL_STATE_EMPTY)
    {
      pMyCell->state = pLeftCell->state;
      pMyCell->type = pLeftCell->type;
      pLeftCell->state = CELL_STATE_EMPTY;
    }
  }
}



static uint32_t
GetScore(int num)
{
  return num*(num-1);
}

static void
Initialize(void)
{
  int i;
  int j;
  tCell* pCell;

  srand(msClock);

  for (i = 0; i < JAW_NUM_ROWS; i++)
  {
    for (j = 0; j < JAW_NUM_COLUMNS; j++)
    {
      pCell =  &(gpBoard->board[i][j]);
      pCell->row = i;
      pCell->col = j;
      pCell->state = CELL_STATE_UNSELECTED;
      pCell->type = rand() % JAW_NUM_TYPES;
    }
  }

  gpBoard->totalScore=0;
  gpBoard->selected.numUsed = 0;
}

static Bool
GameFinished(void)
{
  int i;
  int j;

  for (j = 0; j < JAW_NUM_COLUMNS; j++)
  {
    for (i = JAW_NUM_ROWS-1; i >= 1; i--)
    {
      if (JAW_IS_EMPTY_CELL(i,j))
      {
        // cant be anything on rows above an empty cell
        break;
      }
      else
      {
        // check if cell above is valid
        if (JAW_IS_VALID_CELL(i-1,j) && !JAW_IS_EMPTY_CELL(i-1,j))
        {
          // check if cell above is of same type
          if (JAW_CELL_TYPE(i,j) == JAW_CELL_TYPE(i-1,j))
          {
            // found same color above
            return FALSE;
          }
        }

        // check if cell to the right is valid
        if (JAW_IS_VALID_CELL(i,j+1) && !JAW_IS_EMPTY_CELL(i,j+1))
        {
          // check if cell to the right is of same type
          if (JAW_CELL_TYPE(i,j) == JAW_CELL_TYPE(i,j+1))
          {
            // found same color to the right
            return FALSE;
          }
        }
      }
    }
  }

  return TRUE;
}

static tJawError
ClearSelection(void)
{
  tJawError result = E_OK;
  int num = gpBoard->selected.numUsed;
  if (num > 0)
  {
    int i;
    for (i = 0; (i < num) && (result == E_OK); i++)
    {
      (gpBoard->selected).list[i]->state = CELL_STATE_UNSELECTED;
      result = jaw_UpdateCell((gpBoard->selected).list[i]);
    }
  }
  gpBoard->selected.numUsed = 0;
  return result;
}

static tJawError
UpdateSelection(void)
{
  tJawError result = E_OK;

  int num = gpBoard->selected.numUsed;
  if (num > 0)
  {
    int i;
    for (i = 0; (i < num) && (result == E_OK); i++)
    {
      ((gpBoard->selected).list[i])->state = CELL_STATE_SELECTED;
      result = jaw_UpdateCell((gpBoard->selected).list[i]);
    }
  }
  return result;
}

static void
ShowGameOverScreen(void)
{
  log_i("\r\nGame over your score was %u.\r\n", gpBoard->totalScore);
  jaw_ShowGameOverScreen(gpBoard->totalScore);
}

static void
UpdateScore(void)
{
  gpBoard->totalScore += GetScore(gpBoard->selected.numUsed);

  log_i("Your score is now %u\r\n", gpBoard->totalScore);
  (void)jaw_UpdateTotalScore(gpBoard->totalScore);
}

static void
SetExpectedScore(uint32_t score)
{
  //log_i("Expected score is now %u\r\n", score);
  (void)jaw_UpdateExpectedScore(score);
}

/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

tJawError
jaw_ArrayList_Create(tArrayList** ppArrayList)
{
  if (ppArrayList == NULL)
  {
    return E_INVALID_POINTER;
  }
  if (*ppArrayList != NULL)
  {
    tJawError result = jaw_ArrayList_Destroy(ppArrayList);
    if (result != E_OK)
    {
      return result;
    }
  }
  *ppArrayList = (tArrayList*)lpc_new(sizeof(tArrayList));
  if (*ppArrayList == NULL)
    return E_FAILED_TO_ALLOCATE;

  (*ppArrayList)->numUsed = 0;

  return E_OK;
}

tJawError jaw_ArrayList_Destroy(tArrayList** ppArrayList)
{
  if (ppArrayList == NULL)
  {
    return E_INVALID_POINTER;
  }
  if (*ppArrayList != NULL)
  {
    lpc_free(*ppArrayList);
    *ppArrayList = NULL;
  }
  return E_OK;
}

tJawError jaw_ArrayList_Add(tArrayList* pArrayList, tCell* pCell)
{
  if (pArrayList == NULL)
  {
    return E_INVALID_POINTER;
  }
  pArrayList->list[pArrayList->numUsed++] = pCell;
  return E_OK;
}

tJawError jaw_ArrayList_AddIfUnique(tArrayList* pArrayList, tCell* pCell)
{
  if (pArrayList == NULL)
  {
    return E_INVALID_POINTER;
  }
  else
  {
    int i;
    for (i = 0; i < pArrayList->numUsed; i++)
    {
      if (pArrayList->list[i] == pCell)
      {
        // already in the list
        return E_OK;
      }
    }
    pArrayList->list[pArrayList->numUsed++] = pCell;
  }
  return E_OK;
}

Bool jaw_ArrayList_Contains(tArrayList* pArrayList, tCell* pCell)
{
  if (pArrayList == NULL)
  {
    return E_INVALID_POINTER;
  }
  else
  {
    int i;
    for (i = 0; i < pArrayList->numUsed; i++)
    {
      if (pArrayList->list[i] == pCell)
      {
        return TRUE;
      }
    }
  }
  return FALSE;
}

tJawError
jaw_StartGame(void)
{
  tJawError result;

  log_i("jaw_StartGame: Setting up game data %u\r\n", sizeof(tBoard));
  do
  {
    gpBoard = (tBoard*)lpc_new(sizeof(tBoard));
    if (gpBoard == NULL)
    {
      log_i("jaw_StartGame: failed to allocate board\r\n");
    }
    gpBoard->selected.numUsed = 0;
    gpBoard->totalScore = 0;

    // prepare cells
    do
    {
      Initialize();
    } while (GameFinished());

    // draw board and cells
    result = jaw_InitGraphics();
    if (result != E_OK)
      break;

  } while (0);

  if (result != E_OK)
  {
    jaw_EndGame();
  }

  return result;
}

// user has pressed the red button and the game is terminated immediately
tJawError
jaw_EndGame(void)
{
  if (gpBoard != NULL)
  {
    lpc_free(gpBoard);
    gpBoard = NULL;
  }

  return jaw_DeinitGraphics();
}

// Deals with the user click and returns E_OK if the game should continue,
// E_USER_SELECTED_TO_EXIT if the game should be shut down or any of the other
// error codes if anything goes wrong
tJawError
jaw_ClickAt(int x, int y, int p)
{
  tJawError result = E_OK;
  tCell* pCell = jaw_GetCellForXY(x,y);
  if (pCell != NULL)
  {
    if (pCell->state == CELL_STATE_EMPTY)
    {
      // Click on empty cell => clear selection
      result = ClearSelection();
      SetExpectedScore(0);
    }
    else if (pCell->state == CELL_STATE_SELECTED)
    {
      if (jaw_ArrayList_Contains(&(gpBoard->selected), pCell))
      {
        SetExpectedScore(0);
        UpdateScore();
        result = Remove(&(gpBoard->selected));
      }
      else
      {
        log_i("STRANGE: Cell [%u,%u] is both selected and unselected\r\n", pCell->row, pCell->col);
      }
    }
    else
    {
      result = ClearSelection();
      result = Select(pCell);
      if (gpBoard->selected.numUsed == 1)
      {
        // not allowed to select only one cell
        result = ClearSelection();
      }
      else
      {
        result = UpdateSelection();
      }

      SetExpectedScore(GetScore(gpBoard->selected.numUsed));
    }
  }
  jaw_UpdateAllCells();

  if (GameFinished())
  {
    ShowGameOverScreen();
  }

  return result;
}

tJawError
jaw_CursorAt(uint16_t x, uint16_t y, uint16_t p, Bool released)
{
  Bool ClickedOnNew = FALSE;
  Bool ClickedOnQuit = FALSE;
  if (jaw_IsButtonClick(&ClickedOnNew, &ClickedOnQuit, x, y, p, released))
  {
    if (ClickedOnQuit)
    {
      return E_USER_SELECTED_TO_EXIT;
    }
    else if (ClickedOnNew)
    {
      log_i("User wants to start a new game\r\n");
      Initialize();
      return jaw_RestartGraphics();
    }
  }
  else if (released)
  {
    return jaw_ClickAt(x, y, p);
  }
  return E_OK;
}

