/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include "jaw_game.h"
#include "jaw_graphics.h"
#include "image.h"
#include "lpc_swim.h"
#include "lpc_swim_font.h"
#include "lpc_swim_image.h"
#include "globals.h"
#include "lpc_heap.h"
#include "log.h"
#include "joystick.h"
#include "lpc177x_8x_timer.h"
#include "swim_extensions.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

typedef uint16_t* tImage;

#define JAW_OFFSET_UNSELECTED   0
#define JAW_OFFSET_SELECTED     JAW_OFFSET_UNSELECTED + JAW_NUM_TYPES
#define JAW_OFFSET_BUTTON_UP    JAW_OFFSET_SELECTED + JAW_NUM_TYPES
#define JAW_OFFSET_BUTTON_DOWN  JAW_OFFSET_BUTTON_UP + 1
#define JAW_OFFSET_BACKIMAGE    JAW_OFFSET_BUTTON_DOWN + 1

#define JAW_NUM_PREPARED_IMG  JAW_OFFSET_BACKIMAGE + 1   //Total number of preloaded images

#define JAW_WINDOW_WIDTH   GlobalParameters.display.width
#define JAW_WINDOW_HEIGHT  GlobalParameters.display.height
#define JAW_WINDOW_BPP      16

#define JAW_VERTICAL_SPACING    1
#define JAW_HORIZONTAL_SPACING  1

#define JAW_CELL_WIDTH   20
#define JAW_CELL_HEIGHT  20

#define JAW_TOUCH_MARGIN_X  1
#define JAW_TOUCH_MARGIN_Y  1

#define JAW_BOARD_WIDTH   (JAW_NUM_COLUMNS * JAW_CELL_WIDTH +  (JAW_NUM_COLUMNS+1) * JAW_HORIZONTAL_SPACING)
#define JAW_BOARD_HEIGHT  (JAW_NUM_ROWS    * JAW_CELL_HEIGHT + (JAW_NUM_ROWS+1)    * JAW_VERTICAL_SPACING)
#define JAW_BOARD_START_X  jawBoardStartX
#define JAW_BOARD_START_Y  jawBoardStartY


#define JAW_BUTTON_HEIGHT  20
#define JAW_BUTTON_WIDTH   92

#define JAW_BTN_NEW_X   jawButtonNewX
#define JAW_BTN_NEW_Y   jawButtonNewY
#define JAW_BTN_QUIT_X  jawButtonQuitX
#define JAW_BTN_QUIT_Y  jawButtonQuitY

#define JAW_BTN_DOWN_TEXT_BKG_COLOR  0xE71A
#define JAW_BTN_UP_TEXT_BKG_COLOR    0xF7BE

#define JAW_CELL_X(col)  (JAW_BOARD_START_X + (JAW_CELL_WIDTH + JAW_HORIZONTAL_SPACING) * (col))
#define JAW_CELL_Y(row)  (JAW_BOARD_START_Y + (JAW_CELL_HEIGHT+ JAW_VERTICAL_SPACING) * (row))

#define JAW_GAME_BACKGROUND_COLOR  0xA165 // EA-red
#define JAW_GAME_FOREGROUND_COLOR  WHITE
#define JAW_GAME_FRAME_COLOR       0xBD31
#define JAW_PLAY_BACKGROUND_COLOR  WHITE

#define JAW_GAME_OVER_SCREEN_WIDTH   120
#define JAW_GAME_OVER_SCREEN_HEIGHT   80
#define JAW_GAME_OVER_SCREEN_COLOR   WHITE


/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static SWIM_WINDOW_T GameWindow;

static BackBuffer_t myLcdHnd = BACK_BUFFER_INVALID;
static tImage PreparedImages[JAW_NUM_PREPARED_IMG];
static tImage EmptyCellImage = NULL;

static Bool ButtonNewPressed = FALSE;
static Bool ButtonQuitPressed = FALSE;

static int jawButtonNewX = 0;
static int jawButtonNewY = 0;
static int jawButtonQuitX = 0;
static int jawButtonQuitY = 0;
static int jawBoardStartX = 0;
static int jawBoardStartY = 0;
static int jawLabelExpectedX = 0;
static int jawLabelExpectedY = 0;
static int jawLabelTotalX = 0;
static int jawLabelTotalY = 0;

/******************************************************************************
 * Local function prototypes
 *****************************************************************************/

static tJawError loadImage(char* pFileName, ImageFormat_t FileFormat, int slot, int width, int height);
static tJawError PrepareBackground(void);
static void DrawButtonQuit(Bool pressed, Bool forceDraw);
static void DrawButtonNew(Bool pressed, Bool forceDraw);
static void DrawCenterString(int x, int y, int w, int h, char* pStr);
static uint8_t int2String(uint32_t value, char* pBuf);


/******************************************************************************
 * Local functions
 *****************************************************************************/

static tJawError
loadImage(char* pFileName, ImageFormat_t FileFormat, int slot, int width, int height)
{
  tJawError jawresult = E_OK;
  ImageError_t result;
  uint16_t w;
  uint16_t h;
  uint8_t* pTmp;

  if (PreparedImages[slot] != NULL)
  {
    lpc_free(PreparedImages[slot]);
    PreparedImages[slot] = NULL;
  }

  pTmp = NULL;
  result = decodeImageFile(ONLY_USE_ROMFS, pFileName, &pTmp, &w, &h);
  if (result == IMAGE_ERROR_OK)
  {
    PreparedImages[slot] = (uint16_t*)pTmp;
  }
  else
  {
    log_i("[%s:%d] Failed to load icon %s, result=%d\r\n", __FILE__, __LINE__, pFileName, result);
    jawresult = E_INVALID_IMAGE;
  }

  return jawresult;
}

static tJawError
PrepareBackground(void)
{
  int i, j;
  tCell* pCell;

  if (PreparedImages[JAW_OFFSET_BACKIMAGE] == NULL)
  {
    swim_clear_screen(&GameWindow, JAW_GAME_BACKGROUND_COLOR);
  }
  else
  {
    swimex_put_image_at_position(&GameWindow,
        (COLOR_T*)PreparedImages[JAW_OFFSET_BACKIMAGE],
        0,
        0,
        JAW_WINDOW_WIDTH,
        JAW_WINDOW_HEIGHT);
  }


  // Draw frame
  swimex_fill_rect(&GameWindow,
               JAW_BOARD_START_X - 2,
               JAW_BOARD_START_Y - 2,
               JAW_BOARD_START_X + JAW_BOARD_WIDTH + 2,
               JAW_BOARD_START_Y + JAW_BOARD_HEIGHT + 2,
               JAW_GAME_FRAME_COLOR);


  // Drawing the board
  swimex_fill_rect(&GameWindow,
               JAW_BOARD_START_X,
               JAW_BOARD_START_Y,
               JAW_BOARD_START_X + JAW_BOARD_WIDTH,
               JAW_BOARD_START_Y + JAW_BOARD_HEIGHT,
               JAW_PLAY_BACKGROUND_COLOR);

  // Drawing the cells
  for (i = 0; i < JAW_NUM_ROWS; i++)
  {
    for (j = 0; j < JAW_NUM_COLUMNS; j++)
    {
      pCell = JAW_POINTER_TO_CELL(i,j);
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)PreparedImages[JAW_OFFSET_UNSELECTED + pCell->type],
          JAW_CELL_X(pCell->col),
          JAW_CELL_Y(pCell->row),
          JAW_CELL_WIDTH,
          JAW_CELL_HEIGHT);
      pCell->lastRenderedState = CELL_STATE_UNSELECTED;
    }
  }

  // Draw the buttons
  DrawButtonNew(FALSE, TRUE);
  DrawButtonQuit(FALSE, TRUE);

  return E_OK;
}

static void
DrawButtonNew(Bool pressed, Bool forceDraw)
{
  if (forceDraw || (pressed != ButtonNewPressed))
  {
    if (pressed)
    {
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)PreparedImages[JAW_OFFSET_BUTTON_DOWN],
          JAW_BTN_NEW_X,
          JAW_BTN_NEW_Y,
          JAW_BUTTON_WIDTH,
          JAW_BUTTON_HEIGHT);
      swim_set_pen_color(&GameWindow, BLACK);
      swim_set_bkg_color(&GameWindow, JAW_BTN_DOWN_TEXT_BKG_COLOR);
    }
    else
    {
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)PreparedImages[JAW_OFFSET_BUTTON_UP],
          JAW_BTN_NEW_X,
          JAW_BTN_NEW_Y,
          JAW_BUTTON_WIDTH,
          JAW_BUTTON_HEIGHT);
      swim_set_pen_color(&GameWindow, BLACK);
      swim_set_bkg_color(&GameWindow, JAW_BTN_UP_TEXT_BKG_COLOR);
    }
    DrawCenterString(JAW_BTN_NEW_X, JAW_BTN_NEW_Y, JAW_BUTTON_WIDTH, JAW_BUTTON_HEIGHT, "New Game");
  }

  ButtonNewPressed = pressed;
}

static void
DrawButtonQuit(Bool pressed, Bool forceDraw)
{
  if (forceDraw || (pressed != ButtonQuitPressed))
  {
    if (pressed)
    {
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)PreparedImages[JAW_OFFSET_BUTTON_DOWN],
          JAW_BTN_QUIT_X,
          JAW_BTN_QUIT_Y,
          JAW_BUTTON_WIDTH,
          JAW_BUTTON_HEIGHT);
      swim_set_pen_color(&GameWindow, BLACK);
      swim_set_bkg_color(&GameWindow, JAW_BTN_DOWN_TEXT_BKG_COLOR);
    }
    else
    {
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)PreparedImages[JAW_OFFSET_BUTTON_UP],
          JAW_BTN_QUIT_X,
          JAW_BTN_QUIT_Y,
          JAW_BUTTON_WIDTH,
          JAW_BUTTON_HEIGHT);
      swim_set_pen_color(&GameWindow, BLACK);
      swim_set_bkg_color(&GameWindow, JAW_BTN_UP_TEXT_BKG_COLOR);
    }
    DrawCenterString(JAW_BTN_QUIT_X, JAW_BTN_QUIT_Y, JAW_BUTTON_WIDTH, JAW_BUTTON_HEIGHT, "Quit");
  }

  ButtonQuitPressed = pressed;
}

static void
DrawCenterString(int x, int y, int w, int h, char* pStr)
{
  int offsetX = 0;
  int offsetY = 0;
  int strWidth = 0;
  int strHeight;
  int i;
  for (i = 0; pStr[i] != '\0'; i++)
  {
    strWidth += (GameWindow.font->font_width_table[pStr[i] - GameWindow.font->first_char]);
  }
  strHeight = GameWindow.font->font_height;

  if (strWidth < w)
  {
    offsetX = (w - strWidth) / 2;
  }
  if (h > strHeight)
  {
    offsetY = (h - strHeight) / 2;
  }
  swim_put_text_xy(&GameWindow, pStr, x + offsetX, y + offsetY);
}

static uint8_t
int2String(uint32_t value, char* pBuf)
{
  int8_t  pos     = 0;
  uint8_t  *pAlpha = (uint8_t*)"0123456789ABCDEF";
  uint8_t base = 10;
  uint32_t num     = value;
  uint8_t  len     = 0;

  while(num > 0)
  {
    len++;
    num /= base;
  }

  if(value == 0)
    len = 1;

  pos = len-1;
  num = value;

  for(;pos >= 0; pos--)
  {
    pBuf[pos] = pAlpha[num % base];
    num      /= base;
  }

  pBuf[len] = '\0';
  return len;
}

/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

tJawError
jaw_InitGraphics(void)
{
  int i;
  tJawError result = E_OK;
  char fileName[18] = {'/','g','a','m','e','s','/','j','a','w','_','x','x','.','p','n','g','\0'};

  if (GlobalParameters.display.height < 320)
  {
    // The 4.3" display has a 480x272 resolution and the board will take
    // up the entire height. Move the buttons and status text to the left

    jawBoardStartX = (JAW_WINDOW_WIDTH - JAW_BOARD_WIDTH)/2;
    jawBoardStartY = (JAW_WINDOW_HEIGHT - JAW_BOARD_HEIGHT)/2;
    jawButtonNewX = (JAW_BOARD_START_X - JAW_BUTTON_WIDTH)/2;
    jawButtonNewY = (GlobalParameters.display.height - 4*(5 + JAW_BUTTON_HEIGHT))/2;
    jawButtonQuitX = jawButtonNewX;
    jawButtonQuitY = jawButtonNewY + 5 + JAW_BUTTON_HEIGHT;
    jawLabelExpectedX = jawButtonNewX;
    jawLabelExpectedY = jawButtonQuitY + 5 + JAW_BUTTON_HEIGHT;
    jawLabelTotalX = jawButtonNewX;
    jawLabelTotalY = jawLabelExpectedY + 5 + JAW_BUTTON_HEIGHT;
  }
  else
  {
    jawBoardStartX = (JAW_WINDOW_WIDTH - JAW_BOARD_WIDTH)/2;
    jawBoardStartY = (JAW_WINDOW_HEIGHT - JAW_BOARD_HEIGHT - 60)/2;
    jawButtonNewX = JAW_BOARD_START_X;
    jawButtonNewY = JAW_BOARD_START_Y + JAW_BOARD_HEIGHT + 10;
    jawButtonQuitX = JAW_BOARD_START_X;
    jawButtonQuitY = JAW_BTN_NEW_Y + 5 + JAW_BUTTON_HEIGHT + 5;
    jawLabelExpectedX = JAW_WINDOW_WIDTH/2;
    jawLabelExpectedY = JAW_BTN_NEW_Y+(JAW_BUTTON_HEIGHT-7)/2;
    jawLabelTotalX = JAW_WINDOW_WIDTH/2;
    jawLabelTotalY = JAW_BTN_QUIT_Y+(JAW_BUTTON_HEIGHT-7)/2;
  }

  do
  {
    log_i("jaw_InitGraphics: Loading cell images\r\n");

    // Load all unselected images
    for (i = 0; i < JAW_NUM_TYPES && result==E_OK; i++)
    {
      PreparedImages[i] = NULL;
      fileName[11] = 'u';
      fileName[12] = '0' + i;
      result = loadImage(fileName, IMAGE_FORMAT_PNG, i, JAW_CELL_WIDTH, JAW_CELL_HEIGHT);
    }
    if (result != E_OK)
      break;

    // Load all selected images
    for (i = 0; i < JAW_NUM_TYPES && result==E_OK; i++)
    {
      PreparedImages[JAW_OFFSET_SELECTED+i] = NULL;
      fileName[11] = 's';
      fileName[12] = '0' + i;
      result = loadImage(fileName, IMAGE_FORMAT_PNG, JAW_OFFSET_SELECTED+i, JAW_CELL_WIDTH, JAW_CELL_HEIGHT);
    }
    if (result != E_OK)
      break;

    log_i("jaw_InitGraphics: Loading button images\r\n");

    // load button up image
    PreparedImages[JAW_OFFSET_BUTTON_UP] = NULL;
    result = loadImage("/games/jaw_b_u.png", IMAGE_FORMAT_PNG, JAW_OFFSET_BUTTON_UP, JAW_BUTTON_WIDTH, JAW_BUTTON_HEIGHT);
    if (result != E_OK)
      break;

    // load button down image
    PreparedImages[JAW_OFFSET_BUTTON_DOWN] = NULL;
    result = loadImage("/games/jaw_b_d.png", IMAGE_FORMAT_PNG, JAW_OFFSET_BUTTON_DOWN, JAW_BUTTON_WIDTH, JAW_BUTTON_HEIGHT);
    if (result != E_OK)
      break;

    log_i("jaw_InitGraphics: Loading background image\r\n");

    // load background image (we don't care if it failes - if it does we use background color)
    PreparedImages[JAW_OFFSET_BACKIMAGE] = NULL;
    loadImage("/games/jaw_back.png", IMAGE_FORMAT_PNG, JAW_OFFSET_BACKIMAGE, JAW_WINDOW_WIDTH, JAW_WINDOW_HEIGHT);
    if (PreparedImages[JAW_OFFSET_BACKIMAGE] == NULL)
    {
      log_i("Skip backgound image and use %x as color instead\r\n", JAW_GAME_BACKGROUND_COLOR);
    }

    log_i("jaw_InitGraphics: Creating empty cell image\r\n");

    // Create the image for empty cells
    EmptyCellImage = (tImage)lpc_new(JAW_CELL_WIDTH*JAW_CELL_HEIGHT*JAW_WINDOW_BPP/2);
    if (EmptyCellImage == NULL)
    {
      result = E_FAILED_TO_ALLOCATE;
      break;
    }
    memset(EmptyCellImage, WHITE, JAW_CELL_WIDTH*JAW_CELL_HEIGHT*JAW_WINDOW_BPP/8);

    log_i("jaw_InitGraphics: Setting up LCD\r\n");

    // create graphics buffers
    myLcdHnd = anydisplay_createBackBuffer();
    if (myLcdHnd == BACK_BUFFER_INVALID)
    {
      result = E_LCD_CREATE_FAILED;
      break;
    }

    /* Create a SWIM window */
    swim_window_open(&GameWindow, GlobalParameters.display.width,
        GlobalParameters.display.height, (COLOR_T*)myLcdHnd, 0, 0,
        (GlobalParameters.display.width - 1),
        (GlobalParameters.display.height - 1), 0, BLACK, JAW_PLAY_BACKGROUND_COLOR, JAW_PLAY_BACKGROUND_COLOR);

    log_i("jaw_InitGraphics: Initial drawing\r\n");

    // Draw board
    result = PrepareBackground();
    if (result != E_OK)
      break;

    log_i("jaw_InitGraphics: Init complete\r\n");

    // Displaying it on the screen
    anydisplay_activateBuffer(myLcdHnd);

  } while (0);

  if (result != E_OK)
  {
    log_i("jaw_InitGraphics: Got error %u\r\n", result);
    jaw_DeinitGraphics();
  }

  return result;
}

tJawError
jaw_DeinitGraphics(void)
{
  int i;

  // Unload any images
  for (i = 0; i < JAW_NUM_PREPARED_IMG; i++)
  {
    if (PreparedImages[i] != NULL)
    {
      lpc_free(PreparedImages[i]);
      PreparedImages[i] = NULL;
    }
  }

  // Unload background image
  if (EmptyCellImage != NULL)
  {
    lpc_free(EmptyCellImage);
    EmptyCellImage = NULL;
  }

  // Unload LCD resources (if any)
  if (myLcdHnd != BACK_BUFFER_INVALID)
  {
    swim_window_close(&GameWindow);
    anydisplay_freeBackBuffer(myLcdHnd);
    myLcdHnd = BACK_BUFFER_INVALID;
  }

  return E_OK;
}

tJawError
jaw_RestartGraphics(void)
{
  return PrepareBackground();
}

tJawError
jaw_UpdateCell(tCell* pCell)
{
  //DISABLED: if (pCell->lastRenderedState != pCell->state)
  {
    switch(pCell->state)
    {
    case CELL_STATE_EMPTY:
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)EmptyCellImage,
          JAW_CELL_X(pCell->col),
          JAW_CELL_Y(pCell->row),
          JAW_CELL_WIDTH,
          JAW_CELL_HEIGHT);
      break;

    case CELL_STATE_SELECTED:
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)PreparedImages[JAW_OFFSET_SELECTED + pCell->type],
          JAW_CELL_X(pCell->col),
          JAW_CELL_Y(pCell->row),
          JAW_CELL_WIDTH,
          JAW_CELL_HEIGHT);
      break;

    case CELL_STATE_UNSELECTED:
      swimex_put_image_at_position(&GameWindow,
          (COLOR_T*)PreparedImages[JAW_OFFSET_UNSELECTED + pCell->type],
          JAW_CELL_X(pCell->col),
          JAW_CELL_Y(pCell->row),
          JAW_CELL_WIDTH,
          JAW_CELL_HEIGHT);
      break;

    default:
      return E_INVALID_STATE;
    }

    // mark to avoid repaints
    pCell->lastRenderedState = pCell->state;
  }

  return E_OK;
}

tJawError
jaw_UpdateAllCells(void)
{
  int i;
  int j;
  for (i = 0; i < JAW_NUM_ROWS; i++)
  {
    for (j = 0; j < JAW_NUM_COLUMNS; j++)
    {
      //TMP
      tCell* pCell = JAW_POINTER_TO_CELL(i, j);
      if ((pCell->row != i) || (pCell->col != j))
      {
        log_i("Mismatch: [%d,%d] should have been [%d,%d]\r\n", pCell->row, pCell->col, i, j);
      }
      //TMP
      jaw_UpdateCell(JAW_POINTER_TO_CELL(i, j));
    }
  }
  return E_OK;
}

tCell*
jaw_GetCellForXY(uint16_t x, uint16_t y)
{
  if ((x < JAW_BOARD_START_X) ||
      (y < JAW_BOARD_START_Y) ||
      (x > (JAW_BOARD_START_X+JAW_BOARD_WIDTH)) ||
      (y > (JAW_BOARD_START_Y+JAW_BOARD_HEIGHT)))
  {
    // clicked outside the board
    log_i("clicked outside board (%u, %u)\r\n", x, y);
    return NULL;
  }
  else
  {
    uint16_t boardX = x - JAW_BOARD_START_X;
    uint16_t boardY = y - JAW_BOARD_START_Y;

    // detect cell
    uint16_t row = boardY / (JAW_CELL_HEIGHT + JAW_VERTICAL_SPACING);
    uint16_t col = boardX / (JAW_CELL_WIDTH + JAW_HORIZONTAL_SPACING);

    // make sure we are within cell center
    uint16_t cellX = x - JAW_CELL_X(col);
    uint16_t cellY = y - JAW_CELL_Y(row);

    if ((cellX > JAW_TOUCH_MARGIN_X) &&
        (cellX < (JAW_CELL_WIDTH - JAW_TOUCH_MARGIN_X)) &&
        (cellY > JAW_TOUCH_MARGIN_Y) &&
        (cellY < (JAW_CELL_HEIGHT - JAW_TOUCH_MARGIN_Y)))
    {
      // Clicked inside cell
      return JAW_POINTER_TO_CELL(row, col);
    }
    else
    {
      // missed touch margin
      log_i("Missed boardX=%u, boardY=%u, cellX=%u, cellY=%u, row=%u, col=%u, x=%u, y=%u\r\n", boardX, boardY, cellX, cellY, row, col, x, y);
    }
  }

  return NULL;
}

Bool
jaw_IsButtonClick(Bool* pBtnNew, Bool* pBtnQuit, uint16_t x, uint16_t y, uint16_t p, Bool released)
{
  *pBtnQuit = FALSE;
  *pBtnNew = FALSE;

  if ((x > JAW_BTN_NEW_X) && (x < (JAW_BTN_NEW_X + JAW_BUTTON_WIDTH)) &&
      (y > JAW_BTN_NEW_Y) && (y < (JAW_BTN_NEW_Y + JAW_BUTTON_HEIGHT)))
  {
    if (released)
    {
      *pBtnNew = TRUE;
      DrawButtonNew(FALSE, TRUE);
    }
    else
    {
      DrawButtonNew(TRUE, FALSE);
    }
  }
  else if ((x > JAW_BTN_QUIT_X) && (x < (JAW_BTN_QUIT_X + JAW_BUTTON_WIDTH)) &&
           (y > JAW_BTN_QUIT_Y) && (y < (JAW_BTN_QUIT_Y + JAW_BUTTON_HEIGHT)))
  {
    if (released)
    {
      *pBtnQuit = TRUE;
      DrawButtonQuit(FALSE, TRUE);
    }
    else
    {
      DrawButtonQuit(TRUE, FALSE);
    }
  }
  else
  {
    DrawButtonNew(FALSE, FALSE);
    DrawButtonQuit(FALSE, FALSE);
  }

  return (*pBtnNew) || (*pBtnQuit);
}

tJawError
jaw_ShowGameOverScreen(uint32_t totalScore)
{
  char buf[10];
  int x = (JAW_WINDOW_WIDTH-JAW_GAME_OVER_SCREEN_WIDTH)/2;
  int y = (JAW_WINDOW_HEIGHT-JAW_GAME_OVER_SCREEN_HEIGHT)/3;
  int w = JAW_GAME_OVER_SCREEN_WIDTH;
  int h = JAW_GAME_OVER_SCREEN_HEIGHT;

  swimex_fill_rect(&GameWindow, x-2, y-2, x+w+2, y+h+2, JAW_GAME_FRAME_COLOR);
  swimex_fill_rect(&GameWindow, x, y, x+w, y+h, JAW_GAME_OVER_SCREEN_COLOR);

  swim_set_pen_color(&GameWindow, BLACK);
  swim_set_bkg_color(&GameWindow, JAW_GAME_OVER_SCREEN_COLOR);
  DrawCenterString(x, y, w, h/4, "GAME OVER!");
  DrawCenterString(x, y+h/2, w, h/4, "Your Score:");

  (void)int2String(totalScore, &(buf[0]));

  DrawCenterString(x, y+h/2+h/4, w, h/4, &(buf[0]));

  jaw_UpdateExpectedScore(0);
  jaw_UpdateTotalScore(0);

  return E_OK;
}

tJawError
jaw_UpdateExpectedScore(uint32_t score)
{
  int i;
  char buff[25] = {'E','x','p','e','c','t','e','d',' ','S','c','o','r','e',':',' ','\0'};
  (void)int2String(score, &(buff[16]));
  for (i = strlen(&(buff[0])); i < 25; i++)
  {
    buff[i] = ' '; // pad to avoid old digits shining through
  }
  buff[24] = '\0';
  swim_set_pen_color(&GameWindow, JAW_GAME_FOREGROUND_COLOR);
  swim_set_bkg_color(&GameWindow, JAW_GAME_BACKGROUND_COLOR);
  swim_put_text_xy(&GameWindow, &(buff[0]), jawLabelExpectedX, jawLabelExpectedY);

  return E_OK;
}

tJawError
jaw_UpdateTotalScore(uint32_t score)
{
  int i;
  char buff[25] = {'T', 'o', 't', 'a','l',' ','S','c','o','r','e',':',' ','\0'};
  (void)int2String(score, &(buff[13]));
  for (i = strlen(&(buff[0])); i < 25; i++)
  {
    buff[i] = ' '; // pad to avoid old digits shining through
  }
  buff[24] = '\0';
  swim_set_pen_color(&GameWindow, JAW_GAME_FOREGROUND_COLOR);
  swim_set_bkg_color(&GameWindow, JAW_GAME_BACKGROUND_COLOR);
  swim_put_text_xy(&GameWindow, &(buff[0]), jawLabelTotalX, jawLabelTotalY);

  return E_OK;
}
