/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/*
 * Code from http://rosettacode.org/wiki/Mandelbrot_set#PPM_non_interactive
 *
 * c program:
 * --------------------------------
 *  1. draws Mandelbrot set for Fc(z)=z*z +c
 *     using Mandelbrot algorithm ( boolean escape time )
 * -------------------------------
 *  2. technique of creating ppm file is  based on the code of Claudio Rocchini
 *  http://en.wikipedia.org/wiki/Image:Color_complex_plot.jpg
 *  create 24 bit color graphic file ,  portable pixmap file = PPM
 *  see http://en.wikipedia.org/wiki/Portable_pixmap
 *  to see the file use external application ( graphic viewer)
 */

/*
 * Improved on code with ideas from http://mrob.com/pub/muency/successiverefinement.html
 */


/******************************************************************************
 * Includes
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "lpc_swim.h"
#include "joystick.h"
#include "application.h"
#include "any_display.h"
#include "swim_extensions.h"
#include "board.h"
#include "log.h"
#include "lpc_heap.h"
#include "math.h"
#include "tsc2046_touch.h"
#include "lpc177x_8x_timer.h"
#include "pca9532.h"

/******************************************************************************
 * Defines, macros, and typedefs
 *****************************************************************************/

#define fabs(_x) ((_x) < 0 ? -(_x) : (_x))

#define SETPIXEL(_val24) \
     ((unsigned short)(((_val24) & 0xF80000 >> 8) | ((_val24) & 0xFC00 >> 5) | ((_val24) & 0xF8 >> 3)))

#define SETPIXEL_RGB(_r, _g, _b) \
    (((unsigned short)(_r) & 0xF8) << 8) | (((unsigned short)(_g) & 0xFC) << 3) | (((unsigned short)(_b) & 0xF8) >> 3)

typedef enum { COLORMAP_TYPE_RED, COLORMAP_TYPE_GREEN, COLORMAP_NUM_TYPES } colormap_t;

typedef struct {
  double xpos;  // defaults to 0.6
  double ypos;  // defaults to 0
  double scale; // defaults to 1/256
} zoominfo_t;

typedef struct {
  int oldX;
  int oldY;
  int oldZ;
  int newX;
  int newY;
  int newZ;
  Bool released;
} clickinfo_t;

#define COUNTER(__y, __x)  (counters[(__y)*GlobalParameters.display.width + (__x)])

#define TL   (COUNTER(y - chunkSize, x - chunkSize) == refCounter)
#define LL   (COUNTER(y,             x - chunkSize) == refCounter)
#define BL   (COUNTER(y + chunkSize, x - chunkSize) == refCounter)

#define TR   (COUNTER(y - chunkSize, x + chunkSize) == refCounter)
#define RR   (COUNTER(y,             x + chunkSize) == refCounter)
#define BR   (COUNTER(y + chunkSize, x + chunkSize) == refCounter)

#define TT   (COUNTER(y - chunkSize, x            ) == refCounter)
#define BB   (COUNTER(y + chunkSize, x            ) == refCounter)


/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static int32_t activeBuffer = 1;
static SWIM_WINDOW_T MainWin;
static Bool started = FALSE;
static Bool userInteraction = FALSE;
static uint16_t* colormap = NULL;
static int colormapIterations = -1;
static colormap_t colormapType = COLORMAP_NUM_TYPES;

static BackBuffer_t myLcdHnd1 = BACK_BUFFER_INVALID;
static BackBuffer_t myLcdHnd2 = BACK_BUFFER_INVALID;
static BackBuffer_t myCounters = BACK_BUFFER_INVALID;

static zoominfo_t ZoomInfo;
static zoominfo_t nextZoomInfo;

static clickinfo_t ClickInfo;

static uint32_t times[] = {0,0,0,0,0};
static int chunk_sizes[] = {16, 8, 4, 2, 1};
static int chunkIndex = 0;

/******************************************************************************
 * Local functions
 *****************************************************************************/

static Bool
allNeighborsSame(int x, int y, int chunkSize)
{
  uint16_t* counters = (uint16_t*)myCounters;
  uint16_t refCounter = counters[y*GlobalParameters.display.width + x];

  if (x < chunkSize)
  {
    if (y < chunkSize)
    {
      /*
       *       TOP LEFT CORNER
       *     ______ ______ ______
       *    |      |      |      |
       *    |      |      |      |
       *    |______|______|______|
       *    |      |      |      |
       *    |      |  REF |  RR  |
       *    |______|______|______|
       *    |      |      |      |
       *    |      |  BB  |  BR  |
       *    |______|______|______|
       *
       */
      return BB && RR && BR;
    }
    else
    {
      /*
       *         LEFT SIDE
       *     ______ ______ ______
       *    |      |      |      |
       *    |      |  TT  |  TR  |
       *    |______|______|______|
       *    |      |      |      |
       *    |      |  REF |  RR  |
       *    |______|______|______|
       *    |      |      |      |
       *    |      |  BB  |  BR  |
       *    |______|______|______|
       *
       */
      return TT && TR && RR && BR && BB;
    }
  }
  else if (y < chunkSize)
  {
    /*
     *           TOP SIDE
     *     ______ ______ ______
     *    |      |      |      |
     *    |      |      |      |
     *    |______|______|______|
     *    |      |      |      |
     *    |  LL  |  REF |  RR  |
     *    |______|______|______|
     *    |      |      |      |
     *    |  BL  |  BB  |  BR  |
     *    |______|______|______|
     *
     */
    return LL && BL && BB && BR && RR;
  }
  else
  {
    /*
     *         IN THE MIDDLE
     *     ______ ______ ______
     *    |      |      |      |
     *    |  TL  |  TT  |  TR  |
     *    |______|______|______|
     *    |      |      |      |
     *    |  LL  |  REF |  RR  |
     *    |______|______|______|
     *    |      |      |      |
     *    |  BL  |  BB  |  BR  |
     *    |______|______|______|
     *
     */
    return LL && BL && BB && BR && RR && TR && TT && TL;
  }
}

static Bool
render(uint16_t* pBuff, int IterationMax)
{
  /* next chunk size */
  int chunkSize = chunk_sizes[chunkIndex++];

  /* screen ( integer) coordinate */
  int iX, iY;
  const int iXmax = GlobalParameters.display.width;
  const int iYmax = GlobalParameters.display.height;

  /* world ( double) coordinate = parameter plane*/
  double Cx, Cy;
  const double CyMin = -2.0;
  const double CyMax = 2.0;

  /* */
  double PixelHeight = (CyMax - CyMin) / iYmax;

  /* Z=Zx+Zy*i  ;   Z0 = 0 */
  double Zx, Zy;
  double Zx2, Zy2; /* Zx2=Zx*Zx;  Zy2=Zy*Zy  */

  /*  */
  int Iteration;

  /* bail-out value , radius of circle ;  */
  const double EscapeRadius = 2;
  double ER2 = EscapeRadius * EscapeRadius;

  Bool DoEvaluate;
  int parentX, parentY;

  /* compute and write image data bytes to the file*/
  for (iY = 0; iY <= (iYmax-chunkSize); iY+=chunkSize) {
    Cy = (iY - iYmax/2) * ZoomInfo.scale + ZoomInfo.ypos;
    if (fabs(Cy) < PixelHeight / 2)
      Cy = 0.0; /* Main antenna */
    for (iX = 0; iX <= (iXmax-chunkSize); iX+=chunkSize) {

      if ((iX%(2*chunkSize)) == 0) {
        parentX = iX;
      } else {
        parentX = iX - chunkSize;
      }
      if ((iY%(2*chunkSize)) == 0) {
        parentY = iY;
      } else {
        parentY = iY - chunkSize;
      }

      if (chunkSize == 16) {
        DoEvaluate = TRUE;
      } else if ((iX == parentX) && (iY == parentY)) {
        DoEvaluate = FALSE;
      } else if (allNeighborsSame(parentX, parentY, chunkSize*2)) {
        DoEvaluate = FALSE;
      } else {
        DoEvaluate = TRUE;
      }

      if (DoEvaluate)
      {
        Cx = (iX - iXmax/2) * ZoomInfo.scale + ZoomInfo.xpos;

        /* initial value of orbit = critical point Z= 0 */
        Zx = 0.0;
        Zy = 0.0;
        Zx2 = Zx * Zx;
        Zy2 = Zy * Zy;

        /* */
        for (Iteration = 0; Iteration < IterationMax && ((Zx2 + Zy2) < ER2); Iteration++) {
          Zy = 2 * Zx * Zy + Cy;
          Zx = Zx2 - Zy2 + Cx;
          Zx2 = Zx * Zx;
          Zy2 = Zy * Zy;
        };
      }
      else
      {
        Iteration = ((uint16_t*)myCounters)[parentY*iXmax + parentX];
      }

      ((uint16_t*)myCounters)[iY*iXmax + iX] = Iteration;

      /* Fill a chunk */
      {
        uint16_t color = colormap[Iteration];
        int pos = iY * iXmax + iX;
        int ax, by;
        for (by = 0; by < chunkSize; by++) {
          for (ax = 0; ax < chunkSize; ax++) {
            pBuff[pos+ax] = color;
          }
          pos += iXmax;
        }
      }
    }
    if (userInteraction) {
      // The user has interrupted the rendering. It should be enough to test
      // this once per drawn line. On a 800x400 display it would mean a delay
      // of 1/400 of the rendering time so it will appear to be immediate (<100ms)
      // long as the complete render time is less than 40s.
      return TRUE;
    }
  }
  times[chunkIndex-1] = msClock;
  return chunkIndex < 5; // five chunk sizes
}

static void
prepareColormap(int iterations, colormap_t type)
{
  int i;

  if (iterations == colormapIterations && type == colormapType) {
    // already prepared
    return;
  }
  if (colormap != NULL) {
    lpc_free(colormap);
  }
  colormap = lpc_new(2*(iterations+1)); // 2 bytes per color, 0..iterations

  if (type == COLORMAP_TYPE_RED) {
    int c;
    double a, b;

    a = log10(iterations - 1);

    colormap[0] = WHITE;
    for (i = 1; i < iterations; i++) {
      b = 3 * log10(i) / a;
      if (b < 1.0) {
        // scale red; green and blue are 0
        c = (int)((63 * b) + 1);
        colormap[i] = SETPIXEL_RGB(c<<3, 0, 0);
      } else if (b < 2.0) {
        // red max; scale green; blue is 0
        c = (int)((127 * (b-1)) + 1);
        colormap[i] = SETPIXEL_RGB(0xF8, c<<2, 0);
      } else {
        // red and green max; scale blue
        c = (int)((63 * (b-2)) + 1);
        colormap[i] = SETPIXEL_RGB(0xF8, 0xFC, c<<3);
      }
    }
    colormap[iterations] = BLACK;
  } else if (type == COLORMAP_TYPE_GREEN) {
    colormap[0] = WHITE;
    for (i = 1; i < iterations; i++) {
      colormap[i] = SETPIXEL(i | (i << 8));
    }
    colormap[iterations] = BLACK;
  }

  colormapIterations = iterations;
  colormapType = type;
}

static void
swapBuffers(void)
{
  if (activeBuffer == 1) {
    anydisplay_activateBuffer(myLcdHnd1);
    MainWin.fb = (COLOR_T*)myLcdHnd2;
    activeBuffer = 2;
    memcpy((uint8_t*)myLcdHnd2, (uint8_t*)myLcdHnd1, GlobalParameters.display.buffersize);
  } else {
    anydisplay_activateBuffer(myLcdHnd2);
    MainWin.fb = (COLOR_T*)myLcdHnd1;
    activeBuffer = 1;
    memcpy((uint8_t*)myLcdHnd1, (uint8_t*)myLcdHnd2, GlobalParameters.display.buffersize);
  }
}

static void runApp(void)
{
  uint32_t iterations = 50;
  Bool draw = TRUE;

  if (!anydisplay_createBackBuffers(&myLcdHnd1, &myLcdHnd2, &myCounters))
  {
    log_e("Failed to allocate back buffers\r\n");
    return;
  }

  /* Create a SWIM window */
  swim_window_open(&MainWin, GlobalParameters.display.width,
      GlobalParameters.display.height, (COLOR_T*)myLcdHnd1, 0, 0,
      (GlobalParameters.display.width - 1),
      (GlobalParameters.display.height - 1), 1, WHITE, BLACK, BLACK);
  activeBuffer = 1;
  userInteraction = FALSE;
  started = TRUE;
  chunkIndex = 0;

  ZoomInfo.scale = 4/256.0;
  ZoomInfo.xpos = -0.7;
  ZoomInfo.ypos = 0;
  nextZoomInfo = ZoomInfo;

  ClickInfo.oldX = ClickInfo.oldY = ClickInfo.oldZ = 0;
  ClickInfo.released = TRUE;

  prepareColormap(iterations, COLORMAP_TYPE_GREEN);

  while (started) {
    if (draw) {
      draw = render((uint16_t*)MainWin.fb, iterations);
      swapBuffers();
      if (!draw) {
        int i;
        for (i = 1; i < 5; i++) {
          log_i("Iteration %d took %d\r\n", i, times[i]-times[i-1]);
        }
      }
    } else {
      TIM_Waitms(50);
    }
    if (userInteraction && started)
    {
      // Know it is a click event, as the quit button/joystick
      // results in setting of started=FALSE

      ZoomInfo = nextZoomInfo;
      draw = TRUE;

      /* Reset the old image */
      swim_clear_screen(&MainWin, BLACK);

      /* Must restore the chunk sizes again */
      chunkIndex = 0;

      // Reset the flag to be ready for next event
      userInteraction = FALSE;
    }
  }
}

static void cleanup(void)
{
  started = FALSE;
  swim_window_close(&MainWin);
  if (colormap != NULL) {
    lpc_free(colormap);
    colormap = NULL;
    colormapIterations = -1;
    colormapType = COLORMAP_NUM_TYPES;
  }

  anydisplay_freeBackBuffers(&myLcdHnd1, &myLcdHnd2, &myCounters);
}

static void timeout(void)
{
  if (started)
  {
    touch_xyz(&ClickInfo.newX, &ClickInfo.newY, &ClickInfo.newZ);
    if (ClickInfo.newZ == 0 && ClickInfo.oldZ != 0) {
      // ignore x and y coordinates as they are wrong when z becomes 0
      ClickInfo.oldZ = 0;
      ClickInfo.released = TRUE;

      // Can only modify the new zoom information when the main loop is
      // not accessing it
      if (!userInteraction)
      {
        // Image is currently centered around ZoomInfo
        nextZoomInfo.xpos = ZoomInfo.xpos + (ClickInfo.oldX - GlobalParameters.display.width / 2) * ZoomInfo.scale;
        nextZoomInfo.ypos = ZoomInfo.ypos + (ClickInfo.oldY - GlobalParameters.display.height/ 2) * ZoomInfo.scale;

        if ((ZoomInfo.scale > (fabs(ClickInfo.oldX) * 1e-16)) && (ZoomInfo.scale > (fabs(ClickInfo.oldY) * 1e-16))) {
          nextZoomInfo.scale = ZoomInfo.scale / 2;
        }
        userInteraction = TRUE;
      }
    } else if ((ClickInfo.oldX != ClickInfo.newX) || (ClickInfo.oldY != ClickInfo.newY) || (ClickInfo.oldZ == 0 && ClickInfo.newZ != 0)) {
      ClickInfo.oldX = ClickInfo.newX;
      ClickInfo.oldY = ClickInfo.newY;
      ClickInfo.oldZ = ClickInfo.newZ;
      if (ClickInfo.released && ClickInfo.oldZ > 0) {
        ClickInfo.released = FALSE;
      }
    } else {
      if (!userInteraction) {
        uint8_t joy = joystick_read();
        if (joy & JOYSTICK_DOWN) {
          nextZoomInfo.ypos = ZoomInfo.ypos + (GlobalParameters.display.height / 8) * ZoomInfo.scale;
        } else if (joy & JOYSTICK_UP) {
          nextZoomInfo.ypos = ZoomInfo.ypos - (GlobalParameters.display.height / 8) * ZoomInfo.scale;
        }
        if (joy & JOYSTICK_LEFT) {
          nextZoomInfo.xpos = ZoomInfo.xpos - (GlobalParameters.display.width / 8) * ZoomInfo.scale;
        } else if (joy & JOYSTICK_RIGHT) {
          nextZoomInfo.xpos = ZoomInfo.xpos + (GlobalParameters.display.width / 8) * ZoomInfo.scale;
        }
        if (joy & JOYSTICK_CENTER) {
          if ((ZoomInfo.scale > (fabs(ClickInfo.oldX) * 1e-16)) && (ZoomInfo.scale > (fabs(ClickInfo.oldY) * 1e-16))) {
            nextZoomInfo.scale = ZoomInfo.scale / 2;
          }
        }
        if (joy) {
          userInteraction = TRUE;
        }
      }
    }

    /* Check SW2 button which causes Exit */
    if (pca9532_getLedState(FALSE) & KEY1) {
      started = FALSE;
      userInteraction = TRUE;
    }
  }
}

/******************************************************************************
 * Public Functions
 *****************************************************************************/

static AppCallback_t mandelbrotCallback = {
    &runApp,
    &cleanup,
    &timeout,
    100,
    FALSE,
};

AppCallback_t* mandelbrotAppInit(void)
{
  return &mandelbrotCallback;
}

