/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include <stdio.h>
#include "lpc_types.h"
#include "application.h"
#include "lpc177x_8x_timer.h"
#include "tsc2046_touch.h"
#include "image.h"
#include "lpc_heap.h"
#include "lpc_swim.h"
#include "lpc_swim_font.h"
#include "lpc_swim_image.h"
#include "globals.h"
#include "log.h"
#include "joystick.h"
#include "progress.h"

/******************************************************************************
 * Defines, macros, and typedefs
 *****************************************************************************/

#define ICON_ANIM_NUM         4
#define ICON_ANIM_TIME_MS   120
#define ICON_ANIM_START_IDX   1

typedef struct
{
  const char*    pName;
  uint16_t*      pIconData[ICON_ANIM_NUM];
  SWIM_WINDOW_T* pIconWindow;
} tApplication;

#define MENU_TITLE_X       0
#define MENU_TITLE_Y       0
#define MENU_TITLE_HEIGHT 16
#define MENU_TITLE_WIDTH  DISPLAY_WIDTH

#define MENU_ICON_TOP_MARGIN   1
#define MENU_ICON_LEFT_MARGIN  12

#define ICON_WIDTH   72
#define ICON_HEIGHT  72
#define ICON_BPP     24

#define NUM_APPLICATIONS (sizeof(applications) / sizeof(tApplication) )

#define NUM_BOOT_IMG (sizeof(bootImgs) / sizeof(tBootImg))

#define MENU_ICON_START_PATH "/menu/icon"
#define MENU_ICON_END_PATH   ".png"
#define MENU_ICON_MID_PATH   "_"

#define BACK_COLOR             (WHITE)
#define MENU_TITLE_COLOR       (0x3ACD)  /* OLIVE */
#define MENU_TITLE_FONT_COLOR  (GREEN)


/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static SWIM_WINDOW_T* pMessageWindow;
static SWIM_WINDOW_T* pTitleWindow;
static SWIM_WINDOW_T* pBaseWindow;

static uint8_t selectedIcon = 4;
static int16_t releasedIcon = -1;
static uint8_t iconAnimIndex = ICON_ANIM_START_IDX;
static int8_t iconAnimDir = 1;

static tApplication applications[12];
static SWIM_WINDOW_T AppWindows[NUM_APPLICATIONS];
static SWIM_WINDOW_T MessageWin;
static SWIM_WINDOW_T TitleWin;
static SWIM_WINDOW_T BaseWindow;

static int numIconRows = 4;
static int numIconCols = 3;

/******************************************************************************
 * Local functions
 *****************************************************************************/

static ImageError_t
uiLoadIcons(void)
{
  int i = 0;
  int j = 0;
  uint16_t w = 0;
  uint16_t h = 0;
  ImageError_t result;
  char path[30];

  for (j = 0; j < ICON_ANIM_NUM; j++)
  {
    progress_set((j+1)*20);
    for (i = 0; i < NUM_APPLICATIONS; i++)
    {
      if (applications[i].pIconData[j] == NULL) {
        sprintf(path, MENU_ICON_START_PATH "%d" MENU_ICON_MID_PATH "%d" MENU_ICON_END_PATH, i, j);
        result = decodeImageFile(ONLY_USE_ROMFS, path, (uint8_t**)(&(applications[i].pIconData[j])), &w, &h);
        if (result != IMAGE_ERROR_OK) {
          return result;
        } else if ((w != ICON_WIDTH) || (h != ICON_HEIGHT)) {
          return IMAGE_ERROR_BAD_IMAGE_SIZE;
        }
      }
    }
  }

  return IMAGE_ERROR_OK;
}

static void
uiPaintSelectedIcon(void)
{
  swim_put_image(applications[selectedIcon].pIconWindow,
      (const COLOR_T*)applications[selectedIcon].pIconData[iconAnimIndex],
      ICON_WIDTH, ICON_HEIGHT);

  if (iconAnimIndex + iconAnimDir == ICON_ANIM_NUM)
    iconAnimDir = -1;
  else if (iconAnimIndex == ICON_ANIM_START_IDX && iconAnimDir < 0)
    iconAnimDir = 1;


  iconAnimIndex = (iconAnimIndex + iconAnimDir);
}

static void
uiEraseMessageArea(void)
{
  swim_clear_screen(pMessageWindow, BACK_COLOR);
}

static void
uiPaintIconArea(void)
{
  int i = 0;

  for (i = 0; i < NUM_APPLICATIONS; i++)
  {
    swim_put_image(applications[i].pIconWindow,
        (const COLOR_T*)applications[i].pIconData[0],
        ICON_WIDTH, ICON_HEIGHT);
  }

  uiPaintSelectedIcon();
  uiEraseMessageArea();
}

static void
uiPaintMessage(char* pMsg)
{
  swim_set_xy(pMessageWindow, 0, 0);
  swim_put_text(pMessageWindow, pMsg);
}

static void
uiPaintTitle(const char* pTitle)
{
  swim_clear_screen(pTitleWindow, MENU_TITLE_COLOR);
  swim_set_xy(pTitleWindow, 5, 1);
  swim_put_text(pTitleWindow, pTitle);
}

static void
uiPaintBackground(void)
{
  swim_clear_screen(&BaseWindow, BACK_COLOR);
}

static void
uiRedraw(void)
{
  uiPaintBackground();
  uiPaintTitle(applications[selectedIcon].pName);
  uiPaintIconArea();
}

static void
selectedIconChanged(void)
{
  iconAnimIndex = ICON_ANIM_START_IDX;
  uiPaintIconArea();
  uiPaintTitle(applications[selectedIcon].pName);
  releasedIcon = -1;
}

static void
selectedIconClicked(void)
{
  releasedIcon = -1;

  if (appStartApp(applications[selectedIcon].pName))
  {
    // Time to remove all traces of the running app
    uiRedraw();
  }
  else
  {
    uiPaintMessage("Failed to start application");
    printf("Failed to start application %d (%s)\r\n", selectedIcon,
        applications[selectedIcon].pName);
  }
}

static void
selectIcon(uint16_t x, uint16_t y, Bool released)
{
  uint16_t topX = MENU_ICON_LEFT_MARGIN;
  uint16_t topY = MENU_ICON_TOP_MARGIN + MENU_TITLE_HEIGHT;
  uint16_t width = numIconCols*ICON_WIDTH;
  uint16_t height = numIconRows*ICON_HEIGHT;
  uint16_t sel = 0;
  uint8_t oldValue = selectedIcon;

  if (x < topX || y < topY)
    return;
  if(x > topX+width || y > topY+height)
    return;

  if (released)
  {
    releasedIcon = selectedIcon;
    return;
  }

  // get selected icon
  sel = (x - MENU_ICON_LEFT_MARGIN) / ICON_WIDTH;
  sel += ((y-MENU_ICON_TOP_MARGIN-MENU_TITLE_HEIGHT) / ICON_HEIGHT)*numIconCols;

  selectedIcon = sel;
  if (oldValue != selectedIcon)
  {
    selectedIconChanged();
  }
  else if (releasedIcon == selectedIcon)
  {
    selectedIconClicked();
  }
}

static SWIM_WINDOW_T* createWindowNoBorder(SWIM_WINDOW_T* pWin, int x, int y, int width, int height)
{
  if (width == -1) {
    width = GlobalParameters.display.width - x; /* use remaining width */
  }
  if (height == -1) {
    height = GlobalParameters.display.height - y; /* use remaining height */
  }
  swim_window_open(pWin,
      GlobalParameters.display.width,
      GlobalParameters.display.height,
      (COLOR_T*)GlobalParameters.display.framebuffer,
      x, y,
      x + width - 1,
      y + height - 1,
      0, WHITE, BLACK, BLACK);
  return pWin;
}

static void cleanup(void)
{
  int i, j;
  for (i = 0; i < NUM_APPLICATIONS; i++)
  {
    for (j = 0; j < ICON_ANIM_NUM; j++)
    {
      if (applications[i].pIconData[j] != NULL) {
        lpc_free(applications[i].pIconData[j]);
        applications[i].pIconData[j] = NULL;
      }
    }
    if (applications[i].pIconWindow != NULL) {
      swim_window_close(applications[i].pIconWindow);
      applications[i].pIconWindow = NULL;
    }
    applications[i].pName = NULL;
  }
  if (pMessageWindow != NULL) {
    swim_window_close(pMessageWindow);
    pMessageWindow = NULL;
  }
  if (pTitleWindow != NULL) {
    swim_window_close(pTitleWindow);
    pTitleWindow = NULL;
  }
  if (pBaseWindow != NULL) {
    swim_window_close(pBaseWindow);
    pBaseWindow = NULL;
  }
}

static void showMenu(void)
{
  int i;
  int x, y, z;
  int x2, y2, z2;
  Bool released = FALSE;

  joystick_init();

  /* Initialize everything to allow cleanup */
  memset(applications, 0, sizeof(applications));
  pMessageWindow = NULL;
  pTitleWindow = NULL;
  pBaseWindow = NULL;

  pBaseWindow = createWindowNoBorder(&BaseWindow, 0, 0, -1, -1);

  if (GlobalParameters.display.height < 320)
  {
    // The 4.3" display has a 480x272 resolution on which the four
    // rows of menu icons won't fit. Change to 3 rows of 4 icons instead.
    numIconRows = 3;
    numIconCols = 4;
  }

  for (i = 0; i < NUM_APPLICATIONS; i++) {
    applications[i].pName = appGetAppName(i+1); /* Skip ourself */

    applications[i].pIconWindow = createWindowNoBorder(&(AppWindows[i]),
        MENU_ICON_LEFT_MARGIN + ((i%numIconCols)*ICON_WIDTH),
        MENU_ICON_TOP_MARGIN + MENU_TITLE_HEIGHT + ((i/numIconCols)*ICON_HEIGHT),
        ICON_WIDTH,
        ICON_HEIGHT);
  }

  pMessageWindow = createWindowNoBorder(&MessageWin, 0,
      MENU_TITLE_HEIGHT + numIconRows*ICON_HEIGHT + MENU_ICON_TOP_MARGIN,
      GlobalParameters.display.width,
      -1);

  pTitleWindow = createWindowNoBorder(&TitleWin,
      MENU_TITLE_X,
      MENU_TITLE_Y,
      -1,
      MENU_TITLE_HEIGHT);
  swim_set_pen_color(pTitleWindow, MENU_TITLE_FONT_COLOR);
  swim_set_bkg_color(pTitleWindow, MENU_TITLE_COLOR);

  progress_init();

  if (uiLoadIcons() != IMAGE_ERROR_OK) {
    cleanup();
    log_e("failed to load icons for menu.\r\n");
  }

  progress_destroy();

  uiRedraw();

  x = y = z = 0;
  released = TRUE;
  while (1) {
    TIM_Waitms(ICON_ANIM_TIME_MS);

    touch_xyz(&x2, &y2, &z2);
    if (z2 == 0 && z != 0) {
      // ignore x and y coordinates as they are wrong when z becomes 0
      z = 0;
      released = TRUE;
      selectIcon(x, y, released);
    } else if ((x != x2) || (y != y2) || (z == 0 && z2 != 0)) {
      x = x2;
      y = y2;
      z = z2;
      if (released && z > 0) {
        released = FALSE;
      }

      selectIcon(x, y, released);
    } else {
      uint8_t joyState = 0;
      uint8_t change = 0;
      joyState = joystick_read();

      if (joyState & JOYSTICK_UP) {
        change += NUM_APPLICATIONS - numIconCols;
      }

      if (joyState & JOYSTICK_DOWN) {
        change += numIconCols;
      }

      if (joyState & JOYSTICK_LEFT) {
        change += NUM_APPLICATIONS - 1;
      }

      if (joyState & JOYSTICK_RIGHT) {
        change++;
      }

      if (change != 0) {
        selectedIcon = (selectedIcon + change) % NUM_APPLICATIONS;
        selectedIconChanged();
      } else if (joyState & JOYSTICK_CENTER) {
        selectedIconClicked();
      }
    }
    uiPaintSelectedIcon();
  }
}

/******************************************************************************
 * Public Functions
 *****************************************************************************/

static AppCallback_t menuCallback = {
    &showMenu,
    &cleanup,
    FALSE,
};

AppCallback_t* menuAppInit(void)
{
  return &menuCallback;
}
