/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include "slideshow.h"
#include "transitions.h"
#include "image.h"
#include "trans_fs.h"
#include "log.h"
#include "lpc_heap.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define TRANSITIONS_NUM_PREPARED_IMAGES  40

#define SLIDESHOW_SCRIPT_MAX_SIZE  80

#define SLIDE_MIN(a,b) (((a) <= (b)) ? (a) : (b))
#define SLIDE_MAX_SLEEP  50

#define NO_SLOT  -1

//#define C_(x) x
#define C_(x)

typedef enum
{
  COMMAND_CLEAR,
  COMMAND_GOTO,
  COMMAND_LOAD_IMG,
  COMMAND_SHOW,
  COMMAND_WAIT
} TCommand;

typedef struct
{
  TCommand Cmd;
  int Info;
  char* pFileName;

  TTransition Transition;
} TCommandInfo;

typedef struct
{
  uint8_t* pLabel;
  int index;
} TLabelInfo;

/******************************************************************************
 * External variables
 *****************************************************************************/

extern volatile uint32_t msClock;

/******************************************************************************
 * Public variables
 *****************************************************************************/

// The back buffer will be of twice normal screen size for some
// image magics
TImage ImageBackBuffer = NULL;

/******************************************************************************
 * Local variables
 *****************************************************************************/

static TImage PreparedImages[TRANSITIONS_NUM_PREPARED_IMAGES] = {0};

static TCommandInfo* Sequence = NULL;
static int allocatedSequenceSize = 0;
static int nextSequence = 0;

static int CurrentSlot = NO_SLOT;

static Bool* gpAbort;
static Bool* gpShouldRender;

// Position in the current script
static int seqIndex = 0;

// Indicates when it is time to do something when in wait mode
static uint32_t targetTime = 0;

/******************************************************************************
 * Local function prototypes
 *****************************************************************************/

// Parsing functions
static int getNextLine(uint8_t* pBuf, int* pOffset, uint8_t** ppLine);
static int splitLine(uint8_t* pLine, uint8_t** ppPart1, uint8_t** ppPart2, uint8_t** ppPart3);
static int findLabel(TLabelInfo* pLabels, int numLabels, uint8_t* pLabel);
static void freeSequence(void);
static int parseScript(uint8_t* pBuf, uint32_t* pDelayFactor);

// Command related functions
static uint16_t loadImage(char* pFileName, int slot);
static void doTransition(void);


/******************************************************************************
 * Local functions
 *****************************************************************************/

// pBuf in, pOffset in/out, pLine out
// returns 0 as long as a token is found
static int
getNextLine(uint8_t* pBuf, int* pOffset, uint8_t** ppLine)
{
  int pos = *pOffset;
  int result = -1;

  // trim whitespace from start of line
  while ((pBuf[pos] == ' ') || (pBuf[pos] == '\t'))
  {
    pos++;
  }
  *ppLine = &(pBuf[pos]);

  while (pBuf[pos] != '\0')
  {
    if ((pBuf[pos] == '\r') || (pBuf[pos] == '\n'))
    {
      // found the next end of line
      pBuf[pos++] = '\0';
      result = 0;

      // move past all end-of-line characters
      while ((pBuf[pos] == '\r') || (pBuf[pos] == '\n'))
      {
        pos++;
      }
      break;
    }
    pos++;
  }

  *pOffset = pos;
  return result;
}

// pLine in, ppPart1 out, ppPart2 out, ppPart3 out
// returns number of found parts
static int
splitLine(uint8_t* pLine, uint8_t** ppPart1, uint8_t** ppPart2, uint8_t** ppPart3)
{
  int pos = 0;
  int found = 0;

  *ppPart1 = NULL;
  *ppPart2 = NULL;
  *ppPart3 = NULL;

  if (*pLine != '\0')
  {
    *ppPart1 = &(pLine[0]);
    found++;

    while (pLine[pos] != '\0')
    {
      if (pLine[pos] == ' ')
      {
        // found the next token separator
        pLine[pos++] = '\0';
        found++;

        // move past all token separator characters
        while (pLine[pos] == ' ')
        {
          pos++;
        }

        // start looking for end of next token
        if (found == 2)
        {
          *ppPart2 = &(pLine[pos]);
        }
        else if (found == 3)
        {
          *ppPart3 = &(pLine[pos]);
        }
      }
      pos++;
    }
  }

  return found;
}

// returns index of pLabel or -1 if it doesn't exist
static int
findLabel(TLabelInfo* pLabels, int numLabels, uint8_t* pLabel)
{
  int i;
  for (i = 0; i < numLabels; i++)
  {
    if (strcmp((char*)(pLabels[i].pLabel), (char*)pLabel) == 0)
    {
      return pLabels[i].index;
    }
  }
  return -1;
}

// releases all resources used by the current sequence (if any)
static void
freeSequence(void)
{
  int i;
  if (Sequence != NULL)
  {
    for (i = 0; i < nextSequence; i++)
    {
      if ((Sequence[i].Cmd == COMMAND_LOAD_IMG) && (Sequence[i].pFileName != NULL))
      {
        lpc_free(Sequence[i].pFileName);
        Sequence[i].pFileName = NULL;
      }
    }

    lpc_free(Sequence);
  }
  nextSequence = 0;
  Sequence = NULL;
}

// allocates space for the new sequence, reallocating if needed
static void
allocateSequence(int wantedSize)
{
  if (Sequence == NULL)
  {
    nextSequence = 0;
    Sequence = lpc_new(sizeof(TCommandInfo) * wantedSize);
    memset(Sequence, 0, sizeof(TCommandInfo) * wantedSize);
    allocatedSequenceSize = wantedSize;
  }
  else
  {
    TCommandInfo* NewSequence = lpc_new(sizeof(TCommandInfo) * wantedSize);
    memset(NewSequence, 0, sizeof(TCommandInfo) * wantedSize);
    memcpy(NewSequence, Sequence, sizeof(TCommandInfo) * allocatedSequenceSize);
    lpc_free(Sequence);
    Sequence = NewSequence;
    allocatedSequenceSize = wantedSize;

    // Don't update the nextSequence
  }
}

static int
parseScript(uint8_t* pBuf, uint32_t* pDelayFactor)
{
  uint8_t* pLine = NULL;
  int offset = 0;
  TLabelInfo Labels[10];
  int numLabels = 0;
  TLabelInfo UnresolvedGotos[10];
  int numUnresolvedGotos = 0;
  int i;

  // initializing data
  memset(Labels, 0, sizeof(TLabelInfo)*10);
  memset(UnresolvedGotos, 0, sizeof(TLabelInfo)*10);

  // cleanup old sequences
  freeSequence();

  // prepare the new one
  allocateSequence(SLIDESHOW_SCRIPT_MAX_SIZE);

  // start parsing the new sequence
  while (getNextLine(pBuf, &offset, &pLine) == 0)
  {
    if (*pLine == '#')
    {
      // found a comment line
    }
    else
    {
      uint8_t* pCommand;
      uint8_t* pArg1;
      uint8_t* pArg2;
      int num = splitLine(pLine, &pCommand, &pArg1, &pArg2);

      if ((num == 1) && (strcmp((char*)pCommand, "clear") == 0))
      {
        Sequence[nextSequence].Cmd = COMMAND_CLEAR;
        Sequence[nextSequence].Transition = CLEAR_SCREEN;
      }
      else if ((num == 2) && (strcmp((char*)pCommand, "delayfactor") == 0))
      {
        *pDelayFactor = atoi((char*)pArg1);
      }
      else if ((num == 3) && (strcmp((char*)pCommand, "show") == 0))
      {
        Sequence[nextSequence].Cmd = COMMAND_SHOW;
        Sequence[nextSequence].Info = atoi((char*)pArg1);

        for (i = 0; Transitions[i].pTransitionString != NULL; i++)
        {
          if (strcmp(Transitions[i].pTransitionString, (char*)pArg2) == 0)
          {
            Sequence[nextSequence].Transition = Transitions[i].Transition;
            break;
          }
        }
        if (Transitions[i].pTransitionString == NULL)
        {
          log_i("Found invalid transition '%s'. Aborting...\r\n", pArg2);
          return -1;
        }
      }
      else if ((num == 2) && (strcmp((char*)pCommand, "wait") == 0))
      {
        Sequence[nextSequence].Cmd = COMMAND_WAIT;
        Sequence[nextSequence].Info = atoi((char*)pArg1);
      }
      else if ((num == 2) && (strcmp((char*)pCommand, "label") == 0))
      {
        int index = findLabel(&(Labels[0]), numLabels, pArg1);
        if (index == -1)
        {
          // found a new label
          Labels[numLabels].index = nextSequence;
          Labels[numLabels].pLabel = pArg1;
          numLabels++;

          // A label doesn't occupy a slot in the sequence
          nextSequence--;
        }
        else
        {
          // label already declared
          log_i("Found a second declaration of label '%s'. Aborting...\r\n", pArg1);
          return -2;
        }
      }
      else if ((num == 2) && (strcmp((char*)pCommand, "goto") == 0))
      {
        int index = findLabel(&(Labels[0]), numLabels, pArg1);
        if (index == -1)
        {
          // couldn't find the label we are looking for so we
          // wait for now
          UnresolvedGotos[numUnresolvedGotos].index = nextSequence;
          UnresolvedGotos[numUnresolvedGotos].pLabel = pArg1;
          numUnresolvedGotos++;
        }

        // Create the command
        Sequence[nextSequence].Cmd = COMMAND_GOTO;
        Sequence[nextSequence].Info = index;
      }
      else if ((num == 3) && (strcmp((char*)pCommand, "load") == 0))
      {
        size_t numBytes;

        Sequence[nextSequence].Cmd = COMMAND_LOAD_IMG;
        Sequence[nextSequence].Info = atoi((char*)pArg2);

        // copy filename
        numBytes = sizeof(char) * (strlen((char*)pArg1) + 1);
        Sequence[nextSequence].pFileName = lpc_new(numBytes);
        memcpy(Sequence[nextSequence].pFileName, pArg1, numBytes);
      }
      else
      {
        // unknown command
        log_i("Found unknown command '%s'. Aborting...\r\n", pCommand);
        return -3;
      }

      // start looking for next part in the sequence
      nextSequence++;

      // assure we don't pass memory limit
      if (nextSequence >= allocatedSequenceSize)
      {
        log_i("Increasing sequence size from %d to %d\r\n", allocatedSequenceSize, allocatedSequenceSize * 2);
        allocateSequence(allocatedSequenceSize * 2);
      }
    }
  }

  // Resolve any unresolved gotos. Happens when the label is on
  // a line with a higher line number than the goto statement.
  for (i = 0; i < numUnresolvedGotos; i++)
  {
    int index = findLabel(&(Labels[0]), numLabels, UnresolvedGotos[i].pLabel);
    if (index == -1)
    {
      log_i("Unable to find label '%s' used in goto statement. Aborting...\r\n", UnresolvedGotos[i].pLabel);
      return -4;
    }
    else
    {
      // Update the goto element with the correct index of the label
      Sequence[UnresolvedGotos[i].index].Info = index;
    }
  }

  if (nextSequence == 0)
  {
    log_i("Found no sequence. Aborting...\r\n");
    return -5;
  }
  return 0;
}

static void
doTransition(void)
{
  *gpShouldRender = TRUE;
}

static uint16_t
loadImage(char* pFileName, int slot)
{
  uint16_t result = 0;
  ImageError_t imgResult;
  uint16_t w;
  uint16_t h;
  uint8_t* pTmp;
  uint32_t time = msClock;

  if (PreparedImages[slot] != NULL)
  {
    lpc_free(PreparedImages[slot]);
    PreparedImages[slot] = NULL;
  }

  pTmp = NULL;
  imgResult = decodeImageFile(FATFS_OVERRIDES_ROMFS, pFileName, &pTmp, &w, &h);
  if (imgResult == IMAGE_ERROR_OK)
  {
    PreparedImages[slot] = (uint16_t*)pTmp;
  }
  else
  {
    log_i("[%s:%d] Failed to load image %s into slot %d, result=%d\r\n", __FILE__, __LINE__, pFileName, slot, imgResult);
    result = SLIDE_ERR_RUNTIME;
  }

  time = msClock - time;
  log_i("It took %d ms to load image\r\n", time);

  return result;
}


/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

tSlideError
slideInit(char* pScriptFile, Bool* pAbort, Bool* pShouldRender, uint32_t* pDelayFactor)
{
  uint8_t* pBuf = NULL;
  uint32_t fileSize = 0;
  tFsFileSystem fsType;
  tSlideError result = SLIDE_OK;

  // save references to flags
  gpAbort = pAbort;
  gpShouldRender = pShouldRender;

  // Set the default delay factor
  *pDelayFactor = 1;

  // make sure the script starts to run on the first line
  seqIndex = 0;

  // reset time variables
  targetTime = 0;

  do
  {
    if (initTransitions() != 0)
    {
      result = SLIDE_ERR_MALLOC;
      break;
    }
    if (ImageBackBuffer == NULL)
    {
      // Back buffer will be able to hold two images
      ImageBackBuffer = (TImage) lpc_new(SLIDE_WINDOW_NUM_BYTES*2);
      if (ImageBackBuffer == NULL)
      {
        result = SLIDE_ERR_MALLOC;
        break;
      }
    }

    // Read the contents of the file into a newly allocated buffer
    if (transfs_loadFileIntoNewBuff(pScriptFile, FATFS_OVERRIDES_ROMFS, &pBuf, &fileSize, &fsType) != TFS_OK)
    {
      result = SLIDE_ERR_NO_SUCH_SCRIPT;
      break;
    }

    log_i("Parsing buffer...\r\n");

    // Parse buffer to create the script sequence
    if (parseScript(pBuf, pDelayFactor) != 0)
    {
      result = SLIDE_ERR_INVALID_SCRIPT;
      break;
    }

  } while (0);

  // Release file buffer (as it is no longer needed)
  if (pBuf != NULL)
  {
    lpc_free(pBuf);
    pBuf = NULL;
  }

  if (result != SLIDE_OK)
  {
    if (ImageBackBuffer != NULL)
    {
      lpc_free(ImageBackBuffer);
      ImageBackBuffer = NULL;
    }
  }

  return result;
}

tSlideError
slideRunSlideshow()
{
  uint32_t lastTime = msClock;

  while (!(*gpAbort))
  {
    switch (Sequence[seqIndex].Cmd)
    {
    case COMMAND_CLEAR:
      C_(log_i("[%d] Clear screen\r\n", seqIndex));
      prepareTransition(CLEAR_SCREEN, NULL, NULL);
      doTransition();
      CurrentSlot = NO_SLOT;
      seqIndex++;
      return SLIDE_OK; // time to render

    case COMMAND_GOTO:
      C_(log_i("[%d] Goto %d\r\n", seqIndex, Sequence[seqIndex].Info));
      seqIndex = Sequence[seqIndex].Info;
      break;

    case COMMAND_LOAD_IMG:
      C_(log_i("[%d] Load '%s' into slot %d\r\n", seqIndex, Sequence[seqIndex].pFileName, Sequence[seqIndex].Info));
      if (loadImage(Sequence[seqIndex].pFileName, Sequence[seqIndex].Info) != 0)
      {
        log_i("Failed to load image. Aborting...\r\n");
        return SLIDE_ERR_RUNTIME;
      }
      seqIndex++;
      break;

    case COMMAND_SHOW:
      C_(log_i("[%d] Show image %d with transition %d\r\n", seqIndex, Sequence[seqIndex].Info, Sequence[seqIndex].Transition));
      prepareTransition(Sequence[seqIndex].Transition,
                        (CurrentSlot==NO_SLOT ? NULL : PreparedImages[CurrentSlot]),
                        PreparedImages[Sequence[seqIndex].Info]);
      doTransition();
      CurrentSlot = Sequence[seqIndex].Info;
      seqIndex++;
      return SLIDE_OK; // time to render

    case COMMAND_WAIT:
      if (targetTime == 0)
      {
        // first time in wait state
        targetTime = lastTime + Sequence[seqIndex].Info;
      }

      if (msClock >= targetTime)
      {
        // passed our time => time to continue
        targetTime = 0;
        seqIndex++;
        lastTime = msClock;
      }
      return SLIDE_OK; // must let time pass

    default:
      log_i("Found unknown command: %d\r\n", Sequence[seqIndex].Cmd);
      return SLIDE_ERR_INVALID_SCRIPT;
    }
  }

  if (*gpAbort)
  {
    return SLIDE_ERR_USER_ABORT;
  }
  else
  {
    return SLIDE_ERR_SCRIPT_END;
  }
}

tSlideError
slideRender(int32_t myLcdHnd, uint32_t tick)
{
  if (renderTransition(myLcdHnd))
  {
    // When done with the rendering we should notify slideshow loop to continue
    *gpShouldRender = FALSE;
  }

  return SLIDE_OK;
}

void
slideCleanup(void)
{
  int slot;
  freeSequence();

  if (ImageBackBuffer != NULL)
  {
    lpc_free(ImageBackBuffer);
    ImageBackBuffer = NULL;
  }

  for (slot = 0; slot < TRANSITIONS_NUM_PREPARED_IMAGES; slot++)
  {
    if (PreparedImages[slot] != NULL)
    {
      lpc_free(PreparedImages[slot]);
      PreparedImages[slot] = NULL;
    }
  }

  cleanupTransitions();
}
