/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "application.h"
#include "slideshow.h"
#include <string.h>
#include "globals.h"
#include "log.h"
#include "lpc177x_8x_timer.h"
#include "trans_fs.h"
#include "lpc_swim_font.h"
#include "swim_extensions.h"
#include "pca9532.h"
#include "joystick.h"
#include "lpc_heap.h"
#include "rom_fs.h"
#include "ff.h"
#include "trans_fs.h"
#include "random.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define MAX_SLIDESHOWS_TO_FIND  15

// Delay (in ms) between checking the joystick value
#define JOYSTICK_POLL_DELAY 100

/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static Bool gAbortSlideshow = FALSE;
static Bool gShouldRender = FALSE;
static uint32_t gReachedTimeout = 0;
static uint32_t initialDelay = 0;

static SWIM_WINDOW_T MainWin;

static BackBuffer_t myLcdHnd1 = BACK_BUFFER_INVALID;
static BackBuffer_t myLcdHnd2 = BACK_BUFFER_INVALID;

static uint8_t   activeArea;

/******************************************************************************
 * Local functions
 *****************************************************************************/

static Bool
printError(tSlideError errorCode)
{
  Bool waitForExitButton = FALSE;
  char buff[40];
  buff[0] = '\0';

  switch (errorCode)
  {
  case SLIDE_ERR_NO_SUCH_SCRIPT:
    sprintf(buff, FATFS_PATH_PREFIX "/slides/%dx%d/", GlobalParameters.display.width, GlobalParameters.display.height);
    swimex_put_text_centered(&MainWin, 50, "Unable to load presentation!");
    swimex_put_text_centered(&MainWin, 100, "Store presentation on");
    swimex_put_text_centered(&MainWin, 120, "a memory card in");
    swimex_put_text_centered(&MainWin, 200, buff);
    waitForExitButton = TRUE;
    log_e("SLIDESHOW: script does not exist\r\n");
    break;
  case SLIDE_ERR_INVALID_SCRIPT:
    swimex_put_text_centered(&MainWin, 50, "Unable to load presentation!");
    swimex_put_text_centered(&MainWin, 100, "Script is invalid.");
    waitForExitButton = TRUE;
    log_e("SLIDESHOW: script is invalid\r\n");
    break;
  case SLIDE_ERR_RUNTIME:
    swimex_put_text_centered(&MainWin, 50, "Unable to run presentation!");
    swimex_put_text_centered(&MainWin, 100, "Error during execution.");
    waitForExitButton = TRUE;
    log_e("SLIDESHOW: got a runtime error\r\n");
    break;
  case SLIDE_ERR_USER_ABORT:
    log_i("SLIDESHOW: user aborted script execution\r\n");
    break;
  case SLIDE_ERR_SCRIPT_END:
    log_i("SLIDESHOW: reached end of script\r\n");
    break;
  default:
    swimex_put_text_centered(&MainWin, 50, "Unable to run presentation!");
    swimex_put_text_centered(&MainWin, 100, "Error during execution.");
    waitForExitButton = TRUE;
    log_e("SLIDESHOW: found unknown result code: %d\r\n", errorCode);
    break;
  }
  if (waitForExitButton)
  {
    swimex_put_text_centered(&MainWin, 280, "Press button to exit...");
  }
  return waitForExitButton;
}

static void
start(char* pFileName)
{
  tSlideError result;
  uint32_t delayFactor = 1;
  log_i("slideshowApp: start %s\r\n", pFileName);

  if (!anydisplay_createBackBuffers(&myLcdHnd1, &myLcdHnd2, NULL))
  {
    log_e("Failed to allocate back buffers\r\n");
    return;
  }

  gAbortSlideshow = FALSE;
  gShouldRender = FALSE;
  gReachedTimeout = 0;

  swim_window_open(&MainWin, GlobalParameters.display.width,
      GlobalParameters.display.height, (COLOR_T*)myLcdHnd1, 0, 0,
      (GlobalParameters.display.width - 1),
      (GlobalParameters.display.height - 1), 0, WHITE, BLACK, BLACK);

  // initialize random
  random_init(msClock);

  result = slideInit(pFileName, &gAbortSlideshow, &gShouldRender, &delayFactor);
  if (result != SLIDE_OK)
  {
    swim_clear_screen(&MainWin, WHITE);
    swim_set_pen_color(&MainWin, BLACK);
    swim_set_bkg_color(&MainWin, WHITE);

    if (printError(result))
    {
      anydisplay_activateBuffer(myLcdHnd1);
      while (!gAbortSlideshow)
      {
      }
    }
    return;
  }

  activeArea = 2;
  while(!gAbortSlideshow)
  {
    if (gReachedTimeout > delayFactor)
    {
      if (gShouldRender)
      {
        if (activeArea == 2)
        {
          MainWin.fb = (COLOR_T*)myLcdHnd2;
          slideRender(myLcdHnd2, msClock);
          anydisplay_activateBuffer(myLcdHnd2);
          activeArea = 1;
        }
        else //if (activeArea == 1)
        {
          MainWin.fb = (COLOR_T*)myLcdHnd1;
          slideRender(myLcdHnd1, msClock);
          anydisplay_activateBuffer(myLcdHnd1);
          activeArea = 2;
        }
        gReachedTimeout = 1;
      }
      else
      {
        gReachedTimeout = 1;
        result = slideRunSlideshow();
        if (result != SLIDE_OK)
        {
          printError(result);
          gAbortSlideshow = TRUE;
        }
      }
    }
  }
}

static void
stop(void)
{
  log_i("SLIDESHOW: stop\r\n");
  gAbortSlideshow = TRUE;
  initialDelay = 0;

  slideCleanup();

  anydisplay_freeBackBuffers(&myLcdHnd1, &myLcdHnd2, NULL);
}


static void timeout(void)
{
  if (!gAbortSlideshow && initialDelay>10)
  {
    /* Check SW2 button which causes Exit */
    if (pca9532_getLedState(FALSE) & KEY1) {
      gAbortSlideshow = TRUE;
    }
  }
  initialDelay++;

  gReachedTimeout++;
}

static int copyIfValid(char** ppOut, char* pFileName)
{
  int len = strlen(pFileName);
  *ppOut = NULL;
  if (len > 4 && (pFileName[len-4]=='.') &&
      (pFileName[len-3]=='T' || pFileName[len-3]=='t') &&
      (pFileName[len-2]=='X' || pFileName[len-2]=='x') &&
      (pFileName[len-1]=='T' || pFileName[len-1]=='t'))
  {
    // Found a slide show for the correct display size
    *ppOut = lpc_new(len+1);
    if (*ppOut == NULL)
    {
      // out of memory
      return -1;
    }
    strcpy(*ppOut, pFileName);
  }
  return 0;
}

static char* selectSlideShow(void)
{
  char pDirName[40];
  char* Shows[MAX_SLIDESHOWS_TO_FIND];
  int found = 0;
  int result = 0;
  char* chosen = NULL;
  int i;

  // list files in romfs
  if (result == 0)
  {
    Dir_t *pDir;
    DirEnt_t *pDirEnt;
    pDirName[0] = 0;
    sprintf(pDirName, "/slides/%dx%d", GlobalParameters.display.width, GlobalParameters.display.height);
    pDir = fsOpenDir((uint8_t *)pDirName);
    if (pDir != NULL)
    {
      pDirEnt = fsReadDir(pDir);
      while (pDirEnt != NULL && found<MAX_SLIDESHOWS_TO_FIND)
      {
        if (!fsIsEntryDir(pDirEnt))
        {
          if (copyIfValid(&(Shows[found]), (char*)fsDirEntName(pDirEnt)) != 0)
          {
            // out of memory
            result = -1;
            break;
          }
          if (Shows[found] != NULL) {
            found++;
          }
        }
        pDirEnt = fsReadDir(pDir);
      }
    }
  }

  // list files on SD card
  if (result == 0 && transfs_initializeFATFS()==TFS_OK)
  {
    DIR dir;
    FILINFO info;
    FRESULT fres;
    pDirName[0] = 0;
    sprintf(pDirName, FATFS_PATH_PREFIX "/slides/%dx%d", GlobalParameters.display.width, GlobalParameters.display.height);
    fres = f_opendir(&dir, pDirName);
    if (fres == FR_OK)
    {
      fres = f_readdir(&dir, &info);
      while ((fres == FR_OK) && (info.fname[0] != 0) && (found < MAX_SLIDESHOWS_TO_FIND))
      {
        if ((info.fattrib & AM_DIR) == 0)
        {
          if (copyIfValid(&(Shows[found]), info.fname) != 0)
          {
            // out of memory
            result = -1;
            break;
          }
          if (Shows[found] != NULL) {
            found++;
          }
        }

        fres = f_readdir(&dir, &info);
      }
    }
  }

  // create window and show list
  if (result == 0 && found > 1)
  {
    int joyval;
    int selected = 0;
    int itemHeight = 14;
    int w = 160;
    int h = itemHeight * MIN(MAX_SLIDESHOWS_TO_FIND, found+2);
    int x = (GlobalParameters.display.width - w) / 2;
    int y = (GlobalParameters.display.height - h) / 2;
    swim_window_open(&MainWin,
        GlobalParameters.display.width, GlobalParameters.display.height,
        (COLOR_T*)GlobalParameters.display.framebuffer,
        x, y, x+w-1, y+h-1, 1, BLACK, WHITE, WHITE);
    swim_set_font_trasparency(&MainWin, 0);

    // If the user starts the slideshow with the help of the joystick
    // and keeps the joystick pressed until the file system has been
    // read then the first slideshow will automatically be started,
    // which is not the desired behaviour. Make sure that the joystick
    // is not pressed.
    while ((joyval = joystick_read()) != 0)
    {
      TIM_Waitms(50);
    }

    do
    {
      for (i = 0; i < found; i++)
      {
        swim_set_fill_color(&MainWin, (i==selected ? BLUE : WHITE));
        swim_set_pen_color(&MainWin, (i==selected ? BLUE : WHITE));
        swim_put_box(&MainWin, 1, 1+i*itemHeight, w-4, (i+1)*itemHeight);

        swim_set_pen_color(&MainWin, (i==selected ? WHITE : BLUE));
        swim_put_text_xy(&MainWin, Shows[i], 2, 1+i*itemHeight);
      }

      while ((joyval = joystick_read()) == 0)
      {
        TIM_Waitms(50);
      }
      if (joyval & JOYSTICK_CENTER)
      {
        // user made a selection
        pDirName[0] = 0;
        sprintf(pDirName, "/slides/%dx%d/%s", GlobalParameters.display.width, GlobalParameters.display.height, Shows[selected]);
        chosen = lpc_new(strlen(pDirName)+1);
        if (chosen == NULL) {
          result = -3;
          break;
        }
        strcpy(chosen, pDirName);
        break;
      }
      else if (joyval & JOYSTICK_UP)
      {
        selected = (selected - 1 + found) % found;
      }
      else if (joyval & JOYSTICK_DOWN)
      {
        selected = (selected + 1) % found;
      }
      else
      {
        // user pressed left or right which means exit
        result = -2;
        break;
      }

      // Delay so that the scrolling is not too fast
      TIM_Waitms(200);
    } while(TRUE);
  }

  if (result == 0 && found == 1)
  {
    // user did not have to make a selection as there is only one alternative
    pDirName[0] = 0;
    sprintf(pDirName, "/slides/%dx%d/%s", GlobalParameters.display.width, GlobalParameters.display.height, Shows[0]);
    chosen = lpc_new(strlen(pDirName)+1);
    if (chosen == NULL) {
      result = -4;
    } else {
      strcpy(chosen, pDirName);
    }
  }

  // cleanup all slideshow file names
  for (i = 0; i < found; i++)
  {
    lpc_free(Shows[i]);
  }
  return chosen;
}

static void start1(void) {
  char* pSelectedFile = selectSlideShow();
  if (pSelectedFile != NULL)
  {
    start(pSelectedFile);
    lpc_free(pSelectedFile);
  }
}


/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

static AppCallback_t slideshowAppCallback = {
    &start1,
    &stop,
    &timeout,
    JOYSTICK_POLL_DELAY,
    FALSE,
};

AppCallback_t* slideshowAppInit(void)
{
  return &slideshowAppCallback;
}
