/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "transitions.h"
#include "slideshow.h"
#include <string.h>
#include <stdlib.h>
#include "log.h"
#include "lpc_heap.h"
#include "random.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

// Configuration of transition left-right
#define TRANSITION_LEFT_RIGHT_PIXELS_TO_SKIP  20
#define TRANSITION_LEFT_RIGHT_COPY_STMT(offset) \
  memcpy(ImageBackBuffer + (offset),            \
         NewImage + (offset),                   \
         TRANSITION_LEFT_RIGHT_PIXELS_TO_SKIP*sizeof(uint16_t))
#define TRANSITION_LEFT_RIGHT_NUM_CALLS   ((SLIDE_WINDOW_WIDTH - TRANSITION_LEFT_RIGHT_PIXELS_TO_SKIP) / TRANSITION_LEFT_RIGHT_PIXELS_TO_SKIP)


// Configuration of transition top-down
#define TRANSITION_TOP_DOWN_LINE_SKIP    20
#define TRANSITION_TOP_DOWN_NUM_CALLS   (((SLIDE_WINDOW_HEIGHT - 1) / TRANSITION_TOP_DOWN_LINE_SKIP) + 1)


// Configuration of blinds transition
#define TRANSITION_BLINDS_LINES_PER_BLOCK  10
#define TRANSITION_BLINDS_BEAM_COLOR       0xFFFF
#define TRANSITION_BLINDS_BACK_COLOR       0x0000
#define TRANSITION_BLINDS_BLOCK_NUM_PIXELS  TRANSITION_BLINDS_LINES_PER_BLOCK * SLIDE_WINDOW_WIDTH
#define TRANSITION_BLINDS_BLOCK_NUM_BYTES   TRANSITION_BLINDS_BLOCK_NUM_PIXELS * sizeof(uint16_t)
#define TRANSITION_BLINDS_HALFWAY     ((SLIDE_WINDOW_NUM_PIXELS) / (TRANSITION_BLINDS_BLOCK_NUM_PIXELS))
#define TRANSITION_BLINDS_NUM_CALLS   TRANSITION_BLINDS_HALFWAY * 2


// Configuration of explode transition
#define TRANSITION_EXPLODE_NUM_CALLS     10 // no idea yet
#define TRANSITION_EXPLODE_CHUNK_WIDTH   40
#define TRANSITION_EXPLODE_CHUNK_HEIGHT  40
#define TRANSITION_EXPLODE_NUM_CHUNKS_W  (SLIDE_WINDOW_WIDTH / TRANSITION_EXPLODE_CHUNK_WIDTH)
#define TRANSITION_EXPLODE_NUM_CHUNKS_H  (SLIDE_WINDOW_HEIGHT / TRANSITION_EXPLODE_CHUNK_HEIGHT)
#define TRANSITION_EXPLODE_NUM_CHUNKS    (TRANSITION_EXPLODE_NUM_CHUNKS_W * TRANSITION_EXPLODE_NUM_CHUNKS_H)
#define TRANSITION_EXPLODE_MAX(a,b)      (((a) > (b)) ? (a) : (b))
#define TRANSITION_EXPLODE_MIN(a,b)      (((a) < (b)) ? (a) : (b))
#define TRANSITION_EXPLODE_CHUNK_INSIDE(chunk) ( \
    (((chunk).currX + (chunk).chunkWidth) > 0)  && \
    ((chunk).currX < SLIDE_WINDOW_WIDTH)    && \
    (((chunk).currY + (chunk).chunkHeight) > 0) && \
    ((chunk).currY < SLIDE_WINDOW_HEIGHT) )
#define TRANSITION_EXPLODE_BEZIER(p0, p1, p2, p3, t) \
    (( (100-(t)) * (100-(t)) * (100-(t)) * (p0)   \
        + 3 * (t) * (100-(t)) * (100-(t)) * (p1)  \
        + 3 * (t) * (t) * (100-(t)) * (p2)        \
        + (t) * (t) * (t) * (p3) ) / 1000000 )
#define TRANSITION_EXPLODE_BEZIER_X(chunk, t) \
    TRANSITION_EXPLODE_BEZIER((chunk).chunkX, (chunk).xpoints[0], (chunk).xpoints[1], (chunk).xpoints[2], t)
#define TRANSITION_EXPLODE_BEZIER_Y(chunk, t) \
    TRANSITION_EXPLODE_BEZIER((chunk).chunkY, (chunk).ypoints[0], (chunk).ypoints[1], (chunk).ypoints[2], t)

typedef struct
{
  TTransitionInfo* pInfo;
  uint16_t numMadeRenderings;
  TImage FromImage;
  TImage ToImage;
} TCurrentTransition;

typedef struct
{
  int chunkX;
  int chunkY;
  int chunkWidth;
  int chunkHeight;
  int currX;
  int currY;
  int xpoints[3];
  int ypoints[3];
  int numCalls;
} TExplosion;

/******************************************************************************
 * External variables
 *****************************************************************************/

/******************************************************************************
 * Local function prototypes
 *****************************************************************************/

// transition functions
static uint16_t showImageNoTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo);
static uint16_t showImageLeftRightTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo);
static uint16_t showImageTopDownTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo);
static uint16_t showImageWindowBlinds(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo);
static uint16_t showImageExplodeTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo);
static uint16_t clearScreen(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo);

// Calls to prepare for a transition - not for any actual drawing
static uint16_t prepareLeftRightTransition(TImage CurrentImage, TImage NewImage);
static uint16_t prepareTopDownTransition(TImage CurrentImage, TImage NewImage);
static uint16_t prepareWindowBlinds(TImage CurrentImage, TImage NewImage);
static uint16_t prepareExplodeVerticalTransition(TImage CurrentImage, TImage NewImage);
static uint16_t prepareExplodeHorizontalTransition(TImage CurrentImage, TImage NewImage);

/******************************************************************************
 * Public variables
 *****************************************************************************/

TTransitionInfo Transitions[TRANSITIONS_NUM_AVAILABLE+1] = {
  {"none",       TRANSITION_NONE,       0,                             &showImageNoTransition,        NULL},
  {"left-right", TRANSITION_LEFT_RIGHT, 0, /* set later */             &showImageLeftRightTransition, &prepareLeftRightTransition},
  {"top-down",   TRANSITION_TOP_DOWN,   0, /* set later */             &showImageTopDownTransition,   &prepareTopDownTransition},
  {"blinds",     TRANSITION_BLINDS,     0, /* set later */             &showImageWindowBlinds,        &prepareWindowBlinds},
  {"explode_v",  TRANSITION_EXPLODE_V,  TRANSITION_EXPLODE_NUM_CALLS,  &showImageExplodeTransition,   &prepareExplodeVerticalTransition},
  {"explode_h",  TRANSITION_EXPLODE_H,  TRANSITION_EXPLODE_NUM_CALLS,  &showImageExplodeTransition,   &prepareExplodeHorizontalTransition},
  {"????",       CLEAR_SCREEN,          0,                             &clearScreen,                  NULL},
  {NULL,         TRANSITION_NONE,       0,                             NULL,                          NULL} // marker
};


/******************************************************************************
 * Local variables
 *****************************************************************************/

static TCurrentTransition CurrentTransition = { NULL, 0, NULL, NULL };

// Variables used by the "Blinds" transition
static TImage blindsBeamBlock;
static TImage blindsBkgBlock;

// Variables used by the "Explode" transition
static TExplosion* ExplodingChunks = NULL;


/******************************************************************************
 * Local functions
 *****************************************************************************/

static uint16_t
showImageNoTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo)
{
  //log_i("Transition[None] callNo = %u\r\n", callNo);
  memcpy((void*)myLcdHnd, (uint8_t*)(NewImage), SLIDE_WINDOW_NUM_BYTES);
  return 0;
}

static uint16_t
prepareLeftRightTransition(TImage CurrentImage, TImage NewImage)
{
  // Copy the entire old image
  memcpy(ImageBackBuffer, CurrentImage, SLIDE_WINDOW_NUM_BYTES);
  return 0;
}

static uint16_t
showImageLeftRightTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo)
{
  int x, y, off;

  if (callNo < TRANSITION_LEFT_RIGHT_NUM_CALLS)
  {
    x = callNo * TRANSITION_LEFT_RIGHT_PIXELS_TO_SKIP;
    off = 0;
    for (y = 0; y < SLIDE_WINDOW_HEIGHT; y++)
    {
      TRANSITION_LEFT_RIGHT_COPY_STMT(off+x);
      off += SLIDE_WINDOW_WIDTH;
    }

    // Show the updated image
    memcpy((void*)myLcdHnd, ImageBackBuffer, SLIDE_WINDOW_NUM_BYTES);
  }
  else
  {
    // show final image
    memcpy((void*)myLcdHnd, NewImage, SLIDE_WINDOW_NUM_BYTES);
  }

  return 0;
}

static uint16_t
prepareTopDownTransition(TImage CurrentImage, TImage NewImage)
{
  // Create a buffer with the two images after each other
  memcpy(ImageBackBuffer, CurrentImage, SLIDE_WINDOW_NUM_BYTES);
  memcpy(ImageBackBuffer+SLIDE_WINDOW_NUM_PIXELS, NewImage, SLIDE_WINDOW_NUM_BYTES);
  return 0;
}

static uint16_t
showImageTopDownTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo)
{
  int i;

  if (callNo < TRANSITION_TOP_DOWN_NUM_CALLS)
  {
    i = 1 + callNo * TRANSITION_TOP_DOWN_LINE_SKIP;
    memcpy((void*)myLcdHnd, ImageBackBuffer+i*SLIDE_WINDOW_WIDTH, SLIDE_WINDOW_NUM_BYTES);
  }
  else
  {
    // show final image
    memcpy((void*)myLcdHnd, NewImage, SLIDE_WINDOW_NUM_BYTES);
  }

  return 0;
}

static uint16_t
prepareWindowBlinds(TImage CurrentImage, TImage NewImage)
{
  blindsBeamBlock = ImageBackBuffer + SLIDE_WINDOW_NUM_PIXELS;
  blindsBkgBlock = blindsBeamBlock + TRANSITION_BLINDS_BLOCK_NUM_PIXELS;

  // Create a buffer with the old image
  memcpy(ImageBackBuffer, CurrentImage, SLIDE_WINDOW_NUM_BYTES);

  // Create the two coloured blocks
  memset(blindsBeamBlock, TRANSITION_BLINDS_BEAM_COLOR, TRANSITION_BLINDS_BLOCK_NUM_BYTES);
  memset(blindsBkgBlock, TRANSITION_BLINDS_BACK_COLOR, TRANSITION_BLINDS_BLOCK_NUM_BYTES);

  return 0;
}

static uint16_t
showImageWindowBlinds(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo)
{
  int i;

  if (callNo < TRANSITION_BLINDS_HALFWAY)
  {
    // Draw the moving beam, erasing the old image
    i = callNo * TRANSITION_BLINDS_BLOCK_NUM_PIXELS;
    memcpy(ImageBackBuffer + i, blindsBeamBlock, TRANSITION_BLINDS_BLOCK_NUM_BYTES);

    // Fill upp behind the beam with background color
    if (i > 0)
    {
      memcpy(ImageBackBuffer + i - TRANSITION_BLINDS_BLOCK_NUM_PIXELS, blindsBkgBlock, TRANSITION_BLINDS_BLOCK_NUM_BYTES);
    }

    // Show the updated image
    memcpy((void*)myLcdHnd, ImageBackBuffer, SLIDE_WINDOW_NUM_BYTES);
  }
  else if (callNo < TRANSITION_BLINDS_NUM_CALLS)
  {
    if (callNo == TRANSITION_BLINDS_HALFWAY)
    {
      // Fill last part (after the bar) with background color
      i = callNo * TRANSITION_BLINDS_BLOCK_NUM_PIXELS;
      memcpy(ImageBackBuffer + i - TRANSITION_BLINDS_BLOCK_NUM_PIXELS, blindsBkgBlock, TRANSITION_BLINDS_BLOCK_NUM_BYTES);
    }

    // Draw the moving beam, erasing the old image
    i = (callNo % TRANSITION_BLINDS_HALFWAY) * TRANSITION_BLINDS_BLOCK_NUM_PIXELS;
    memcpy(ImageBackBuffer + i, blindsBeamBlock, TRANSITION_BLINDS_BLOCK_NUM_BYTES);

    // Fill upp behind the beam with the new image
    if (i > 0)
    {
      memcpy(ImageBackBuffer + i - TRANSITION_BLINDS_BLOCK_NUM_PIXELS, NewImage + i - TRANSITION_BLINDS_BLOCK_NUM_PIXELS, TRANSITION_BLINDS_BLOCK_NUM_BYTES);
    }

    // Show the updated image
    memcpy((void*)myLcdHnd, ImageBackBuffer, SLIDE_WINDOW_NUM_BYTES);
  }
  else
  {
    // show final image
    memcpy((void*)myLcdHnd, NewImage, SLIDE_WINDOW_NUM_BYTES);
  }

  return 0;
}

static uint16_t
prepareExplodeVerticalTransition(TImage CurrentImage, TImage NewImage)
{
  int i;

  // only setup do no actual transformation
  for (i = 0; i < TRANSITION_EXPLODE_NUM_CHUNKS; i++)
  {
    ExplodingChunks[i].chunkX = (i % TRANSITION_EXPLODE_NUM_CHUNKS_W) * TRANSITION_EXPLODE_CHUNK_WIDTH;
    ExplodingChunks[i].chunkY = (i / TRANSITION_EXPLODE_NUM_CHUNKS_W) * TRANSITION_EXPLODE_CHUNK_HEIGHT;
    ExplodingChunks[i].chunkWidth = TRANSITION_EXPLODE_CHUNK_WIDTH;
    ExplodingChunks[i].chunkHeight = TRANSITION_EXPLODE_CHUNK_HEIGHT;

    ExplodingChunks[i].xpoints[0] = random() % SLIDE_WINDOW_WIDTH;
    ExplodingChunks[i].xpoints[1] = SLIDE_WINDOW_WIDTH;
    ExplodingChunks[i].xpoints[2] = random() % SLIDE_WINDOW_WIDTH;

    ExplodingChunks[i].ypoints[0] = 0;
    ExplodingChunks[i].ypoints[1] = random() % SLIDE_WINDOW_HEIGHT;
    ExplodingChunks[i].ypoints[2] = SLIDE_WINDOW_HEIGHT;

    ExplodingChunks[i].currX = TRANSITION_EXPLODE_BEZIER_X(ExplodingChunks[i], 0);
    ExplodingChunks[i].currY = TRANSITION_EXPLODE_BEZIER_Y(ExplodingChunks[i], 0);

    ExplodingChunks[i].numCalls = 1;
  }
  return 0;
}

static uint16_t
prepareExplodeHorizontalTransition(TImage CurrentImage, TImage NewImage)
{
  int i;

  // only setup do no actual transformation
  for (i = 0; i < TRANSITION_EXPLODE_NUM_CHUNKS; i++)
  {
    ExplodingChunks[i].chunkX = (i % TRANSITION_EXPLODE_NUM_CHUNKS_W) * TRANSITION_EXPLODE_CHUNK_WIDTH;
    ExplodingChunks[i].chunkY = (i / TRANSITION_EXPLODE_NUM_CHUNKS_W) * TRANSITION_EXPLODE_CHUNK_HEIGHT;
    ExplodingChunks[i].chunkWidth = TRANSITION_EXPLODE_CHUNK_WIDTH;
    ExplodingChunks[i].chunkHeight = TRANSITION_EXPLODE_CHUNK_HEIGHT;

    ExplodingChunks[i].xpoints[0] = SLIDE_WINDOW_WIDTH;
    ExplodingChunks[i].xpoints[1] = random() % SLIDE_WINDOW_WIDTH;
    ExplodingChunks[i].xpoints[2] = 0;

    ExplodingChunks[i].ypoints[0] = random() % SLIDE_WINDOW_HEIGHT;
    ExplodingChunks[i].ypoints[1] = SLIDE_WINDOW_HEIGHT;
    ExplodingChunks[i].ypoints[2] = random() % SLIDE_WINDOW_HEIGHT;

    ExplodingChunks[i].currX = TRANSITION_EXPLODE_BEZIER_X(ExplodingChunks[i], 0);
    ExplodingChunks[i].currY = TRANSITION_EXPLODE_BEZIER_Y(ExplodingChunks[i], 0);

    ExplodingChunks[i].numCalls = 1;
  }
  return 0;
}

static uint16_t
showImageExplodeTransition(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo)
{
  int i;
  Bool moreChunksToDraw = FALSE;
  for (i = 0; i < TRANSITION_EXPLODE_NUM_CHUNKS; i++)
  {
    if (ExplodingChunks[i].numCalls > 0)
    {
      moreChunksToDraw = TRUE;
      break;
    }
  }

  if (moreChunksToDraw && (callNo <= 9))
  {
    // start with the back image
    memcpy(ImageBackBuffer, NewImage, SLIDE_WINDOW_NUM_BYTES);

    for (i = 0; i < TRANSITION_EXPLODE_NUM_CHUNKS; i++)
    {
      if (ExplodingChunks[i].numCalls > 0)
      {
        int line;
        int numLines = TRANSITION_EXPLODE_MIN(SLIDE_WINDOW_HEIGHT - ExplodingChunks[i].currY, ExplodingChunks[i].chunkHeight);
        int xoff = TRANSITION_EXPLODE_MAX(0, -ExplodingChunks[i].currX);
        int width = TRANSITION_EXPLODE_MIN(SLIDE_WINDOW_WIDTH - (ExplodingChunks[i].currX + xoff), ExplodingChunks[i].chunkWidth);

        // Draw chunk
        for (line = 0; line < numLines; line++)
        {
          memcpy(ImageBackBuffer + (ExplodingChunks[i].currX + xoff) + ((ExplodingChunks[i].currY + line) * SLIDE_WINDOW_WIDTH),
                 CurrentImage + (ExplodingChunks[i].chunkX + xoff) + ((ExplodingChunks[i].chunkY + line) * SLIDE_WINDOW_WIDTH),
                 (width - xoff)*sizeof(uint16_t));
        }

        // Update chunks next position
        ExplodingChunks[i].currX = TRANSITION_EXPLODE_BEZIER_X(ExplodingChunks[i], (callNo+1)*10);
        ExplodingChunks[i].currY = TRANSITION_EXPLODE_BEZIER_Y(ExplodingChunks[i], (callNo+1)*10);

        if (TRANSITION_EXPLODE_CHUNK_INSIDE(ExplodingChunks[i]))
        {
          ExplodingChunks[i].numCalls = 1; // simple interpretation. At least one pixel inside => redraw
        }
        else
        {
          ExplodingChunks[i].numCalls = 0; // no pixels inside visible area => stop drawing
        }
      }
    }

    // Show the updated image
    memcpy((void*)myLcdHnd, ImageBackBuffer, SLIDE_WINDOW_NUM_BYTES);
  }
  else
  {
    // show final image
    memcpy((void*)myLcdHnd, NewImage, SLIDE_WINDOW_NUM_BYTES);
  }

  return 0;
}

static uint16_t
clearScreen(int32_t myLcdHnd, TImage CurrentImage, TImage NewImage, uint16_t callNo)
{
//  log_i("Transition[Clear] callNo = %u\r\n", callNo);
  memset((void*)myLcdHnd, 0, SLIDE_WINDOW_NUM_BYTES);
  return 0;
}



/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

// sets next transition to be rendered
int prepareTransition(TTransition t, TImage CurrentImage, TImage NewImage)
{
  int i;

  for (i = 0; i < TRANSITIONS_NUM_AVAILABLE; i++)
  {
    if (Transitions[i].Transition == t)
    {
      CurrentTransition.pInfo = &(Transitions[i]);
      CurrentTransition.numMadeRenderings = 0;
      CurrentTransition.FromImage = CurrentImage;
      CurrentTransition.ToImage = NewImage;

      // Do an initialization call to allow lengthy initialization to be made
      if (CurrentTransition.pInfo->prepare != NULL)
      {
        if (CurrentTransition.pInfo->prepare(CurrentTransition.FromImage,
                                             CurrentTransition.ToImage))
        {
          log_i("prepareTransition: ERROR: Init failed for transition %d\r\n", t);
          return -2;
        }
      }

      return 0;
    }
  }
  log_i("prepareTransition: ERROR: no valid transition for %d\r\n", t);
  return -1;
}

// renders next part of current transition, returns true if last step was rendered
Bool renderTransition(int32_t myLcdHnd)
{
  if (CurrentTransition.pInfo->showImage(myLcdHnd,
                                         CurrentTransition.FromImage,
                                         CurrentTransition.ToImage,
                                         CurrentTransition.numMadeRenderings))
  {
    log_i("Failed to show image. Aborting...\r\n");
    return TRUE;
  }
  CurrentTransition.numMadeRenderings++;
  if (CurrentTransition.numMadeRenderings > CurrentTransition.pInfo->NumRenderCallsNeeded)
  {
    // Have rendered the last frame
    return TRUE;
  }

  // Need to continue rendering
  return FALSE;
}

int initTransitions(void)
{
  int i;
  ExplodingChunks = lpc_new(sizeof(TExplosion)*TRANSITION_EXPLODE_NUM_CHUNKS);
  if (ExplodingChunks == NULL)
  {
    return -1;
  }

  // Initialize transitions with the non-constant data
  for (i = 0; i <= TRANSITIONS_NUM_AVAILABLE; i++)
  {
    switch (Transitions[i].Transition)
    {
    case TRANSITION_LEFT_RIGHT:
      Transitions[i].NumRenderCallsNeeded = TRANSITION_LEFT_RIGHT_NUM_CALLS;
      break;
    case TRANSITION_TOP_DOWN:
      Transitions[i].NumRenderCallsNeeded = TRANSITION_TOP_DOWN_NUM_CALLS;
      break;
    case TRANSITION_BLINDS:
      Transitions[i].NumRenderCallsNeeded = TRANSITION_BLINDS_NUM_CALLS;
      break;
    default:
      break;
    }
  }
  return 0;
}

void cleanupTransitions(void)
{
  if (ExplodingChunks != NULL)
  {
    lpc_free(ExplodingChunks);
    ExplodingChunks = NULL;
  }
}

