/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include "lodepng.h"
#include "lpc_heap.h"
#include "image.h"
#include "log.h"
#include "rom_fs.h"
#include "tinyjpeg.h"
#include "bmp.h"


/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

typedef struct
{
  ImageFormat_t type;
  char* pExtension;
} FileTypes_t;

/******************************************************************************
 * Global variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static FileTypes_t FileTypes[] =
{
  {IMAGE_FORMAT_BMP, ".bmp"},
  {IMAGE_FORMAT_PNG, ".png"},
  {IMAGE_FORMAT_JPEG, ".jpg"},
  {IMAGE_FORMAT_JPEG, ".jpeg"},
  {IMAGE_FORMAT_UNKNOWN, NULL},
};

/******************************************************************************
 * Local Functions
 *****************************************************************************/

/*
 * The decoded image produced by LodePNG is in 24 bit format but the display wants it in 16 bit
 * format (565). This function converts the image to 565 format.
 */
/*****************************************************************************
 *
 * Description:
 *    Uses the LodePNG library to decode the pSrc data into a 24bit
 *    image. That image is then converted into the 565 format used by
 *    the display.
 *
 *    Memory will be allocated for the new image and returned in the
 *    ppDest parameter. If the return value is IMAGE_ERROR_OK then it is
 *    the responsibility of the caller to deallocate (using lpc_free()) the
 *    image when it is no longer needed.
 *
 * Params:
 *    [in]  pSrc       - The image file to decode
 *    [in]  srcSize    - The number of bytes in pSrc
 *    [out] ppDest     - The decoded image in 565 format
 *    [out] pWidth     - The width of the decoded image
 *    [out] pHeight    - The height of the decoded image
 *
 * Return:
 *    IMAGE_ERROR_OK if an image was decoded and stored in ppDest,
 *    otherwise one of the IMAGE_ERROR_* error codes.
 *
 ****************************************************************************/
static unsigned decodePNG(const uint8_t* pSrc, uint32_t srcSize, uint8_t** ppDest,
    uint16_t* pWidth, uint16_t* pHeight)
{
  unsigned char* pTmp;
  unsigned width, height;
  unsigned error = LodePNG_decode24(&pTmp, &width, &height, pSrc, srcSize);

  if (error == 0)
  {
    int x, y;
    uint16_t* pConverted;
    uint8_t r;
    uint8_t g;
    uint8_t b;
    int off = 0;

    *pWidth = width;
    *pHeight = height;

    *ppDest = lpc_new(width * height * 2);
    if (*ppDest == NULL)
    {
      log_e("Failed to decode JPEG: Not enough memory to store result\r\n");
      error = IMAGE_ERROR_MALLOC;
    }
    else
    {
      pConverted = (uint16_t*)*ppDest;

      for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++) {
          r = pTmp[off    ];
          g = pTmp[off + 1];
          b = pTmp[off + 2];
          *pConverted = (((unsigned short)r & 0xF8) << 8) |
                         (((unsigned short)g & 0xFC) << 3) |
                         (((unsigned short)b & 0xF8) >> 3);
          pConverted++;
          off += 3;
        }
      }
    }
    lpc_free(pTmp);
  }
  else
  {
    log_e("Failed to decode PNG. Error %d\r\n", error);
  }
  return error;
}


/*****************************************************************************
 *
 * Description:
 *    Uses the TinyJPEG library to decode the pSrc data into a 24bit
 *    image. That image is then converted into the 565 format used by
 *    the display.
 *
 *    Memory will be allocated for the new image and returned in the
 *    ppDest parameter. If the return value is IMAGE_ERROR_OK then it is
 *    the responsibility of the caller to deallocate (using lpc_free()) the
 *    image when it is no longer needed.
 *
 * Params:
 *    [in]  pSrc       - The image file to decode
 *    [in]  srcSize    - The number of bytes in pSrc
 *    [out] ppDest     - The decoded image in 565 format
 *    [out] pWidth     - The width of the decoded image
 *    [out] pHeight    - The height of the decoded image
 *
 * Return:
 *    IMAGE_ERROR_OK if an image was decoded and stored in ppDest,
 *    otherwise one of the IMAGE_ERROR_* error codes.
 *
 ****************************************************************************/
static ImageError_t decodeJpeg(uint8_t* pSrc, uint32_t srcSize, uint8_t** ppDest,
    uint16_t* pWidth, uint16_t* pHeight)
{
  struct jdec_private *jdec;
  unsigned int width, height;
  unsigned char *components[3];

  /* Decompress it */
  jdec = tinyjpeg_init();
  if (jdec == NULL)
  {
    log_e("Failed to decode JPEG: Not enough memory to alloc the structure need for decompressing\r\n");
    return IMAGE_ERROR_MALLOC;
  }

  if (tinyjpeg_parse_header(jdec, pSrc, srcSize)<0)
  {
    log_e("Failed to decode JPEG: %s\r\n", tinyjpeg_get_errorstring(jdec));
    lpc_free(jdec);
    return IMAGE_ERROR_JPEG_DECODE;
  }

  /* Get the size of the image */
  tinyjpeg_get_size(jdec, &width, &height);

  if (tinyjpeg_decode(jdec, TINYJPEG_FMT_RGB24) < 0)
  {
    log_e("Failed to decode JPEG: %s\r\n", tinyjpeg_get_errorstring(jdec));
    tinyjpeg_free(jdec);
    return IMAGE_ERROR_JPEG_DECODE;
  }

  /*
   * Get address for each plane (not only max 3 planes is supported), and
   * depending of the output mode, only some components will be filled
   * RGB: 1 plane, YUV420P: 3 planes, GREY: 1 plane
   */
  tinyjpeg_get_components(jdec, components);

  /* Validate output buffer */
  if (*ppDest == NULL)
  {
    // Must allocate memory
    *ppDest = lpc_new(width * height * 2);
  }
  if (*ppDest == NULL)
  {
    log_e("Failed to decode JPEG: Not enough memory to store result\r\n");
    tinyjpeg_free(jdec);
    return IMAGE_ERROR_MALLOC;
  }

  /* Save it */
  {
    int x, y;
    uint8_t* pTmp = components[0];
    uint16_t* pConverted = (uint16_t*)(*ppDest);
    uint8_t r;
    uint8_t g;
    uint8_t b;
    int off = 0;
    for (y = 0; y < height; y++) {
      for (x = 0; x < width; x++) {
        r = pTmp[off    ];
        g = pTmp[off + 1];
        b = pTmp[off + 2];
        *pConverted = (((unsigned short)r & 0xF8) << 8) |
                       (((unsigned short)g & 0xFC) << 3) |
                       (((unsigned short)b & 0xF8) >> 3);
        pConverted++;
        off += 3;
      }
    }
  }

  *pWidth = (uint16_t)width;
  *pHeight = (uint16_t)height;

  /* Only called this if the buffers were allocated by tinyjpeg_decode() */
  tinyjpeg_free(jdec);
  /* else called just free(jdec); */
  return IMAGE_ERROR_OK;
}


/*****************************************************************************
 *
 * Description:
 *    Looks at the file extension to see which type of file it is.
 *
 *    If pType already is set to a valid file type then that will
 *    be returned.
 *
 * Params:
 *    [in]     pFileName  - The name of the file to decode.
 *    [in/out] pType      - The image type
 *
 * Return:
 *    IMAGE_ERROR_OK if the image type could be found,
 *    otherwise IMAGE_ERROR_WRONG_FORMAT.
 *
 ****************************************************************************/
static ImageError_t decodeImageType(char* pFileName, ImageFormat_t* pType)
{
  if (*pType != IMAGE_FORMAT_UNKNOWN)
  {
    // already a valid image type
    return IMAGE_ERROR_OK;
  }
  else
  {
    *pType = IMAGE_FORMAT_UNKNOWN;
    if (pFileName != NULL)
    {
      int i;
      for (i = 0; FileTypes[i].type != IMAGE_FORMAT_UNKNOWN; i++)
      {
        if (strstr(pFileName, FileTypes[i].pExtension))
        {
          *pType = FileTypes[i].type;
          return IMAGE_ERROR_OK;
        }
      }
    }
    return IMAGE_ERROR_WRONG_FORMAT;
  }
}

/******************************************************************************
 * Public Functions
 *****************************************************************************/


/*****************************************************************************
 *
 * Description:
 *    Loads an image file from the ROM file system and then decodes it
 *    into the 565 format used by the display.
 *
 *    Memory will be allocated for the new image and returned in the
 *    ppDecoded parameter. If the return value is IMAGE_ERROR_OK then it
 *    is the responsibility of the caller to deallocate (using lpc_free())
 *    the image when it is no longer needed.
 *
 * Params:
 *    [in]  format     - The type of image file. If UNKNOWN then the file
 *                       extension will be looked at
 *    [in]  pFileName  - The name of the file to decode.
 *    [out] ppDecoded  - The decoded image in 565 format
 *    [out] pWidth     - The width of the decoded image
 *    [out] pHeight    - The height of the decoded image
 *
 * Return:
 *    IMAGE_ERROR_OK if an image was decoded and stored in ppDecoded,
 *    otherwise one of the IMAGE_ERROR_* error codes.
 *
 ****************************************************************************/
ImageError_t decodeImageFileRomFS(ImageFormat_t format,
   char* pFileName,
   uint8_t** ppDecoded,
   uint16_t* pWidth,
   uint16_t* pHeight)
{
  return decodeImageFile(ONLY_USE_ROMFS, pFileName, ppDecoded, pWidth, pHeight);
}


/*****************************************************************************
 *
 * Description:
 *    Loads an image file from one of the available file systems and then
 *    decodes it into the 565 format used by the display.
 *
 *    The fileLocation parameter decides the search order used when attempting
 *    to locate the image file.
 *
 *    Memory will be allocated for the new image and returned in the
 *    ppDecoded parameter. If the return value is IMAGE_ERROR_OK then it is
 *    the responsibility of the caller to deallocate (using lpc_free()) the
 *    image when it is no longer needed.
 *
 * Params:
 *    [in]  fileLocation  - Which location (FATFS, ROMFS) to start searching in
 *    [in]  pFileName     - The name of the file to decode.
 *    [out] ppDecoded     - The decoded image in 565 format
 *    [out] pWidth        - The width of the decoded image
 *    [out] pHeight       - The height of the decoded image
 *
 * Return:
 *    IMAGE_ERROR_OK if an image was decoded and stored in ppDecoded,
 *    otherwise one of the IMAGE_ERROR_* error codes.
 *
 ****************************************************************************/
ImageError_t decodeImageFile(
    tFsOverride fileLocation,
    char* pFileName,
    uint8_t** ppDecoded,
    uint16_t* pWidth,
    uint16_t* pHeight)
{
  ImageError_t result = IMAGE_ERROR_OK;
  ImageFormat_t format = IMAGE_FORMAT_UNKNOWN;
  tTransFsError fsError;
  tFsFileSystem fsType;
  uint8_t* pData = NULL;
  uint32_t size = 0;
  unsigned error;

  do {
    result = decodeImageType(pFileName, &format);
    if (result != IMAGE_ERROR_OK) {
      break;
    }

    fsError = transfs_loadFileIntoNewBuff(pFileName, fileLocation, &pData, &size, &fsType);
    if (fsError == TFS_ERR_MALLOC_FAILED)
    {
      result = IMAGE_ERROR_MALLOC;
      break;
    }
    if (fsError != TFS_OK)
    {
      result = IMAGE_ERROR_FSOPEN;
      break;
    }
    if (size == 0)
    {
      result = IMAGE_ERROR_FSSIZE_ZERO;
      break;
    }

    if (format == IMAGE_FORMAT_PNG) {
      error = decodePNG(pData, size, ppDecoded, pWidth, pHeight);
      if (error != 0) {
        result = IMAGE_ERROR_PNG_DECODE;
        break;
      }
    } else if (format == IMAGE_FORMAT_JPEG) {
      error = decodeJpeg(pData, size, ppDecoded, pWidth, pHeight);
      if (error != 0) {
        result = IMAGE_ERROR_JPEG_DECODE;
        break;
      }
    } else if (format == IMAGE_FORMAT_BMP) {
      *pWidth = (uint16_t)(((struct BMPHeader *) pData)->width);
      *pHeight = (uint16_t)(((struct BMPHeader *) pData)->height);
      *ppDecoded = lpc_new((*pWidth)*(*pHeight)*3);
      if (*ppDecoded == NULL) {
        result = IMAGE_ERROR_MALLOC;
        break;
      }
      error = BMP_Decode(pData, *ppDecoded, *pWidth, *pHeight, 24);
      if (error != 0) {
        result = IMAGE_ERROR_BMP_DECODE;
        break;
      }
    } else {
      result = IMAGE_ERROR_WRONG_FORMAT;
      break;
    }
  } while(0);

  if (pData != NULL)
  {
    lpc_free(pData);
  }

  return result;
}
