/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "lpc_swim.h"
#include "lpc_heap.h"
#include "swim_extensions.h"
#include "image.h"
#include "globals.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

#define NUM_PROGRESS_STEPS  6

/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static SWIM_WINDOW_T* pProgressWindow = NULL;
static uint8_t* images[NUM_PROGRESS_STEPS] = {NULL, NULL, NULL, NULL, NULL, NULL};
static uint16_t width = 0;
static uint16_t height = 0;

/******************************************************************************
 * Local functions
 *****************************************************************************/


/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/


/*****************************************************************************
 *
 * Description:
 *    Creates a progress window (a white full screen window) and loads the
 *    necessary images but does not show the initial progress bar.
 *
 *    Note that a call to progress_set(0) should be made immediately after
 *    progress_init() to display the progress bar.
 *
 * Params:
 *    None
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void progress_init(void)
{
  ImageError_t result;
  char file[] = {'/','p','r','o','g','r','e','s','s','/','p','0','0','0','.','p','n','g','\0'};
  do {
    pProgressWindow = lpc_new(sizeof(SWIM_WINDOW_T));
    if (pProgressWindow == NULL) {
      break;
    }

    swim_window_open(pProgressWindow,
        GlobalParameters.display.width,
        GlobalParameters.display.height,
        (COLOR_T*)GlobalParameters.display.framebuffer,
        0, 0,
        GlobalParameters.display.width - 1,
        GlobalParameters.display.height - 1,
        0, BLACK, WHITE, WHITE);

    result = decodeImageFileRomFS(IMAGE_FORMAT_PNG, file, &(images[0]), &width, &height);
    if (result != IMAGE_ERROR_OK) {
      break;
    }

    file[12] = '2';
    result = decodeImageFileRomFS(IMAGE_FORMAT_PNG, file, &(images[1]), &width, &height);
    if (result != IMAGE_ERROR_OK) {
      break;
    }

    file[12] = '4';
    result = decodeImageFileRomFS(IMAGE_FORMAT_PNG, file, &(images[2]), &width, &height);
    if (result != IMAGE_ERROR_OK) {
      break;
    }

    file[12] = '6';
    result = decodeImageFileRomFS(IMAGE_FORMAT_PNG, file, &(images[3]), &width, &height);
    if (result != IMAGE_ERROR_OK) {
      break;
    }

    file[12] = '8';
    result = decodeImageFileRomFS(IMAGE_FORMAT_PNG, file, &(images[4]), &width, &height);
    if (result != IMAGE_ERROR_OK) {
      break;
    }

    file[11] = '1';
    file[12] = '0';
    result = decodeImageFileRomFS(IMAGE_FORMAT_PNG, file, &(images[5]), &width, &height);
    if (result != IMAGE_ERROR_OK) {
      break;
    }
  } while (FALSE);
}


/*****************************************************************************
 *
 * Description:
 *    Sets the current progress in percent.
 *
 * Params:
 *    [in] percent   - The progress so far (range 0-100 inclusive)
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void progress_set(uint8_t percent)
{
  int idx = percent / 20;
  if (idx >= NUM_PROGRESS_STEPS) {
    idx = NUM_PROGRESS_STEPS-1;
  }
  swimex_put_image_at_position(pProgressWindow, (COLOR_T*)images[idx],
      (GlobalParameters.display.width - width)/2,
      (GlobalParameters.display.height - height)/2,
      width, height);
}


/*****************************************************************************
 *
 * Description:
 *    Destroys the progress window and unloads the images.
 *
 * Params:
 *    None
 *
 * Return:
 *    None
 *
 ****************************************************************************/
void progress_destroy(void)
{
  int i;
  if (pProgressWindow != NULL) {
    swim_window_close(pProgressWindow);
    lpc_free(pProgressWindow);
    pProgressWindow = NULL;
  }
  for (i = 0; i < NUM_PROGRESS_STEPS; i++) {
    if (images[i] != NULL) {
      lpc_free(images[i]);
      images[i] = NULL;
    }
  }
}

