/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "rom_fs.h"
#include "rom_fsstubs.h"
#include "log.h"


/******************************************************************************
 * Defines, macros, and typedefs
 *****************************************************************************/

#define MAX_OPEN 16

/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static File_t files[MAX_OPEN];
static uint8_t *pFS = NULL;

/******************************************************************************
 * Local function prototypes
 *****************************************************************************/
static FSFile_t * findEntry(FSFile_t * pDir,
                           uint8_t*      pName,
                           uint16_t      nameLen);


/******************************************************************************
 * Public function prototypes
 *****************************************************************************/

void fsInit(void* pImage);

File_t* fsOpen(uint8_t* pName);

uint8_t* fsGetFile(File_t * pFile,
               uint32_t *  pSize);

void fsClose(File_t* pFile);

Dir_t * fsOpenDir(uint8_t * pName);

DirEnt_t * fsReadDir(Dir_t * pDir);

void fsCloseDir(Dir_t* pDir);


/******************************************************************************
 * Implementation of local functions
 *****************************************************************************/

/*****************************************************************************
 *
 * Description:
 *    Locate a file name in an open directory. The search is of one-level
 *    depth.
 *
 * Params:
 *    [in] pDir    - An open directory.
 *    [in] pName   - Name of directory entry. pName does not have to be
 *                   null-terminated.
 *    [in] nameLen - The number of characters in the name.
 *
 * Returns:
 *    Address of the located directory entry or NULL.
 *
 ****************************************************************************/
static FSFile_t *
findEntry(FSFile_t * pDir,
          uint8_t*   pName,
          uint16_t   nameLen)
{
  FSFile_t dummy;
  FSFile_t *pPrev;
  FSFile_t *pEntry;
  uint8_t  *pName2 = NULL;
  uint16_t i       = 0;


  dummy.__fileFlags = 0;
  pPrev  = &dummy;
  pEntry = (FSFile_t *)(pFS + pDir->__fileStart);

  while ((pPrev->__fileFlags & FS_MASK_LAST) != FS_IS_LAST)
  {

    pName2 = pFS + pEntry->__fileNameStart;

    for (i=0;
         i < nameLen;
         i++)
    {
      if (pName[i] != pName2[i])
        break;
    }
    if (i == nameLen &&
        pName2[i] == '\0')
      return pEntry;            /* Match found */

    pPrev = pEntry;
    pEntry++;
  }
  return NULL;                  /* Path component not found */
}

/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/


/*****************************************************************************
 *
 * Description:
 *    Initialize the ROM file system.
 *
 * Params:
 *    [in] pImage - Address of ROM file system image.
 *
 ****************************************************************************/
void
fsInit(void* pImage)
{
  uint16_t i = 0;

  pFS = pImage;

  for(i = 0;
      i < MAX_OPEN;
      i++)
  {
    files[i].__isUsed = FALSE;
  }
}

/*****************************************************************************
 *
 * Description:
 *    Open a file.
 *
 * Params:
 *    [in] pName - Absolute path name to the file that should be opened.
 *                 Directory names end with a '/'.
 *
 * Returns:
 *    Reference to file meta structure
 *
 ****************************************************************************/
File_t*
fsOpen(uint8_t* pName)
{
  FSFile_t *pEntry     = NULL; /* current entry             */
  uint8_t     *pBaseBegin = NULL; /* beginning of name segment */
  uint8_t     *pBaseEnd   = NULL; /* end of name segment       */
  File_t   *pFile      = NULL; /* file structure to return  */
  uint16_t    i           = 0;    /* loop variable             */
  FSFile_t *pDir;              /* current directory         */

  /* Enter critical region */
  m_lock_fs;

  for (i = 0;
       i < MAX_OPEN;
       i++)
  {
    if (files[i].__isUsed == FALSE)
      break;
  }

  if (i >= MAX_OPEN)
  {
    log_e("\r\n*** FS: no free handles!!! ***\r\n");
    m_unlock_fs;
    return NULL;
  }

  pFile           = &files[i];

  /* leave critical region */
  m_unlock_fs;

  pFile->__offset = 0;
  pFile->__isUsed = TRUE;

  pDir   = (FSFile_t *)pFS;
  pEntry = pDir;

  pBaseBegin = pName + 1;

  /* locate the correct directory  */
  while ( (pBaseEnd = (uint8_t*)m_fs_strchr((char*)pBaseBegin, '/')) != NULL)
  {
    pEntry = findEntry(pDir, pBaseBegin, pBaseEnd - pBaseBegin);

    if (pEntry == NULL)
    {
      pFile->__isUsed = FALSE;
      log_i("\r\n*** FS: cannot find file (1) '%s' ***\r\n", pName);
      return NULL;
    }

    pBaseBegin = pBaseEnd + 1;
    pDir       = pEntry;
  }

  /* locate the correct entry in that directory */
  if (*pBaseBegin != '\0')
  {
    pEntry = findEntry(pDir, pBaseBegin, m_fs_strlen((char*)pBaseBegin));

    if (pEntry == NULL)
    {
      log_i("\r\n*** FS: cannot find file (2) '%s' ***\r\n", pName);
      pFile->__isUsed = FALSE;
      return NULL;
    }
  }
  pFile->__pEntry = pEntry;

  return pFile;
}

/*****************************************************************************
 *
 * Description:
 *    Return a reference to the file data, which is located as a continuous
 *    array in memory.
 *
 * Params:
 *    [in]  pFile - Open file.
 *    [out] pSize - The size of the file.
 *
 * Returns:
 *    A refernece to the file data.
 *
 ****************************************************************************/
uint8_t*
fsGetFile(File_t * pFile,
          uint32_t *  pSize)
{
  if (pSize != NULL)
    *pSize = pFile->__pEntry->__fileSize;

  return (pFS + pFile->__pEntry->__fileStart);
}

/*****************************************************************************
 *
 * Description:
 *    Close an open file.
 *
 * Params:
 *    [in] pFile - An open file to be closed.
 *
 ****************************************************************************/
void
fsClose(File_t* pFile)
{
  pFile->__isUsed = FALSE;
}

/*****************************************************************************
 *
 * Description:
 *    Open a directory for reading.
 *
 * Params:
 *    [in] pName - Absolute path name to the directory.
 *
 * Returns:
 *    A reference to the directory.
 *
 ****************************************************************************/
Dir_t *
fsOpenDir(uint8_t * pName)
{
  Dir_t *pDir = NULL;

  pDir = fsOpen(pName);

  if(pDir != NULL)
  {
    pDir->__seenLast = FALSE;
  }

  return pDir;
}

/*****************************************************************************
 *
 * Description:
 *    Fetch the next entry from a directory.
 *
 * Params:
 *    [in] pDir - An open directory.
 *
 * Returns:
 *    A reference to an entry structure. The __d_name member will indicate the
 *    name of this entry.
 *
 ****************************************************************************/
DirEnt_t *
fsReadDir(Dir_t * pDir)
{
  FSFile_t *pEntry = NULL;


  if (pDir->__seenLast == FALSE  &&
      pDir->__offset < pDir->__pEntry->__fileSize)
  {

    pEntry = (FSFile_t *)(pFS + pDir->__pEntry->__fileStart + pDir->__offset);

    pDir->__pName = pFS + pEntry->__fileNameStart;

    if ((pEntry->__fileFlags & FS_MASK_LAST) == FS_IS_LAST)
      pDir->__seenLast = TRUE;


    pDir->__offset += sizeof(FSFile_t);
    pDir->__flags = pEntry->__fileFlags;
    pDir->__size  = pEntry->__fileSize;

    return pDir;
  }

  return NULL;
}

/*****************************************************************************
 *
 * Description:
 *    Checks if a directory entry is a directory or a file.
 *
 * Params:
 *    [in] pDir - A directory entry
 *
 * Returns:
 *    TRUE of the entry is a directory; otherwise FALSE
 *
 ****************************************************************************/
Bool
fsIsEntryDir(DirEnt_t *pDir)
{
  return ((pDir->__flags & FS_MASK_TYPE) == FS_DIR_TYPE);
}

/*****************************************************************************
 *
 * Description:
 *    Get the size of a directory entry
 *
 * Params:
 *    [in] pDir - A directory entry
 *
 * Returns:
 *    the size of the entry
 *
 ****************************************************************************/
uint32_t
fsEntrySize(DirEnt_t *pDir)
{
  return pDir->__size;
}

/*****************************************************************************
 *
 * Description:
 *    Close an open directory.
 *
 * Params:
 *    [in] pDir - The open directory to close.
 *
 ****************************************************************************/
void
fsCloseDir(Dir_t* pDir)
{
  fsClose(pDir);

}

