/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include "lpc_swim.h"
#include "lpc_swim_font.h"

/******************************************************************************
 * Defines, macros, and typedefs
 *****************************************************************************/

/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

/******************************************************************************
 * Local functions
 *****************************************************************************/

static void
plot4points(SWIM_WINDOW_T *win, int32_t cx, int32_t cy, int32_t x, int32_t y, Bool Filled)
{
  int16_t x0, x1, y0, y1;
  y0 = cy + y;
  y1 = cy - y;
  if (Filled) {
    for (x0 = cx - x; x0 <= cx + x; x0++) {
      *(win->fb + x0 + (y0 * win->xpsize)) = win->pen;
      *(win->fb + x0 + (y1 * win->xpsize)) = win->pen;
    }
  } else {
    x0 = cx + x;
    x1 = cx - x;
    *(win->fb + x0 + (y0 * win->xpsize)) = win->pen;
    if (x != 0)
      *(win->fb + x1 + (y0 * win->xpsize)) = win->pen;
    if (y != 0)
      *(win->fb + x0 + (y1 * win->xpsize)) = win->pen;
    if (x != 0 && y != 0)
      *(win->fb + x1 + (y1 * win->xpsize)) = win->pen;
  }
}

static void
plot8points(SWIM_WINDOW_T *win, int32_t cx, int32_t cy, int32_t x, int32_t y, Bool Filled)
{
  plot4points(win, cx, cy, x, y, Filled);
  if (x != y)
    plot4points(win, cx, cy, y, x, Filled);
}

/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/

/******************************************************************************
 *
 * Description:
 *    Draws a circle in the virtual window. Reason for function is that the
 *    SWIM library does not support drawing of circles.
 *
 * Params:
 *    [in] win     - The window identifier
 *    [in] cx      - Virtual X position of the circle's center
 *    [in] cy      - Virtual Y position of the circle's center
 *    [in] radius  - The radius of the circle
 *    [in] filled  - TRUE if the circle should be filled with the PEN color
 *
 * Returns:
 *    None
 *
 *****************************************************************************/
void swimex_put_circle(SWIM_WINDOW_T *win, int32_t cx, int32_t cy, int32_t radius, Bool Filled)
{
  int32_t Error = -radius;
  int16_t x = radius;
  int16_t y = 0;

  while (x >= y) {
    plot8points(win, cx, cy, x, y, Filled);

    Error += y;
    ++y;
    Error += y;

    if (Error >= 0) {
      --x;
      Error -= x;
      Error -= x;
    }
  }
}


/******************************************************************************
 *
 * Description:
 *    Puts an raw image in a window unscaled, clips off edges. The image
 *    is drawn with it's top left corner at the specified position.
 *    The reason for this function is that the swim_put_image function in the
 *    SWIM library only puts the image at 0,0.
 *
 * Params:
 *    [in] win     - The window identifier
 *    [in] image   - The image
 *    [in] xpos    - Virtual X position to put the top left corner of the image
 *    [in] ypos    - Virtual Y position to put the top left corner of the image
 *    [in] xsize   - The width of the image
 *    [in] ysize   - The height of the image
 *
 * Returns:
 *    None
 *
 *****************************************************************************/
void swimex_put_image_at_position(SWIM_WINDOW_T *win, const COLOR_T *image,
    int32_t xpos, int32_t ypos, int32_t xsize, int32_t ysize)
{
  COLOR_T *fb;
  INT_32 x, y;

  y = ypos + win->ypvmin;

  xsize = xsize + xpos + win->xpvmin;
  ysize = ysize + ypos + win->ypvmin;

  /* Move image to window pixel by pixel */
  while ((y <= win->ypvmax) && (y < ysize))
  {
    /* Set physical frame buffer address */
    x = xpos + win->xpvmin;
    fb = win->fb + x + (y * win->xpsize);

    /* Render a single line */
    while ((x <= win->xpvmax) && (x < xsize))
    {
      *fb = *image;
      fb++;
      image++;
      x++;
    }

    /* Adjust to end of line if the image was clipped */
    image = image + (xsize - x);

    y++;
  }
}


/******************************************************************************
 *
 * Description:
 *    Draws the text horizontally centered in the virtual window.
 *
 * Params:
 *    [in] win     - The window identifier
 *    [in] y       - Virtual Y position to put the text
 *    [in] pText   - The text to draw
 *
 * Returns:
 *    None
 *
 *****************************************************************************/
void swimex_put_text_centered(SWIM_WINDOW_T *win, int32_t y, char *pText)
{
  int32_t width = 0;
  int32_t i = 0;
  int32_t charindex;

  /* Continue until the entire string is output */
  while (pText[i] != '\0')
  {
    if (((uint8_t) pText[i] >= win->font->first_char)
        && ((uint8_t) pText[i] <= win->font->last_char))
    {
      charindex = (int32_t) pText[i] - (int32_t) win->font->first_char;

      width += win->font->font_width_table [charindex];
    }
    i++;
  }

  swim_put_text_xy(win, pText, (win->xvsize - width)/2, y);
}


/******************************************************************************
 *
 * Description:
 *    Draws a filled rectangle.
 *
 * Params:
 *    [in] win     - The window identifier
 *    [in] x1      - Virtual X position of the rectangle's top left corner
 *    [in] y1      - Virtual Y position of the rectangle's top left corner
 *    [in] x2      - Virtual X position of the rectangle's bottom right corner
 *    [in] y2      - Virtual Y position of the rectangle's bottom right corner
 *    [in] color   - The color to use
 *
 * Returns:
 *    None
 *
 *****************************************************************************/
void swimex_fill_rect(SWIM_WINDOW_T *win, int x1, int y1, int x2, int y2, COLOR_T color)
{
  swim_set_pen_color(win, color);
  swim_set_fill_color(win, color);
  swim_put_box(win, x1, y1, x2, y2);
}


/******************************************************************************
 *
 * Description:
 *    Draws a filled circle.
 *
 * Params:
 *    [in] win     - The window identifier
 *    [in] x1      - Virtual X position of the circle's center
 *    [in] y1      - Virtual Y position of the circle's center
 *    [in] radius  - The radius of the circle
 *    [in] color   - The color to use
 *
 * Returns:
 *    None
 *
 *****************************************************************************/
void swimex_fill_circle(SWIM_WINDOW_T *win, int x, int y, int radius, COLOR_T color)
{
  swim_set_pen_color(win, color);
  swimex_put_circle(win, x, y, radius, TRUE);
}


/******************************************************************************
 *
 * Description:
 *    Draws a line with the specified color.
 *
 * Params:
 *    [in] win     - The window identifier
 *    [in] x1      - Virtual X position of the line's starting point
 *    [in] y1      - Virtual Y position of the line's starting point
 *    [in] x2      - Virtual X position of the line's end point
 *    [in] y2      - Virtual Y position of the line's end point
 *    [in] color   - The color to use
 *
 * Returns:
 *    None
 *
 *****************************************************************************/
void swimex_put_colored_line(SWIM_WINDOW_T *win, int x1, int y1, int x2, int y2, COLOR_T color)
{
  swim_set_pen_color(win, color);
  swim_put_line(win, x1, y1, x2, y2);
}

