/*****************************************************************************
 *
 *   Copyright(C) 2012, Embedded Artists AB
 *   All rights reserved.
 *
 ******************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * Embedded Artists AB assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. Embedded Artists AB
 * reserves the right to make changes in the software without
 * notification. Embedded Artists AB also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *****************************************************************************/

/******************************************************************************
 * Includes
 *****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include "trans_fs.h"
#include "rom_fs.h"
#include "ff.h"
#include "log.h"
#include "lpc_heap.h"
#include "pca9532.h"

/******************************************************************************
 * Typedefs and defines
 *****************************************************************************/

/******************************************************************************
 * Public variables
 *****************************************************************************/

/******************************************************************************
 * Local variables
 *****************************************************************************/

static Bool fatfsMounted = FALSE;
static FATFS lFatfs;

/******************************************************************************
 * Local functions
 *****************************************************************************/


/*****************************************************************************
 *
 * Description:
 *    Checks if a MMC card is inserted or not.
 *
 * Params:
 *    None
 *
 * Return:
 *    TRUE if a MMC card is present
 *
 ****************************************************************************/
static Bool isMmcInserted(void)
{
  /* The CD (CardDetect) signal is connected to the PCA9532 Port Expander */
  return (pca9532_getLedState(FALSE) & MMC_CD);
}


/*****************************************************************************
 *
 * Description:
 *    Allocates a new buffer and loads the specified file (located on the
 *    ROM file system) into that buffer.
 *
 * Params:
 *    [in]  pFileName  - the name of the file to locate.
 *    [out] ppDest     - will point to newly allocated memory with the file.
 *    [out] pSize      - the file size and number of bytes stored in ppDest.
 *
 * Return:
 *    TFS_OK if the file was located and its contents stored in ppDest,
 *    otherwise one of the TFS_ERR_* error codes.
 *
 ****************************************************************************/
static tTransFsError loadRomfsFileIntoNewBuff(char* pFileName, uint8_t** ppDest,
    uint32_t* pSize)
{
  tTransFsError result = TFS_OK;
  uint8_t* pBufIn = NULL;
  File_t* fileHandle = NULL;
  uint32_t fileSize = 0;

  fileHandle = fsOpen((uint8_t*)pFileName);
  if(fileHandle != NULL)
  {
    pBufIn = fsGetFile(fileHandle, &fileSize);
    if (pBufIn != NULL)
    {
      *ppDest = lpc_new(fileSize);
      if (*ppDest != NULL)
      {
        memcpy(*ppDest, pBufIn, fileSize);
      }
      else
      {
         result = TFS_ERR_MALLOC_FAILED;
      }
      *pSize = fileSize;
    }

    fsClose(fileHandle);
  }
  else
  {
    result = TFS_ERR_NO_SUCH_FILE;
  }

  return result;
}


/*****************************************************************************
 *
 * Description:
 *    Allocates a new buffer and loads the specified file (located on the
 *    FAT file system) into that buffer.
 *
 *    The function will check if a MMC card is present and initialize that
 *    file system if needed.
 *
 * Params:
 *    [in]  pFileName  - the name of the file to locate.
 *    [out] ppDest     - will point to newly allocated memory with the file.
 *    [out] pSize      - the file size and number of bytes stored in ppDest.
 *
 * Return:
 *    TFS_OK if the file was located and its contents stored in ppDest,
 *    otherwise one of the TFS_ERR_* error codes.
 *
 ****************************************************************************/
static tTransFsError loadFatfsFileIntoNewBuff(char* pFileName, uint8_t** ppDest,
    uint32_t* pSize)
{
  tTransFsError result = TFS_OK;
  uint8_t* pBuf = NULL;
  FRESULT fsResult = FR_OK;
  FIL fileHandle;
  uint32_t fileSize = 0;
  uint32_t bytesRead = 0;
  uint32_t pos = 0;

  if (!isMmcInserted())
  {
    fatfsMounted = FALSE;
    return TFS_ERR_NO_SUCH_FS;
  }
  if (!fatfsMounted) {
    fsResult = f_mount(0, &lFatfs);
    if (fsResult != FR_OK) {
      log_e("trfs: Failed to mount file system\r\n");
      return TFS_ERR_NO_SUCH_FS;
    }
    fatfsMounted = TRUE;
  }

  fsResult = f_open(&fileHandle, pFileName, FA_READ);
  if(fsResult == FR_OK)
  {
    fileSize = fileHandle.fsize;
    *ppDest = lpc_new(fileSize);
    if (*ppDest == NULL)
    {
      f_close(&fileHandle);
      return TFS_ERR_MALLOC_FAILED;
    }
    pBuf = *ppDest;
    *pSize = fileSize;
  }
  else
  {
    result = TFS_ERR_NO_SUCH_FILE;
  }

  if (fsResult == FR_OK && pBuf != NULL)
  {
    pos = 0;
    while(fileSize > 0)
    {
      fsResult = f_read(&fileHandle, &pBuf[pos], fileHandle.fsize, &bytesRead);
      if(fsResult != FR_OK)
      {
        log_e("trfs: failed to read from file\n");
        lpc_free(pBuf);
        pBuf = NULL;
        *ppDest = NULL;
        result = TFS_ERR_INVALID_INPUT;
        break;
      }

      fileSize -= bytesRead;
      pos += bytesRead;
    }

    f_close(&fileHandle);
  }

  return result;
}


/*****************************************************************************
 *
 * Description:
 *    Creates a new file on the FAT file system (possibly overwriting an
 *    old one) and writes the buffer data into it.
 *
 *    The function will check if a MMC card is present and initialize that
 *    file system if needed.
 *
 * Params:
 *    [in] pFileName  - the name of the file to write to.
 *    [in] pSrc       - the data to store
 *    [in] size       - the number of bytes to store
 *
 * Return:
 *    TFS_OK if the file was created and all bytes written,
 *    otherwise one of the TFS_ERR_* error codes.
 *
 ****************************************************************************/
static tTransFsError
storeBuffIntoFatfsFile(char* pFileName, uint8_t* pSrc, uint32_t size)
{
  tTransFsError result = TFS_OK;
  FRESULT fsResult = FR_OK;
  FIL fileHandle;
  uint8_t* pData = pSrc;
  uint32_t bytesLeft = size;
  uint32_t bytesWritten = 0;

  if (!isMmcInserted())
  {
    fatfsMounted = FALSE;
    return TFS_ERR_NO_SUCH_FS;
  }
  if (!fatfsMounted) {
    fsResult = f_mount(0, &lFatfs);
    if (fsResult != FR_OK) {
      log_e("trfs: Failed to mount file system\r\n");
      return TFS_ERR_NO_SUCH_FS;
    }
    fatfsMounted = TRUE;
  }

  fsResult = f_open(&fileHandle, pFileName, FA_CREATE_ALWAYS|FA_WRITE);
  if(fsResult == FR_OK)
  {
    while (fsResult == FR_OK && bytesLeft > 0) {
      fsResult = f_write(&fileHandle, pData, bytesLeft, &bytesWritten);
      bytesLeft -= bytesWritten;
      pData += bytesWritten;
    }
    f_close(&fileHandle);
  }
  return result;
}


/******************************************************************************
 * Implementation of public functions
 *****************************************************************************/


/*****************************************************************************
 *
 * Description:
 *    The function will check if a MMC card is present and initialize that
 *    file system if needed.
 *
 * Params:
 *    None
 *
 * Return:
 *    TFS_OK if a MMC card is present and initialized,
 *    otherwise one of the TFS_ERR_* error codes.
 *
 ****************************************************************************/
tTransFsError transfs_initializeFATFS(void)
{
  if (!isMmcInserted())
  {
    fatfsMounted = FALSE;
    return TFS_ERR_NO_SUCH_FS;
  }
  if (!fatfsMounted) {
    FRESULT fsResult = f_mount(0, &lFatfs);
    if (fsResult != FR_OK) {
      log_e("trfs: Failed to mount file system\r\n");
      return TFS_ERR_NO_SUCH_FS;
    }
    fatfsMounted = TRUE;
  }
  return TFS_OK;
}


/*****************************************************************************
 *
 * Description:
 *    Allocates a new buffer and loads the specified file into that buffer.
 *    The value of the behavior parameter determines the search order when
 *    locating the file.
 *
 * Params:
 *    [in]  pFileName  - the name of the file to locate.
 *    [in]  behavior   - what should be done if file is found i multiple locations.
 *    [out] ppDest     - will point to newly allocated memory with the file.
 *    [out] pSize      - the file size and number of bytes stored in ppDest.
 *    [out] pFsType    - where was the file located.
 *
 * Return:
 *    TFS_OK if the file was located and its contents stored in ppDest,
 *    otherwise one of the TFS_ERR_* error codes.
 *
 ****************************************************************************/
tTransFsError transfs_loadFileIntoNewBuff(char* pFileName, tFsOverride behavior,
    uint8_t** ppDest, uint32_t* pSize, tFsFileSystem* pFsType)
{
  tTransFsError result;
  char path[50];

  switch (behavior)
  {
  case ROMFS_OVERRIDES_FATFS:
    {
      result = transfs_loadFileIntoNewBuff(pFileName, ONLY_USE_ROMFS, ppDest, pSize, pFsType);
      if (result != TFS_OK)
      {
        result = transfs_loadFileIntoNewBuff(pFileName, ONLY_USE_FATFS, ppDest, pSize, pFsType);
      }
    }
    break;

  case FATFS_OVERRIDES_ROMFS:
    {
      result = transfs_loadFileIntoNewBuff(pFileName, ONLY_USE_FATFS, ppDest, pSize, pFsType);
      if (result != TFS_OK)
      {
        result = transfs_loadFileIntoNewBuff(pFileName, ONLY_USE_ROMFS, ppDest, pSize, pFsType);
      }
    }
    break;

  case ONLY_USE_ROMFS:
    {
      result = loadRomfsFileIntoNewBuff(pFileName, ppDest, pSize);
      *pFsType = TFS_FS_ROM;
    }
    break;

  case ONLY_USE_FATFS:
    {
      path[0] = '\0';
      strcpy(path, FATFS_PATH_PREFIX);
      strcat(path, pFileName);
      result = loadFatfsFileIntoNewBuff(&path[0], ppDest, pSize);
      *pFsType = TFS_FS_FAT;
    }
    break;

  default:
    result = TFS_ERR_INVALID_OVERRIDE;
  }

  return result;
}


/*****************************************************************************
 *
 * Description:
 *    Creates a new file (possibly overwriting an old one) and writes the
 *    buffer data into it.
 *
 * Params:
 *    [in] pFileName  - the name of the file to write to.
 *    [in] fsType     - which file system to store the file in
 *    [in] pBuff      - the data to store
 *    [in] buffSize   - the number of bytes in buff
 *
 * Return:
 *    TFS_OK if the file was created and all bytes written,
 *    otherwise one of the TFS_ERR_* error codes.
 *
 ****************************************************************************/
tTransFsError transfs_storeBuffInFile(char* pFileName, tFsFileSystem fsType,
    uint8_t* pBuff, uint32_t buffSize)
{
  tTransFsError result;
  char path[50];

  switch (fsType)
  {
  case TFS_FS_FAT:
    path[0] = '\0';
    strcpy(path, FATFS_PATH_PREFIX);
    strcat(path, pFileName);
    result = storeBuffIntoFatfsFile(&path[0], pBuff, buffSize);
    break;

  case TFS_FS_ROM:
    result = TFS_ERR_FS_IS_READONLY;
    break;

  default:
    result = TFS_ERR_NO_SUCH_FS;
  }

  return result;
}
